@@ -1,3 +1,5 @@
+# -*- coding: latin-1 -*-
+
 """ BeeStorage - Flatfile data storage facility.
 
     Definitions:
@@ -6,7 +8,7 @@
       record: Header + content + padding
 
     Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 
@@ -51,7 +53,7 @@ HOT = 1
 COLD = 0
 
 # Output debugging info
-__debug__ = 1
+_debug = 0
 
 ### Errors
 
@@ -155,7 +157,7 @@ class BeeStorage:
 
         self.readonly = readonly
 
-        if __debug__:
+        if _debug:
             log.call(SYSTEM_DEBUG)
 
         # Adjust min_recordsize
@@ -195,7 +197,7 @@ class BeeStorage:
             if readonly:
                 raise IOError,why
             # New file: write header and state marker
-            if __debug__:
+            if _debug:
                 log(SYSTEM_INFO,'Creating a new storage file %s' % filename)
             self.file = file = open(filename,'w+b')
             self.write_fileheader(file)
@@ -259,7 +261,7 @@ class BeeStorage:
             return
         if self.readonly:
             raise Error,'storage is read-only'
-        if __debug__:
+        if _debug:
             log(SYSTEM_DEBUG,
                 'Marking the file "%s": %s',
                 self.filename,((state==HOT)*'HOT' or 'COLD'))
@@ -413,7 +415,7 @@ class BeeStorage:
 
     def __del__(self):
 
-        if __debug__:
+        if _debug:
             log(SYSTEM_DEBUG,'del %s',self)
         if self.file:
             self.close()
@@ -432,7 +434,7 @@ class BeeStorage:
 
     def read_header(self,position,
 
-                    unpack=struct.unpack,BLOCKSIZE=BLOCKSIZE,None=None,
+                    unpack=struct.unpack,BLOCKSIZE=BLOCKSIZE,
                     ID=ID,headertypes=(OLD,VALID)):
 
         """ Read the header located at position and return
@@ -500,7 +502,7 @@ class BeeStorage:
 
     def write_record(self,data,position,minsize=0,rtype=VALID,
 
-                     BLOCKSIZE=BLOCKSIZE,None=None,pack=struct.pack,
+                     BLOCKSIZE=BLOCKSIZE,pack=struct.pack,
                      join=string.join,ID=ID,HOT=HOT):
 
         """ Write a record of given rtype (defaults to VALID)
@@ -529,7 +531,7 @@ class BeeStorage:
             if datasize < datalen:
                 # Drop old record
                 if rtype == VALID:
-                    if __debug__:
+                    if _debug:
                         log(SYSTEM_INFO,
                             'Could not update %i record in place: '
                             'old size = %i, datalen = %i, data = %s',
@@ -540,7 +542,7 @@ class BeeStorage:
             else:
                 # Write new data
                 file.write(data)
-                if __debug__:
+                if _debug:
                     log(SYSTEM_INFO,'Data written to position %i: %s',
                         position,data)
                 return position
@@ -554,7 +556,7 @@ class BeeStorage:
 
         # Write the header + data + padding
         file.seek(position)
-        file.write(join((ID,pack('l',recordsize),rtype,         # Header
+        file.write(join((ID,pack('<l',recordsize),rtype,        # Header
                          data,                                  # Data
                          '\0' * (recordsize - 6 - datalen)      # Padding
                          ),''))
@@ -563,7 +565,7 @@ class BeeStorage:
 
         # Update EOF
         self.EOF = file.tell()
-        if __debug__:
+        if _debug:
             log(SYSTEM_DEBUG,'New EOF = %i',self.EOF)
             log(SYSTEM_DEBUG,'Data written to position %i: %s',
                 position,data)
@@ -611,9 +613,7 @@ class BeeStorage:
     delete = free
     __delitem__ = free
 
-    def write(self,obj,position=None,
-
-              None=None):
+    def write(self,obj,position=None):
 
         """ Write the encoded object to the file and
             return the file address where the data was written.
@@ -1085,7 +1085,7 @@ freeze(BeeKeyValueStorage)
 
 if __name__ == '__main__':
 
-    f = BeePickleStorage('test-BeePickleStorage.dat',cache=1)
+    f = BeePickleStorage('test-BeePickleStorage.dat',cache=1,lock=1)
     l = [1,'blabla','Hi there',2.3434,4+7j] + range(1000)
     k = map(f.write,l)
     m = map(f.read,k)
