@@ -20,25 +20,14 @@
 
 */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <string.h>
-
-#include "btr.h"
-
-#ifdef MAL_DEBUG
-
 /* Logging file used by debugging facility */
-# ifndef MAL_DEBUG_OUTPUTFILE
-#  define MAL_DEBUG_OUTPUTFILE "mxBeeBase.log"
-# endif
-
-# include "mx.h"
-#else
-# define DPRINTF printf
+#ifndef MAL_DEBUG_OUTPUTFILE
+# define MAL_DEBUG_OUTPUTFILE "mxBeeBase.log"
 #endif
 
+#include "mxstdlib.h"
+#include "btr.h"
+
 /* --- Globals ------------------------------------------------------------ */
 
 /* line number for last IO or memory error */
@@ -245,15 +234,19 @@ void dumpBuf(bHandle *h,
         return;
     }
     k = key(fkey(buf));
-    DPRINTF("\n%s: buf[%04x], ct=%d, leaf=%d", 
-        msg, buf->adr, ct(buf), leaf(buf));
-    if (childLT(k)) DPRINTF(", LT(%04x)", childLT(k));
-    if (leaf(buf)) DPRINTF(", prev(%04x), next(%04x)", prev(buf), next(buf));
+    DPRINTF("\n%s: buf[%04lx], ct=%d, leaf=%d", 
+            msg, (unsigned long)buf->adr, ct(buf), leaf(buf));
+    if (childLT(k)) 
+	DPRINTF(", LT(%04lx)", childLT(k));
+    if (leaf(buf)) 
+	DPRINTF(", prev(%04lx), next(%04lx)", 
+		(unsigned long)prev(buf), (unsigned long)next(buf));
     DPRINTF("\n");
     for (i = 0; i < ct(buf); i++) {
-        DPRINTF("  key %3d: %08x rec(%08x)",
-          i, *(int *)key(k), rec(k));
-        if (childGE(k)) DPRINTF(" GE(%04x)", childGE(k));
+        DPRINTF("  key %3d: %08x rec(%08lx)",
+		i, *(int *)key(k), rec(k));
+        if (childGE(k)) 
+	    DPRINTF(" GE(%04lx)", (unsigned long)childGE(k));
         DPRINTF("\n");
         k += ks(1);
     }
@@ -327,47 +320,55 @@ int _validateTree(bHandle *h,
     buf->p = (bNode *)p;
     dumpBuf(h,"validate", buf);
     if (visited[buf->adr >> 8]) {
-        DPRINTF("previous visit, buf[%04x]\n", buf->adr);
+        DPRINTF("previous visit, buf[%04lx]\n", (unsigned long)buf->adr);
         return -1;
     }
     visited[buf->adr >> 8] = 1;
     DPRINTF("\n");
     if (ct(buf)) {
         if (!leaf(buf)) {
-            DPRINTF("level %d: recursing on buf[%04x] LT\n", level, childLT(fkey(buf)));
+            DPRINTF("level %d: recursing on buf[%04lx] LT\n", 
+		    level, childLT(fkey(buf)));
             if ((rc = readDisk(h, childLT(fkey(buf)), &cbuf)) != 0) {
-                DPRINTF("unable to read buffer %04x\n", childLT(fkey(buf)));
+                DPRINTF("unable to read buffer %04lx\n", 
+			childLT(fkey(buf)));
                 return -1;
             }
             if (*(unsigned int *)key(lkey(cbuf)) > *(unsigned int *)key(fkey(buf))) {
-                DPRINTF("last element in child buf[%04x] LT > first element of parent buf[%04x]\n", cbuf->adr, buf->adr);
+                DPRINTF("last element in child buf[%04lx] LT "
+			"> first element of parent buf[%04lx]\n", 
+			(unsigned long)cbuf->adr, 
+			(unsigned long)buf->adr);
                 return -1;
             }
             _validateTree(h, cbuf, visited, level+1);
             k = fkey(buf);
             for (i = 0; i < ct(buf); i++) {
-                DPRINTF("level %d: recursing on buf[%04x] GE[%d]\n", level, key(childGE(k)), i);
+                DPRINTF("level %d: recursing on buf[%04lx] GE[%d]\n", 
+			level, key(childGE(k)), i);
                 if ((rc = readDisk(h, childGE(k), &cbuf)) != 0) {
-                    DPRINTF("unable to read buffer %04x\n", childGE(k));
+                    DPRINTF("unable to read buffer %04lx\n", childGE(k));
                     return -1;
                 }
                 if (*(unsigned int *)key(fkey(cbuf)) < *(unsigned int *)key(k)) {
-                    DPRINTF("first element in child buf[%04x] < parent buf[%04x] GE (%08x < %08x)\n", 
-                    cbuf->adr,
-                    buf->adr,
-                    *(int *)key(fkey(cbuf)),
-                    *(int *)key(k));
+                    DPRINTF("first element in child buf[%04lx] "
+			    "< parent buf[%04lx] GE (%08x < %08x)\n", 
+			    (unsigned long)cbuf->adr,
+			    (unsigned long)buf->adr,
+			    *(int *)key(fkey(cbuf)),
+			    *(int *)key(k));
 
                     dumpBuf(h,"buf", buf);
                     dumpBuf(h,"cbuf", cbuf);
                     return -1;
                 }
                 if (!leaf(cbuf) && *(unsigned int *)key(fkey(cbuf)) ==  *(unsigned int *)key(k)) {
-                    DPRINTF("first element in child buf[%04x] = parent buf[%04x] GE (%08x < %08x)\n",
-                    cbuf->adr,
-                    buf->adr,
-                    *(int *)key(fkey(cbuf)),
-                    *(int *)key(k));
+                    DPRINTF("first element in child buf[%04lx] "
+			    "= parent buf[%04lx] GE (%08x < %08x)\n",
+			    (unsigned long)cbuf->adr,
+			    (unsigned long)buf->adr,
+			    *(int *)key(fkey(cbuf)),
+			    *(int *)key(k));
                     dumpBuf(h,"buf", buf);
                     dumpBuf(h,"cbuf", cbuf);
                     return -1;
@@ -462,7 +463,7 @@ int search(bHandle *h,
     /* Handle set of duplicates */
     if (h->dupKeys && (mode == MODE_FIRST) && foundDup) {
 #if 0
-	printf("found dups: cc=%i, lb=%i, ub=%i, key(*mkey)=%i next=%i\n",
+	DPRINTF("found dups: cc=%i, lb=%i, ub=%i, key(*mkey)=%i next=%i\n",
 	       cc,lb,ub,*(int*)key(*mkey),*(int*)key((*mkey+ks(1))));
 #endif
         if (cc == CC_GT)
@@ -954,7 +955,7 @@ bError bFindKey(bHandle *h,
 		c->key = mkey;
                 return bErrOk;
             } else {
-		printf("not found; cc=%i\n", cc);
+		DPRINTF("not found; cc=%i\n", cc);
 		
                 return bErrKeyNotFound;
             }
@@ -1385,8 +1386,8 @@ int bValidateTree(bHandle *h)
     if (!visited)
 	return -1;
     flushAll(h);
-    DPRINTF("Validating BTree with handle %0x, root buffer at %0x",
-	    (int)h,(int)&h->root);
+    DPRINTF("Validating BTree with handle %0lx, root buffer at %0lx",
+	    (long)h,(long)&h->root);
     return _validateTree(h,&h->root,visited,1);
 }
 
@@ -1402,7 +1403,7 @@ int comp(const void *key1, const void *key2) {
 
 #define DO(xyz) \
     if ((rc = xyz) != bErrOk) {				\
-        printf("Error in line %d: rc = %d\n", __LINE__, rc);	\
+        DPRINTF("Error in line %d: rc = %d\n", __LINE__, rc);	\
         exit(0);					\
     }
 
@@ -1429,10 +1430,10 @@ int main(void) {
     DO (bInsertKey(handle, &key, value));
 
     DO (bFindKey(handle, &c, &key, &value));
-    printf("Found key %x with value %x\n",key,value);
+    DPRINTF("Found key %x with value %x\n",key,value);
 
     bFlush(handle);
-    printf("Buffers flushed.\n");
+    DPRINTF("Buffers flushed.\n");
 
     key = 0x222;
     value = 0xbeef;
@@ -1440,44 +1441,44 @@ int main(void) {
 
     key = 0x123;
     DO (bFindKey(handle, &c, &key, &value));
-    printf("Found key %x with value %x\n",key,value);
+    DPRINTF("Found key %x with value %x\n",key,value);
 
     key = 0x222;
     DO (bFindKey(handle, &c, &key, &value));
-    printf("Found key %x with value %x\n",key,value);
+    DPRINTF("Found key %x with value %x\n",key,value);
 
-    printf("statistics:\n");
-    printf("    maximum height: %8d\n", handle->maxHeight);
-    printf("    nodes inserted: %8d\n", handle->nNodesIns);
-    printf("    nodes deleted:  %8d\n", handle->nNodesDel);
-    printf("    keys inserted:  %8d\n", handle->nKeysIns);
-    printf("    keys deleted:   %8d\n", handle->nKeysDel);
-    printf("    disk reads:     %8d\n", handle->nDiskReads);
-    printf("    disk writes:    %8d\n", handle->nDiskWrites);
+    DPRINTF("statistics:\n");
+    DPRINTF("    maximum height: %8d\n", handle->maxHeight);
+    DPRINTF("    nodes inserted: %8d\n", handle->nNodesIns);
+    DPRINTF("    nodes deleted:  %8d\n", handle->nNodesDel);
+    DPRINTF("    keys inserted:  %8d\n", handle->nKeysIns);
+    DPRINTF("    keys deleted:   %8d\n", handle->nKeysDel);
+    DPRINTF("    disk reads:     %8d\n", handle->nDiskReads);
+    DPRINTF("    disk writes:    %8d\n", handle->nDiskWrites);
 
     bClose(handle);
 
     /* Second time... */
-    printf("\nSecond time...\n\n");
+    DPRINTF("\nSecond time...\n\n");
 
     DO (bOpen(info, &handle));
 
     key = 0x123;
     DO (bFindKey(handle, &c, &key, &value));
-    printf("Found key %x with value %x\n",key,value);
+    DPRINTF("Found key %x with value %x\n",key,value);
 
     key = 0x222;
     DO (bFindKey(handle, &c, &key, &value));
-    printf("Found key %x with value %x\n",key,value);
+    DPRINTF("Found key %x with value %x\n",key,value);
 
-    printf("statistics:\n");
-    printf("    maximum height: %8d\n", handle->maxHeight);
-    printf("    nodes inserted: %8d\n", handle->nNodesIns);
-    printf("    nodes deleted:  %8d\n", handle->nNodesDel);
-    printf("    keys inserted:  %8d\n", handle->nKeysIns);
-    printf("    keys deleted:   %8d\n", handle->nKeysDel);
-    printf("    disk reads:     %8d\n", handle->nDiskReads);
-    printf("    disk writes:    %8d\n", handle->nDiskWrites);
+    DPRINTF("statistics:\n");
+    DPRINTF("    maximum height: %8d\n", handle->maxHeight);
+    DPRINTF("    nodes inserted: %8d\n", handle->nNodesIns);
+    DPRINTF("    nodes deleted:  %8d\n", handle->nNodesDel);
+    DPRINTF("    keys inserted:  %8d\n", handle->nKeysIns);
+    DPRINTF("    keys deleted:   %8d\n", handle->nKeysDel);
+    DPRINTF("    disk reads:     %8d\n", handle->nDiskReads);
+    DPRINTF("    disk writes:    %8d\n", handle->nDiskWrites);
 
     bClose(handle);
 
