@@ -2,7 +2,7 @@
   mxBeeBase -- BeeBase C extension
 
   Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author (mailto:mal@lemburg.com).
 
@@ -20,11 +20,14 @@
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXBEEBASE
 
+/* Mark the module as Py_ssize_t clean. */
+#define PY_SSIZE_T_CLEAN 1
+
 #include "mx.h"
 #include "mxBeeBase.h"
 
 /* Version number: Major.Minor.Patchlevel */
-#define VERSION "2.0.2"
+#define MXBEEBASE_VERSION "3.0.0"
 
 /* Define these to have the module use free lists (saves malloc calls) */
 /*#define MXBEEINDEX_FREELIST*/
@@ -34,10 +37,10 @@
 
 static char *Module_docstring = 
 
- MXBEEBASE_MODULE" -- BeeBase objects and functions. Version "VERSION"\n\n"
+ MXBEEBASE_MODULE" -- BeeBase objects and functions. Version "MXBEEBASE_VERSION"\n\n"
 
  "Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com\n"
- "Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
+ "Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -62,6 +65,9 @@ static mxBeeIndexObject *mxBeeIndex_FreeList = NULL;
 static mxBeeCursorObject *mxBeeCursor_FreeList = NULL;
 #endif
 
+/* Flag telling us whether the module was initialized or not. */
+static int mxBeeBase_Initialized = 0;
+
 /* --- forward declarations ----------------------------------------------- */
 
 staticforward PyTypeObject mxBeeIndex_Type;
@@ -71,7 +77,7 @@ staticforward PyTypeObject mxBeeCursor_Type;
 staticforward PyMethodDef mxBeeCursor_Methods[];
 
 staticforward
-mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *index,   /* Index object */
+mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *beeindex,   /* Index object */
 				   bCursor *c                 /* bCursor */
 				   );
 
@@ -95,8 +101,31 @@ PyObject *insexc(PyObject *moddict,
 {
     PyObject *v;
     char fullname[256];
+    char *modname;
+    char *dot;
     
-    sprintf(fullname,MXBEEBASE_MODULE".%s",name);
+    v = PyDict_GetItemString(moddict, "__name__");
+    if (v == NULL)
+	modname = NULL;
+    else
+	modname = PyString_AsString(v);
+    if (modname == NULL) {
+	PyErr_Clear();
+	modname = MXBEEBASE_MODULE;
+    }
+    /* The symbols from this extension are imported into
+       mx.<packagename>. We trim the name to not confuse the user with
+       an overly long package path. */
+    strcpy(fullname, modname);
+    dot = strchr(fullname, '.');
+    if (dot)
+	dot = strchr(dot+1, '.');
+    if (dot)
+	strcpy(dot+1, name);
+    else
+	sprintf(fullname, "%s.%s", modname, name);
+
+
     v = PyErr_NewException(fullname, NULL, NULL);
     if (v == NULL)
 	return NULL;
@@ -183,7 +212,7 @@ void mxBeeBase_ReportError(bError rc)
 
     case bErrIO:
 	Py_ErrorWith2Args(PyExc_IOError,
-			  "in BeeIndex: '%s' (line %i)",strerror(errno),
+			  "in BeeIndex: '%s' (btr.c line %i)",strerror(errno),
 			  bErrLineNo);
 
     case bErrMemory:
@@ -192,7 +221,7 @@ void mxBeeBase_ReportError(bError rc)
 
     default:
 	Py_Error(PyExc_SystemError,
-		 "unkown error");
+		 "unknown error");
     }
  onError:
     return;
@@ -203,11 +232,11 @@ void mxBeeBase_ReportError(bError rc)
 /* --- Allocation --- */
 
 static
-mxBeeIndexObject *mxBeeIndex_New(char *filename,   /* name of index file */
+mxBeeIndexObject *mxBeeIndex_New(char *filename,   /* name of beeindex file */
 				 int filemode,	   /* how to open the
 						      file */
-				 int keySize,      /* length, in
-						      bytes, of key */
+				 int keySize,      /* length of key in
+						      bytes */
 				 int sectorSize,   /* size of sector
 						      on disk */
 				 bCompFunc comp,   /* key compare
@@ -219,7 +248,7 @@ mxBeeIndexObject *mxBeeIndex_New(char *filename,   /* name of index file */
 						      keys ? */
 				 )
 {
-    mxBeeIndexObject *index;
+    mxBeeIndexObject *beeindex = 0;
     bDescription *info;
     bError rc;
     char *iName = strdup(filename);
@@ -231,21 +260,21 @@ mxBeeIndexObject *mxBeeIndex_New(char *filename,   /* name of index file */
     /* Allocate the object */
 #ifdef MXBEEINDEX_FREELIST
     if (mxBeeIndex_FreeList) {
-	index = mxBeeIndex_FreeList;
+	beeindex = mxBeeIndex_FreeList;
 	mxBeeIndex_FreeList = *(mxBeeIndexObject **)mxBeeIndex_FreeList;
-	index->ob_type = &mxBeeIndex_Type;
-	_Py_NewReference(index);
+	beeindex->ob_type = &mxBeeIndex_Type;
+	_Py_NewReference(beeindex);
     }
     else
 #endif 
 	 {
-	index = PyObject_NEW(mxBeeIndexObject,&mxBeeIndex_Type);
-	if (index == NULL)
+	beeindex = PyObject_NEW(mxBeeIndexObject,&mxBeeIndex_Type);
+	if (beeindex == NULL)
 	    goto onError;
     }
 
     /* Init description */
-    info = &index->info;
+    info = &beeindex->info;
     info->iName = iName;
     info->keySize = keySize;
     info->dupKeys = (allow_dupkeys != 0);
@@ -254,52 +283,54 @@ mxBeeIndexObject *mxBeeIndex_New(char *filename,   /* name of index file */
     info->filemode = filemode;
 
     /* Conversion routines */
-    index->ObjectFromKey = ofk;
-    index->KeyFromObject = kfo;
+    beeindex->ObjectFromKey = ofk;
+    beeindex->KeyFromObject = kfo;
 
     /* Reset update count (also see mxBeeIndex_Clear()) */
-    index->updates = 0;
+    beeindex->updates = 0;
 
     /* Invalidate length cache */
-    index->length = -1;
-    index->length_state = -1;
+    beeindex->length = -1;
+    beeindex->length_state = -1;
     
-    /* Open the index */
-    rc = bOpen(index->info,&(index->handle));
+    /* Open the beeindex */
+    rc = bOpen(beeindex->info, &(beeindex->handle));
     if (rc != bErrOk) {
+	beeindex->handle = 0;
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
 
-    DPRINTF("mxBeeIndex_New: instance at %0x\n",(int)index);
+    DPRINTF("mxBeeIndex_New: instance at %0lx\n",(long)beeindex);
 
-    return index;
+    return beeindex;
 
  onError:
+    Py_XDECREF(beeindex);
     return NULL;
 }
 
 /* --- Deallocation --- */
 
 static
-void mxBeeIndex_Free(mxBeeIndexObject *index)
+void mxBeeIndex_Free(mxBeeIndexObject *beeindex)
 {
-    DPRINTF("mxBeeIndex_Free: instance at %0x\n",(int)index);
+    DPRINTF("mxBeeIndex_Free: instance at %0lx\n",(long)beeindex);
 
-    if (index->handle)
-	/* Close index file, flushing any unsaved data */
-	bClose(index->handle);
+    if (beeindex->handle)
+	/* Close beeindex file, flushing any unsaved data */
+	bClose(beeindex->handle);
 
     /* Free filename */
-    free(index->info.iName);
-    index->info.iName = NULL;
+    free(beeindex->info.iName);
+    beeindex->info.iName = NULL;
     
 #ifdef MXBEEINDEX_FREELIST
     /* Append to free list */
-    *(mxBeeIndexObject **)index = mxBeeIndex_FreeList;
-    mxBeeIndex_FreeList = index;
+    *(mxBeeIndexObject **)beeindex = mxBeeIndex_FreeList;
+    mxBeeIndex_FreeList = beeindex;
 #else
-    PyMem_DEL(index);
+    PyObject_Del(beeindex);
 #endif
 }
 
@@ -307,16 +338,16 @@ void mxBeeIndex_Free(mxBeeIndexObject *index)
 
 /* Use Python strings as keys */
 static
-void *mxBeeIndex_KeyFromString(mxBeeIndexObject *index,
+void *mxBeeIndex_KeyFromString(mxBeeIndexObject *beeindex,
 			       PyObject *key)
 {
     Py_Assert(PyString_Check(key),
 	      PyExc_TypeError,
 	      "keys must be strings");
-    Py_AssertWithArg(PyString_GET_SIZE(key) < index->info.keySize,
+    Py_AssertWithArg((int)PyString_GET_SIZE(key) < beeindex->info.keySize,
 		     PyExc_TypeError,
-		     "keys must not exceed length %i",index->info.keySize - 1);
-    Py_Assert(PyString_GET_SIZE(key) == strlen(PyString_AS_STRING(key)),
+		     "keys must not exceed length %i",beeindex->info.keySize - 1);
+    Py_Assert((unsigned int) PyString_GET_SIZE(key) == strlen(PyString_AS_STRING(key)),
 	      PyExc_TypeError,
 	      "keys may not have embedded null bytes");
 
@@ -327,7 +358,7 @@ void *mxBeeIndex_KeyFromString(mxBeeIndexObject *index,
 }
 
 static
-PyObject *mxBeeIndex_StringFromKey(mxBeeIndexObject *index,
+PyObject *mxBeeIndex_StringFromKey(mxBeeIndexObject *beeindex,
 				   void *key)
 {
     return PyString_FromString((char*)key);
@@ -341,7 +372,7 @@ int mxBeeIndex_CompareStrings(const void *key1, const void *key2)
 
 /* Use Python integer as keys */
 static
-void *mxBeeIndex_KeyFromInteger(mxBeeIndexObject *index,
+void *mxBeeIndex_KeyFromInteger(mxBeeIndexObject *beeindex,
 				PyObject *key)
 {
     Py_Assert(PyInt_Check(key),
@@ -354,7 +385,7 @@ void *mxBeeIndex_KeyFromInteger(mxBeeIndexObject *index,
 }
 
 static
-PyObject *mxBeeIndex_IntegerFromKey(mxBeeIndexObject *index,
+PyObject *mxBeeIndex_IntegerFromKey(mxBeeIndexObject *beeindex,
 				    void *key)
 {
     return PyInt_FromLong(*(long*)key);
@@ -370,7 +401,7 @@ int mxBeeIndex_CompareLongs(const void *key1, const void *key2)
 
 /* Use Python floats as keys */
 static
-void *mxBeeIndex_KeyFromFloat(mxBeeIndexObject *index,
+void *mxBeeIndex_KeyFromFloat(mxBeeIndexObject *beeindex,
 			      PyObject *key)
 {
     Py_Assert(PyFloat_Check(key),
@@ -383,7 +414,7 @@ void *mxBeeIndex_KeyFromFloat(mxBeeIndexObject *index,
 }
 
 static
-PyObject *mxBeeIndex_FloatFromKey(mxBeeIndexObject *index,
+PyObject *mxBeeIndex_FloatFromKey(mxBeeIndexObject *beeindex,
 				  void *key)
 {
     return PyFloat_FromDouble(*(double*)key);
@@ -397,6 +428,55 @@ int mxBeeIndex_CompareDoubles(const void *key1, const void *key2)
     return (a == b) ? CC_EQ : (a > b) ? CC_GT : CC_LT;
 }
 
+/* Python object to record address conversion.
+
+   Returns 0 and raises an exception in case of an error. 
+*/
+static
+bRecAddr mxBeeIndex_RecordAddressFromObject(PyObject *address)
+{
+    unsigned long value;
+    
+    if (!address)
+	goto onError;
+
+    /* Short cut */
+    if (PyInt_Check(address))
+	return (bRecAddr)PyInt_AS_LONG(address);
+
+    /* file.tell() will return longs on platforms which have long file
+       support */
+    if (PyLong_Check(address))
+	value = PyLong_AsUnsignedLong(address);
+    else
+	value = (unsigned long) PyInt_AsLong(address);
+    if (value == (unsigned long) -1 && PyErr_Occurred())
+	goto onError;
+    return (bRecAddr)value;
+    
+ onError:
+    PyErr_SetString(PyExc_TypeError,
+		    "record address must be an integer or long");
+    return 0;
+}
+
+/* Record address to Python object conversion.
+
+   If the value fits into an Python integer object, then an integer is
+   returned. Otherwise, a Python long object is used.
+
+   Returns NULL and raises an exception in case of an error. 
+
+*/
+static
+PyObject *mxBeeIndex_ObjectFromRecordAddress(bRecAddr recaddr)
+{
+    if ((unsigned long)recaddr > INT_MAX)
+	return PyLong_FromUnsignedLong((unsigned long)recaddr);
+    else
+	return PyInt_FromLong((long)recaddr);
+}
+
 /* --- API functions --- */
 
 static
@@ -461,10 +541,9 @@ int mxBeeIndex_SetKey(mxBeeIndexObject *self,
     if (!key)
 	goto onError;
     
-    Py_Assert(recaddr && PyInt_Check(recaddr),
-	      PyExc_TypeError,
-	      "record address must be an integer");
-    record = (bRecAddr)PyInt_AS_LONG(recaddr);
+    record = mxBeeIndex_RecordAddressFromObject(recaddr);
+    if (record == 0 && PyErr_Occurred())
+	goto onError;
 
     /* Either insert or update the key; if dupkeys are allowed, only
        inserts are possible */
@@ -490,7 +569,7 @@ int mxBeeIndex_SetKey(mxBeeIndexObject *self,
     return -1;
 }
 
-/* Clear the index by reopening the file as new file. */
+/* Clear the beeindex by reopening the file as new file. */
 
 static
 int mxBeeIndex_Clear(mxBeeIndexObject *self)
@@ -500,7 +579,7 @@ int mxBeeIndex_Clear(mxBeeIndexObject *self)
 
     Py_Assert(filemode != 1,
 	      PyExc_IOError,
-	      "index is read-only");
+	      "beeindex is read-only");
 
     /* Close the file */
     if (self->handle)
@@ -509,7 +588,7 @@ int mxBeeIndex_Clear(mxBeeIndexObject *self)
     /* Reopen the file as new file */
     self->info.filemode = 2;
 
-    /* Open the index */
+    /* Open the beeindex */
     rc = bOpen(self->info,&(self->handle));
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
@@ -526,7 +605,7 @@ int mxBeeIndex_Clear(mxBeeIndexObject *self)
     /* Restore filemode */
     self->info.filemode = filemode;
 
-    DPRINTF("mxBeeIndex_Clear: instance at %0x\n",(int)self);
+    DPRINTF("mxBeeIndex_Clear: instance at %0lx\n",(long)self);
 
     return 0;
 
@@ -536,7 +615,7 @@ int mxBeeIndex_Clear(mxBeeIndexObject *self)
 
 /* --- Methods --- */
 
-#define index ((mxBeeIndexObject*)self)
+#define beeindex ((mxBeeIndexObject*)self)
 
 Py_C_Function( mxBeeIndex_flush,
 	       "flush()\n\n"
@@ -547,10 +626,10 @@ Py_C_Function( mxBeeIndex_flush,
     
     Py_NoArgsCheck();
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
-    rc = bFlush(index->handle);
+    rc = bFlush(beeindex->handle);
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
 	goto onError;
@@ -570,13 +649,13 @@ Py_C_Function( mxBeeIndex_close,
     
     Py_NoArgsCheck();
 
-    if (index->handle) {
-	rc = bClose(index->handle);
+    if (beeindex->handle) {
+	rc = bClose(beeindex->handle);
 	if (rc != bErrOk) {
 	    mxBeeBase_ReportError(rc);
 	    goto onError;
 	}
-	index->handle = NULL;
+	beeindex->handle = NULL;
     }
     Py_ReturnNone();
     
@@ -591,7 +670,7 @@ Py_C_Function( mxBeeIndex_clear,
 {
     Py_NoArgsCheck();
 
-    if (mxBeeIndex_Clear(index))
+    if (mxBeeIndex_Clear(beeindex))
 	goto onError;
     Py_ReturnNone();
     
@@ -614,15 +693,15 @@ Py_C_Function( mxBeeIndex_get,
 
     Py_Get2Args("O|O",obj,def);
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
     /* Find key */
-    key = index->KeyFromObject(index,obj);
+    key = beeindex->KeyFromObject(beeindex,obj);
     if (!key)
 	goto onError;
-    rc = bFindKey(index->handle,&c,key,&record);
+    rc = bFindKey(beeindex->handle,&c,key,&record);
     if (rc == bErrKeyNotFound) {
 	Py_INCREF(def);
 	return def;
@@ -631,7 +710,7 @@ Py_C_Function( mxBeeIndex_get,
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
-    return PyInt_FromLong((long)record);
+    return mxBeeIndex_ObjectFromRecordAddress(record);
     
  onError:
     return NULL;
@@ -650,15 +729,15 @@ Py_C_Function( mxBeeIndex_has_key,
 
     Py_GetArg("O",obj);
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
     /* Find key */
-    key = index->KeyFromObject(index,obj);
+    key = beeindex->KeyFromObject(beeindex,obj);
     if (!key)
 	goto onError;
-    rc = bFindKey(index->handle,&c,key,&record);
+    rc = bFindKey(beeindex->handle,&c,key,&record);
     if (rc == bErrKeyNotFound) {
 	Py_INCREF(Py_False);
 	return Py_False;
@@ -690,21 +769,21 @@ Py_C_Function( mxBeeIndex_cursor,
 
     Py_Get2Args("O|O",obj,def);
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
     /* Find key */
     if (obj == mxBeeIndex_FirstKey)
-	rc = bFindFirstKey(index->handle,&c,NULL,NULL);
+	rc = bFindFirstKey(beeindex->handle,&c,NULL,NULL);
     else if (obj == mxBeeIndex_LastKey)
-	rc = bFindLastKey(index->handle,&c,NULL,NULL);
+	rc = bFindLastKey(beeindex->handle,&c,NULL,NULL);
     else {
 	void *key;
-	key = index->KeyFromObject(index,obj);
+	key = beeindex->KeyFromObject(beeindex,obj);
 	if (!key)
 	    goto onError;
-	rc = bFindKey(index->handle,&c,key,NULL);
+	rc = bFindKey(beeindex->handle,&c,key,NULL);
     }
     if (rc == bErrKeyNotFound && def) {
 	Py_INCREF(def);
@@ -716,7 +795,7 @@ Py_C_Function( mxBeeIndex_cursor,
     }
 
     /* Create cursor */
-    v = (PyObject*)mxBeeCursor_New(index,&c);
+    v = (PyObject*)mxBeeCursor_New(beeindex,&c);
     if (!v)
 	goto onError;
 
@@ -739,7 +818,7 @@ Py_C_Function( mxBeeIndex_keys,
 
     Py_NoArgsCheck();
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
@@ -748,7 +827,7 @@ Py_C_Function( mxBeeIndex_keys,
 	goto onError;
     
     /* Find first */
-    rc = bFindFirstKey(index->handle,&c,NULL,NULL);
+    rc = bFindFirstKey(beeindex->handle,&c,NULL,NULL);
     if (rc == bErrKeyNotFound)
 	return v;
     if (rc != bErrOk) {
@@ -757,13 +836,13 @@ Py_C_Function( mxBeeIndex_keys,
     }
 
     while (1) {
-	w = index->ObjectFromKey(index,c.key);
+	w = beeindex->ObjectFromKey(beeindex,c.key);
 	if (!w)
 	    goto onError;
 	PyList_Append(v,w);
 	Py_DECREF(w);
 
-	rc = bFindNextKey(index->handle,&c,NULL,NULL);
+	rc = bFindNextKey(beeindex->handle,&c,NULL,NULL);
 	if (rc == bErrKeyNotFound)
 	    break;
 	if (rc != bErrOk) {
@@ -793,7 +872,7 @@ Py_C_Function( mxBeeIndex_values,
 
     Py_NoArgsCheck();
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
@@ -802,7 +881,7 @@ Py_C_Function( mxBeeIndex_values,
 	goto onError;
 
     /* Find first */
-    rc = bFindFirstKey(index->handle,&c,NULL,&rec);
+    rc = bFindFirstKey(beeindex->handle,&c,NULL,&rec);
     if (rc == bErrKeyNotFound)
 	return v;
     if (rc != bErrOk) {
@@ -811,13 +890,13 @@ Py_C_Function( mxBeeIndex_values,
     }
 
     while (1) {
-	w = PyInt_FromLong((long)rec);
+	w = mxBeeIndex_ObjectFromRecordAddress(rec);
 	if (!w)
 	    goto onError;
 	PyList_Append(v,w);
 	Py_DECREF(w);
 
-	rc = bFindNextKey(index->handle,&c,NULL,&rec);
+	rc = bFindNextKey(beeindex->handle,&c,NULL,&rec);
 	if (rc == bErrKeyNotFound)
 	    break;
 	if (rc != bErrOk) {
@@ -846,7 +925,7 @@ Py_C_Function( mxBeeIndex_items,
 
     Py_NoArgsCheck();
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
@@ -855,7 +934,7 @@ Py_C_Function( mxBeeIndex_items,
 	goto onError;
 
     /* Find first */
-    rc = bFindFirstKey(index->handle,&c,NULL,&rec);
+    rc = bFindFirstKey(beeindex->handle,&c,NULL,&rec);
     if (rc == bErrKeyNotFound)
 	return v;
     if (rc != bErrOk) {
@@ -866,10 +945,10 @@ Py_C_Function( mxBeeIndex_items,
     while (1) {
 	PyObject *key,*value,*t;
 	
-	key = index->ObjectFromKey(index,c.key);
+	key = beeindex->ObjectFromKey(beeindex,c.key);
 	if (!key)
 	    goto onError;
-	value = PyInt_FromLong((long)rec);
+	value = mxBeeIndex_ObjectFromRecordAddress(rec);
 	if (!value) {
 	    Py_DECREF(key);
 	    goto onError;
@@ -885,7 +964,7 @@ Py_C_Function( mxBeeIndex_items,
 	PyList_Append(v,t);
 	Py_DECREF(t);
 
-	rc = bFindNextKey(index->handle,&c,NULL,&rec);
+	rc = bFindNextKey(beeindex->handle,&c,NULL,&rec);
 	if (rc == bErrKeyNotFound)
 	    break;
 	if (rc != bErrOk) {
@@ -908,36 +987,38 @@ Py_C_Function( mxBeeIndex_delete,
 	       )
 {
     PyObject *obj;
-    int recaddr = -1;
+    PyObject *recaddr = NULL;
     bError rc = bErrOk;
     bRecAddr record;
     void *key = NULL;
     
-    Py_Get2Args("O|i",obj,recaddr);
+    Py_Get2Args("O|O",obj,recaddr);
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
-    if (index->info.dupKeys)
-	Py_Assert(recaddr > 0,
+    if (beeindex->info.dupKeys)
+	Py_Assert(recaddr != NULL,
 		  PyExc_ValueError,
 		  "record address must be given if dupkeys are allowed");
 
-    key = index->KeyFromObject(index,obj);
+    key = beeindex->KeyFromObject(beeindex,obj);
     if (!key)
 	goto onError;
-    record = (bRecAddr)recaddr;
+    record = mxBeeIndex_RecordAddressFromObject(recaddr);
+    if (record == 0 && PyErr_Occurred())
+	goto onError;
 
     /* Delete key using record address if dupkeys is enabled */
-    rc = bDeleteKey(index->handle,key,&record);
+    rc = bDeleteKey(beeindex->handle,key,&record);
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
 
     /* Increment update count */
-    index->updates++;
+    beeindex->updates++;
 
     Py_ReturnNone();
 
@@ -953,43 +1034,52 @@ Py_C_Function( mxBeeIndex_update,
 {
     PyObject *obj;
     bError rc = bErrOk;
-    int value = 0, oldvalue = -1;
+    PyObject *value;
+    PyObject *oldvalue = NULL;
     bRecAddr record, oldrecord;
     void *key = NULL;
     
-    Py_Get3Args("Oi|i",obj,value,oldvalue);
+    Py_Get3Args("OO|O",obj,value,oldvalue);
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
-    if (index->info.dupKeys)
-	Py_Assert(oldvalue > 0,
+    if (beeindex->info.dupKeys)
+	Py_Assert(oldvalue != NULL,
 		  PyExc_ValueError,
 		  "oldvalue must be given if dupkeys are allowed");
 
-    record = (bRecAddr)value;
-    oldrecord = (bRecAddr)oldvalue;
-    key = index->KeyFromObject(index,obj);
+    record = mxBeeIndex_RecordAddressFromObject(value);
+    if (record == 0 && PyErr_Occurred())
+	goto onError;
+    if (oldvalue) {
+	oldrecord = mxBeeIndex_RecordAddressFromObject(oldvalue);
+	if (record == 0 && PyErr_Occurred())
+	    goto onError;
+    }
+    else
+	oldrecord = 0;
+    key = beeindex->KeyFromObject(beeindex,obj);
     if (!key)
 	goto onError;
 
-    /* Delete key using record address if dupkeys is enabled */
-    rc = bDeleteKey(index->handle,key,&oldrecord);
+    /* Delete key using oldrecord address if dupkeys is enabled */
+    rc = bDeleteKey(beeindex->handle,key,&oldrecord);
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
 
     /* Insert the new key,value pair */
-    rc = bInsertKey(index->handle,key,record);
+    rc = bInsertKey(beeindex->handle,key,record);
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
 
     /* Increment update count */
-    index->updates++;
+    beeindex->updates++;
 
     Py_ReturnNone();
 
@@ -1005,24 +1095,26 @@ Py_C_Function( mxBeeIndex_validate,
 {
     Py_NoArgsCheck();
 
-    Py_Assert(index->handle != NULL,
+    Py_Assert(beeindex->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
-    return PyInt_FromLong(bValidateTree(index->handle) == 0);
+    return PyInt_FromLong(bValidateTree(beeindex->handle) == 0);
     
  onError:
     return NULL;
 }
 
-#undef index
+#undef beeindex
 
 /* --- slots --- */
 
 static
-PyObject *mxBeeIndex_Getattr(mxBeeIndexObject *self,
+PyObject *mxBeeIndex_Getattr(PyObject *obj,
 			     char *name)
 {
+    mxBeeIndexObject *self = (mxBeeIndexObject *)obj;
+    
     if (Py_WantAttr(name,"closed"))
 	return PyInt_FromLong((self->handle == NULL));
 
@@ -1057,9 +1149,10 @@ PyObject *mxBeeIndex_Getattr(mxBeeIndexObject *self,
 }
 
 static
-int mxBeeIndex_Length(mxBeeIndexObject *self)
+Py_ssize_t mxBeeIndex_Length(PyObject *obj)
 {
     /* XXX Much too slow... */
+    mxBeeIndexObject *self = (mxBeeIndexObject *)obj;
     bError rc;
     bCursor c;
     int i;
@@ -1103,38 +1196,41 @@ int mxBeeIndex_Length(mxBeeIndexObject *self)
 }
 
 static
-PyObject *mxBeeIndex_Subscript(mxBeeIndexObject *self,
-			       PyObject *obj)
+PyObject *mxBeeIndex_Subscript(PyObject *obj,
+			       PyObject *key)
 {
+    mxBeeIndexObject *self = (mxBeeIndexObject *)obj;
     bRecAddr record;
 
     Py_Assert(self->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
-    record = mxBeeIndex_FindKey(self,obj);
+    record = mxBeeIndex_FindKey(self, key);
     if (record == -1 && PyErr_Occurred())
 	goto onError;
     
-    return PyInt_FromLong((long)record);
+    return mxBeeIndex_ObjectFromRecordAddress(record);
 
  onError:
     return NULL;
 }
 
 static
-int mxBeeIndex_AssignSubscript(mxBeeIndexObject *self,
-			       PyObject *obj,
+int mxBeeIndex_AssignSubscript(PyObject *obj,
+			       PyObject *key,
 			       PyObject *recaddr)
 {
+    mxBeeIndexObject *self = (mxBeeIndexObject *)obj;
+
     Py_Assert(self->handle != NULL,
 	      mxBeeIndex_Error,
 	      "index is closed");
 
     if (recaddr)
-	return mxBeeIndex_SetKey(self,obj,recaddr);
+	return mxBeeIndex_SetKey(self, key, recaddr);
     else
-	return mxBeeIndex_DeleteKey(self,obj);
+	return mxBeeIndex_DeleteKey(self, key);
 
  onError:
     return -1;
@@ -1143,9 +1239,9 @@ int mxBeeIndex_AssignSubscript(mxBeeIndexObject *self,
 /* Python Type Tables */
 
 static PyMappingMethods mxBeeIndex_TypeAsMapping = {
-	(inquiry)mxBeeIndex_Length, /*mp_length*/
-	(binaryfunc)mxBeeIndex_Subscript, /*mp_subscript*/
-	(objobjargproc)mxBeeIndex_AssignSubscript, /*mp_ass_subscript*/
+	mxBeeIndex_Length, 		/*mp_length*/
+	mxBeeIndex_Subscript, 		/*mp_subscript*/
+	mxBeeIndex_AssignSubscript, 	/*mp_ass_subscript*/
 };
 
 statichere
@@ -1156,23 +1252,23 @@ PyTypeObject mxBeeIndex_Type = {
     sizeof(mxBeeIndexObject),      	/*tp_basicsize*/
     0,			  		/*tp_itemsize*/
     /* slots */
-    (destructor)mxBeeIndex_Free,		/*tp_dealloc*/
-    (printfunc)0,		  	/*tp_print*/
-    (getattrfunc)mxBeeIndex_Getattr,  	/*tp_getattr*/
-    (setattrfunc)0,		  	/*tp_setattr*/
-    (cmpfunc)0,		 		/*tp_compare*/
-    (reprfunc)0,		  	/*tp_repr*/
+    (destructor)mxBeeIndex_Free,	/*tp_dealloc*/
+    0,		  			/*tp_print*/
+    mxBeeIndex_Getattr, 	 	/*tp_getattr*/
+    0,		  			/*tp_setattr*/
+    0,		 			/*tp_compare*/
+    0,		  			/*tp_repr*/
     0,			 		/*tp_as_number*/
     0,					/*tp_as_sequence*/
     &mxBeeIndex_TypeAsMapping,		/*tp_as_mapping*/
-    (hashfunc)0,			/*tp_hash*/
-    (ternaryfunc)0,			/*tp_call*/
-    (reprfunc)0,			/*tp_str*/
-    (getattrofunc)0,			/*tp_getattro*/
-    (setattrofunc)0,			/*tp_setattro*/
+    0,					/*tp_hash*/
+    0,					/*tp_call*/
+    0,					/*tp_str*/
+    0,					/*tp_getattro*/
+    0,					/*tp_setattro*/
     0,					/*tp_as_buffer*/
     0,					/*tp_xxx4*/
-    (char*) 0				/*tp_doc*/
+    0					/*tp_doc*/
 };
 
 /* Python Method Table */
@@ -1198,13 +1294,13 @@ PyMethodDef mxBeeIndex_Methods[] =
 /* --- BeeBase Cursor Object ------------------------------------------- */
 
 statichere
-mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *index,   /* Index object */
+mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *beeindex,   /* Index object */
 				   bCursor *c                 /* bCursor */
 				   )
 {
     mxBeeCursorObject *cursor = 0;
 
-    Py_Assert(index->handle,
+    Py_Assert(beeindex->handle,
 	      mxBeeCursor_Error,
 	      "creating cursor for closed index");
 
@@ -1225,17 +1321,18 @@ mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *index,   /* Index object */
     }
 
     /* Init vars */
-    Py_INCREF(index);
-    cursor->index = index;
+    Py_INCREF(beeindex);
+    cursor->beeindex = beeindex;
     memcpy(&cursor->c,c,sizeof(bCursor));
     cursor->adr = c->buffer->adr;
-    cursor->updates = index->updates;
+    cursor->updates = beeindex->updates;
 
-    DPRINTF("mxBeeCursor_New: instance at %0x\n",(int)cursor);
+    DPRINTF("mxBeeCursor_New: instance at %0lx\n",(long)cursor);
 
     return cursor;
 
  onError:
+    Py_XDECREF(cursor);
     return NULL;
 }
 
@@ -1244,17 +1341,17 @@ mxBeeCursorObject *mxBeeCursor_New(mxBeeIndexObject *index,   /* Index object */
 static
 void mxBeeCursor_Free(mxBeeCursorObject *cursor)
 {
-    DPRINTF("mxBeeCursor_Free: instance at %0x\n",(int)cursor);
+    DPRINTF("mxBeeCursor_Free: instance at %0lx\n",(long)cursor);
 
-    /* Dereference index object */
-    Py_DECREF(cursor->index);
+    /* Dereference beeindex object */
+    Py_DECREF(cursor->beeindex);
     
 #ifdef MXBEEINDEX_FREELIST
     /* Append to free list */
     *(mxBeeCursorObject **)cursor = mxBeeCursor_FreeList;
     mxBeeCursor_FreeList = cursor;
 #else
-    PyMem_DEL(cursor);
+    PyObject_Del(cursor);
 #endif
 }
 
@@ -1263,10 +1360,10 @@ void mxBeeCursor_Free(mxBeeCursorObject *cursor)
 static
 int mxBeeCursor_Invalid(mxBeeCursorObject *self)
 {
-    Py_Assert(self->index->handle != NULL,
+    Py_Assert(self->beeindex->handle != NULL,
 	      mxBeeCursor_Error,
 	      "index is closed - cursor is invalid");
-    Py_Assert(self->index->updates == self->updates,
+    Py_Assert(self->beeindex->updates == self->updates,
 	      mxBeeCursor_Error,
 	      "index was changed - cursor is invalid");
     Py_Assert(self->c.buffer && self->c.buffer->valid,
@@ -1281,7 +1378,7 @@ int mxBeeCursor_Invalid(mxBeeCursorObject *self)
     return -1;
 }
 
-/* Move the cursor to the next index key. If there are no further
+/* Move the cursor to the next beeindex key. If there are no further
    keys, return 0 and leave the cursor where it is. Otherwise return
    1. Returns -1 on error. */
 
@@ -1294,7 +1391,7 @@ int mxBeeCursor_NextKey(mxBeeCursorObject *self)
 	goto onError;
 
     /* Find key (updates cursor only on success) */
-    rc = bFindNextKey(self->index->handle,&self->c,NULL,NULL);
+    rc = bFindNextKey(self->beeindex->handle,&self->c,NULL,NULL);
     if (rc == bErrKeyNotFound)
 	return 0;
     if (rc != bErrOk) {
@@ -1311,7 +1408,7 @@ int mxBeeCursor_NextKey(mxBeeCursorObject *self)
     return -1;
 }
 
-/* Move the cursor to the prev index key. If there are no further
+/* Move the cursor to the prev beeindex key. If there are no further
    keys, return 0 and leave the cursor where it is. Otherwise return
    1. Returns -1 on error. */
 
@@ -1324,7 +1421,7 @@ int mxBeeCursor_PrevKey(mxBeeCursorObject *self)
 	goto onError;
 
     /* Find key (updates cursor only on success) */
-    rc = bFindPrevKey(self->index->handle,&self->c,NULL,NULL);
+    rc = bFindPrevKey(self->beeindex->handle,&self->c,NULL,NULL);
     if (rc == bErrKeyNotFound)
 	return 0;
     if (rc != bErrOk) {
@@ -1350,7 +1447,7 @@ PyObject *mxBeeCursor_GetKey(mxBeeCursorObject *self)
 	goto onError;
 
     /* Convert key to object */
-    v = self->index->ObjectFromKey(self->index,self->c.key);
+    v = self->beeindex->ObjectFromKey(self->beeindex,self->c.key);
     if (!v)
 	goto onError;
 
@@ -1371,13 +1468,13 @@ PyObject *mxBeeCursor_GetValue(mxBeeCursorObject *self)
 	goto onError;
 
     /* Check that cursor is valid and read record address */
-    rc = bCursorReadData(self->index->handle,&self->c,NULL,&rec);
+    rc = bCursorReadData(self->beeindex->handle,&self->c,NULL,&rec);
     if (rc != bErrOk) {
 	mxBeeBase_ReportError(rc);
 	goto onError;
     }
 
-    v = PyInt_FromLong((long)rec);
+    v = mxBeeIndex_ObjectFromRecordAddress(rec);
     if (!v)
 	goto onError;
 
@@ -1454,7 +1551,7 @@ Py_C_Function( mxBeeCursor_copy,
     if (mxBeeCursor_Invalid(cursor))
 	goto onError;
 
-    return (PyObject *)mxBeeCursor_New(cursor->index,&cursor->c);
+    return (PyObject *)mxBeeCursor_New(cursor->beeindex,&cursor->c);
 
  onError:
     return NULL;
@@ -1465,11 +1562,13 @@ Py_C_Function( mxBeeCursor_copy,
 /* --- slots --- */
 
 static
-PyObject *mxBeeCursor_Getattr(mxBeeCursorObject *self,
-			     char *name)
+PyObject *mxBeeCursor_Getattr(PyObject *obj,
+			      char *name)
 {
+    mxBeeCursorObject *self = (mxBeeCursorObject *)obj;
+
     if (Py_WantAttr(name,"closed"))
-	return PyInt_FromLong((self->index->handle == NULL));
+	return PyInt_FromLong((self->beeindex->handle == NULL));
 
     if (Py_WantAttr(name,"key"))
 	return mxBeeCursor_GetKey(self);
@@ -1502,27 +1601,27 @@ statichere
 PyTypeObject mxBeeCursor_Type = {
     PyObject_HEAD_INIT(0)		/* init at startup ! */
     0,			  		/*ob_size*/
-    "BeeCursor",	  			/*tp_name*/
+    "BeeCursor",	  		/*tp_name*/
     sizeof(mxBeeCursorObject),      	/*tp_basicsize*/
     0,			  		/*tp_itemsize*/
     /* slots */
-    (destructor)mxBeeCursor_Free,		/*tp_dealloc*/
-    (printfunc)0,		  	/*tp_print*/
-    (getattrfunc)mxBeeCursor_Getattr,  	/*tp_getattr*/
-    (setattrfunc)0,		  	/*tp_setattr*/
-    (cmpfunc)0,		 		/*tp_compare*/
-    (reprfunc)0,		  	/*tp_repr*/
+    (destructor)mxBeeCursor_Free,	/*tp_dealloc*/
+    0,		  			/*tp_print*/
+    mxBeeCursor_Getattr,  		/*tp_getattr*/
+    0,		  			/*tp_setattr*/
+    0,		 	       		/*tp_compare*/
+    0,				  	/*tp_repr*/
     0,			 		/*tp_as_number*/
     0,					/*tp_as_sequence*/
     0,					/*tp_as_mapping*/
-    (hashfunc)0,			/*tp_hash*/
-    (ternaryfunc)0,			/*tp_call*/
-    (reprfunc)0,			/*tp_str*/
-    (getattrofunc)0,			/*tp_getattro*/
-    (setattrofunc)0,			/*tp_setattro*/
+    0,					/*tp_hash*/
+    0,					/*tp_call*/
+    0,					/*tp_str*/
+    0,					/*tp_getattro*/
+    0,					/*tp_setattro*/
     0,					/*tp_as_buffer*/
     0,					/*tp_xxx4*/
-    (char*) 0				/*tp_doc*/
+    0					/*tp_doc*/
 };
 
 /* Python Method Table */
@@ -1552,8 +1651,10 @@ Py_C_Function_WithKeywords(
     Py_KeywordsGet5Args("si|iii",
 			filename,keysize,dupkeys,filemode,sectorsize);
 
-    return (PyObject *)mxBeeIndex_New(filename,filemode,
-				      keysize+1,sectorsize,
+    return (PyObject *)mxBeeIndex_New(filename,
+				      filemode,
+				      keysize + 1,
+				      sectorsize,
 				      mxBeeIndex_CompareStrings,
 				      mxBeeIndex_StringFromKey,
 				      mxBeeIndex_KeyFromString,
@@ -1576,8 +1677,10 @@ Py_C_Function_WithKeywords(
     Py_KeywordsGet4Args("s|iii",
 			filename,dupkeys,filemode,sectorsize);
 
-    return (PyObject *)mxBeeIndex_New(filename,filemode,
-				      keysize,sectorsize,
+    return (PyObject *)mxBeeIndex_New(filename,
+				      filemode,
+				      keysize,
+				      sectorsize,
 				      mxBeeIndex_CompareLongs,
 				      mxBeeIndex_IntegerFromKey,
 				      mxBeeIndex_KeyFromInteger,
@@ -1600,8 +1703,10 @@ Py_C_Function_WithKeywords(
     Py_KeywordsGet4Args("s|iii",
 			filename,dupkeys,filemode,sectorsize);
 
-    return (PyObject *)mxBeeIndex_New(filename,filemode,
-				      keysize,sectorsize,
+    return (PyObject *)mxBeeIndex_New(filename,
+				      filemode,
+				      keysize,
+				      sectorsize,
 				      mxBeeIndex_CompareDoubles,
 				      mxBeeIndex_FloatFromKey,
 				      mxBeeIndex_KeyFromFloat,
@@ -1631,8 +1736,9 @@ void mxBeeBaseModule_Cleanup(void)
 	while (d != NULL) {
 	    mxBeeIndexObject *v = d;
 	    d = *(mxBeeIndexObject **)d;
-	    free(v);
+	    PyObject_Del(v);
 	}
+	mxBeeIndex_FreeList = NULL;
     }
 #endif
 #ifdef MXBEECURSOR_FREELIST
@@ -1641,10 +1747,13 @@ void mxBeeBaseModule_Cleanup(void)
 	while (d != NULL) {
 	    mxBeeCursorObject *v = d;
 	    d = *(mxBeeCursorObject **)d;
-	    free(v);
+	    PyObject_Del(v);
 	}
+	mxBeeCursor_FreeList = NULL;
     }
 #endif
+    /* Reset mxBeeBase_Initialized flag */
+    mxBeeBase_Initialized = 0;
 }
 
 /* Create PyMethodObjects and register them in the module's dict */
@@ -1653,6 +1762,10 @@ MX_EXPORT(void)
 {
     PyObject *module, *moddict;
 
+    if (mxBeeBase_Initialized)
+	Py_Error(PyExc_SystemError,
+		 "can't initialize "MXBEEBASE_MODULE" more than once");
+
     /* Init type objects */
     PyType_Init(mxBeeIndex_Type);
     PyType_Init(mxBeeCursor_Type);
@@ -1666,6 +1779,14 @@ MX_EXPORT(void)
     if (module == NULL)
 	goto onError;
 
+    /* Init globals */
+#ifdef MXBEEINDEX_FREELIST
+    mxBeeIndex_FreeList = NULL;
+#endif
+#ifdef MXBEECURSOR_FREELIST
+    mxBeeCursor_FreeList = NULL;
+#endif
+
     /* Register cleanup function */
     if (Py_AtExit(mxBeeBaseModule_Cleanup))
 	/* XXX what to do if we can't register that function ??? */;
@@ -1674,7 +1795,7 @@ MX_EXPORT(void)
     moddict = PyModule_GetDict(module);
     if (moddict == NULL)
 	goto onError;
-    insobj(moddict,"__version__",PyString_FromString(VERSION));
+    insobj(moddict,"__version__",PyString_FromString(MXBEEBASE_VERSION));
 
     /* Errors */
     if (!(mxBeeIndex_Error = insexc(moddict,"BeeIndexError")))
@@ -1698,6 +1819,9 @@ MX_EXPORT(void)
     PyDict_SetItemString(moddict,"BeeCursorType",
 			 (PyObject *)&mxBeeCursor_Type);
 
+    /* We are now initialized */
+    mxBeeBase_Initialized = 1;
+
  onError:
     /* Check for errors and report them */
     if (PyErr_Occurred())
