@@ -1,7 +1,7 @@
 """ Python part of the low-level DateTime[Delta] type implementation.
 
     Copyright (c) 1998-2001, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 """
@@ -27,14 +27,32 @@ def _DTD(seconds):
 # Module init
 class modinit:
 
-    global _time,_string,_math
-    import time,string,math
+    global _time,_string,_math,_types
+    import time,string,math,types
     _time = time
     _string = string
     _math = math
+    _types = types
 
 del modinit
 
+### Helpers
+
+def _isstring(arg,
+
+              isinstance=isinstance, types=_types):
+    
+    if isinstance(arg, types.StringType):
+        return 1
+    try:
+        if isinstance(arg, types.UnicodeType):
+            return 1
+    except AttributeError:
+        pass
+    return 0
+
+### Compatibility APIs
+
 # Aliases and functions to make 'from mx.DateTime import *' work much
 # like 'from time import *'
 
@@ -202,43 +220,46 @@ def DateTimeFromMJD(mjd):
     """
     return DateTimeFromAbsDays(mjd + 678575.0)
 
-def DateTimeFrom(*args,**kws):
+def DateTimeFrom(*args, **kws):
 
-    """ DateTimeFrom(*args,**kws)
+    """ DateTimeFrom(*args, **kws)
 
         Generic DateTime instance constructor. Can handle parsing
         strings, numbers and keywords.
 
+        XXX Add support for Unicode.
+
     """
-    import types
     if len(args) == 1:
         # Single argument
         arg = args[0]
-        if type(arg) is types.StringType:
+        argtype = type(arg)
+        if _isstring(arg):
             import Parser
-            return Parser.DateTimeFromString(arg)
-        elif type(arg) is DateTimeType:
+            return apply(Parser.DateTimeFromString, args, kws)
+        elif argtype is DateTimeType:
             return arg
-        elif type(arg) is DateTimeDeltaType:
+        elif argtype is DateTimeDeltaType:
             raise TypeError,'cannot convert DateTimeDelta to DateTime'
         else:
             try:
                 value = float(arg)
-            except TypeError:
+            except (TypeError, ValueError):
                 value = int(arg)
+            assert not kws
             return DateTimeFromTicks(value)
 
     elif len(args) > 1:
         # More than one argument
-        if len(args) == 2 and \
-           type(args[0]) is types.StringType and \
-           type(args[1]) is types.StringType:
+        if len(args) == 2 and _isstring(args[0]) and _isstring(args[1]):
             # interpret as date and time string
             import Parser
-            return Parser.DateTimeFromString(args[0] + ' ' + args[1])
+            return apply(Parser.DateTimeFromString,
+                         (args[0] + ' ' + args[1],),
+                         kws)
 
         # Assume the arguments are the same as for DateTime()
-        return apply(DateTime,args)
+        return apply(DateTime, args, kws)
 
     elif len(kws) > 0:
         # Keyword arguments; add defaults... today at 0:00:00
@@ -254,21 +275,22 @@ def DateTimeFrom(*args,**kws):
     else:
         raise TypeError,'cannot convert arguments to DateTime'
 
-def DateTimeDeltaFrom(*args,**kws):
+def DateTimeDeltaFrom(*args, **kws):
 
-    """ DateTimeDeltaFrom(*args,**kws)
+    """ DateTimeDeltaFrom(*args, **kws)
 
         Generic DateTimeDelta instance constructor. Can handle parsing
         strings, numbers and keywords.
 
+        XXX Add support for Unicode.
+
     """
-    import types
     if len(args) == 1:
         # Single argument
         arg = args[0]
-        if type(arg) is types.StringType:
+        if _isstring(arg):
             import Parser
-            return Parser.DateTimeDeltaFromString(arg)
+            return apply(Parser.DateTimeDeltaFromString, args, kws)
         elif type(arg) is DateTimeDeltaType:
             return arg
         elif type(arg) is DateTimeType:
@@ -278,11 +300,12 @@ def DateTimeDeltaFrom(*args,**kws):
                 value = float(arg)
             except TypeError:
                 value = int(arg)
+            assert not kws
             return DateTimeDeltaFromSeconds(value)
 
     elif len(args) > 1:
         # Assume the arguments are the same as for DateTimeDelta()
-        return apply(DateTimeDelta,args)
+        return apply(DateTimeDelta, args, kws)
 
     elif len(kws) > 0:
         # Keyword arguments; default: 00:00:00:00.00
@@ -295,22 +318,24 @@ def DateTimeDeltaFrom(*args,**kws):
     else:
         raise TypeError,'cannot convert arguments to DateTimeDelta'
 
-def TimeDeltaFrom(*args,**kws):
+def TimeDeltaFrom(*args, **kws):
 
-    """ TimeDeltaFrom(*args,**kws)
+    """ TimeDeltaFrom(*args, **kws)
 
         Generic TimeDelta instance constructor. Can handle parsing
         strings, numbers and keywords.
 
+        XXX Add support for Unicode.
+
     """
     if len(args) > 1:
         # Assume the arguments are the same as for TimeDelta(): without
         # days part !
-        return apply(DateTimeDelta,(0,)+args)
+        return apply(DateTimeDelta, (0,)+args, kws)
     else:
         # Otherwise treat the arguments just like for DateTimeDelta
         # instances.
-        return apply(DateTimeDeltaFrom,args,kws)
+        return apply(DateTimeDeltaFrom, args, kws)
 
 def DateFromTicks(ticks,
                   # Locals:
@@ -323,7 +348,7 @@ def DateFromTicks(ticks,
         The time part is ignored.
 
     """
-    return apply(DateTime,localtime(ticks)[:3])
+    return apply(DateTime, localtime(ticks)[:3])
 
 def TimestampFromTicks(ticks,
                        # Locals:
@@ -335,7 +360,7 @@ def TimestampFromTicks(ticks,
         time indicated by the given ticks value.
 
     """
-    return apply(DateTime,localtime(ticks)[:6])
+    return apply(DateTime, localtime(ticks)[:6])
 
 def TimeFromTicks(ticks,
                   # Locals:
@@ -347,7 +372,7 @@ def TimeFromTicks(ticks,
         indicated by the given ticks value. The date part is ignored.
 
     """
-    return apply(DateTimeDelta,localtime(ticks)[3:6])
+    return apply(DateTimeDelta, (0,) + localtime(ticks)[3:6])
 
 # Aliases
 utctime = gmtime
@@ -357,7 +382,7 @@ DateTimeFromTicks = localtime
 Date = DateTime
 Time = TimeDelta
 Timestamp = DateTime
-DateFrom = DateTimeFrom
+DateFrom = DateTimeFrom # XXX should only parse the date part !
 TimeFrom = TimeDeltaFrom
 TimestampFrom = DateTimeFrom
 GregorianDateTime = DateTime
@@ -437,12 +462,10 @@ Month = {2: 'February', 3: 'March', None: 0, 'July': 7, 11: 'November',
     'January', 0: None}
 
 # Limits (see also the range checks in mxDateTime.c)
-import sys
-MaxDateTime = DateTime(sys.maxint / 366 - 1,12,31)
-MinDateTime = DateTime(-sys.maxint / 367,1,1)
-MaxDateTimeDelta = DateTimeDeltaFromSeconds(sys.maxint * 86400.0)
+MaxDateTime = DateTime(5867440,12,31) 
+MinDateTime = DateTime(-5851455,1,1)
+MaxDateTimeDelta = DateTimeDeltaFromSeconds(2147483647 * 86400.0)
 MinDateTimeDelta = -MaxDateTimeDelta
-del sys
 
 ###
 
@@ -483,6 +506,13 @@ class RelativeDateTime:
     second = None
     weekday = None
 
+    # cached hash value
+    _hash = None
+
+    # For Zope security:
+    __roles__ = None
+    __allow_access_to_unprotected_subobjects__ = 1
+
     def __init__(self,
                  years=0,months=0,days=0,
                  hours=0,minutes=0,seconds=0,
@@ -502,9 +532,9 @@ class RelativeDateTime:
         self.hour = hour
         self.minute = minute
         self.second = second
-        if weekday:
+        if weekday is not None:
             #  Make sure we've got a 2-tuple
-            len(weekday) == 2
+            assert len(weekday) == 2
             self.weekday = weekday
 
     def __add__(self,other,
@@ -539,7 +569,7 @@ class RelativeDateTime:
 
     def __radd__(self,other,
                  # Locals:
-                 None=None,isinstance=isinstance,DateTimeType=DateTimeType,
+                 isinstance=isinstance,DateTimeType=DateTimeType,
                  DateTime=DateTime,DateTimeDelta=DateTimeDelta):
 
         if isinstance(other,DateTimeType):
@@ -580,8 +610,14 @@ class RelativeDateTime:
             # Refit into proper ranges:
             if month < 1 or month > 12:
                 month = month - 1
-                year = year + month / 12
-                month = month % 12 + 1
+                yeardelta, monthdelta = divmod(month, 12)
+                year = year + yeardelta
+                month = monthdelta + 1
+
+            # Make sure we have integers
+            year = int(year)
+            month = int(month)
+            day = int(day)
 
             if self.weekday is None:
                 return DateTime(year, month, 1) + \
@@ -592,8 +628,10 @@ class RelativeDateTime:
             d = DateTime(year, month, 1) + \
                 DateTimeDelta(day-1,hour,minute,second)
             if index == 0:
+                # 0 index: next weekday if no match
                 return d + (day_of_week - d.day_of_week)
             elif index > 0:
+                # positive index (1 == first weekday of month)
                 first = d - (d.day - 1)
                 diff = day_of_week - first.day_of_week
                 if diff >= 0:
@@ -601,6 +639,7 @@ class RelativeDateTime:
                 else:
                     return first + (diff + index * 7)
             else:
+                # negative index (-1 == last weekday of month)
                 last = d + (d.days_in_month - d.day)
                 diff = day_of_week - last.day_of_week
                 if diff <= 0:
@@ -642,7 +681,7 @@ class RelativeDateTime:
 
     def __rsub__(self,other,
                  # Locals:
-                 None=None,isinstance=isinstance,DateTime=DateTime):
+                 isinstance=isinstance,DateTimeType=DateTimeType):
 
         if isinstance(other,DateTimeType):
             # DateTime (other) - RelativeDateTime (self)
@@ -720,12 +759,54 @@ class RelativeDateTime:
         # RelativeDateTime (self) / Number (other)
         return self.__mul__(1/float(other))
 
+    def __eq__(self, other):
+
+        if isinstance(self, RelativeDateTime) and \
+           isinstance(other, RelativeDateTime):
+            # RelativeDateTime (self) == RelativeDateTime (other)
+            if (self.years == other.years and
+                self.months == other.months and
+                self.days == other.days and
+                self.year == other.year and
+                self.day == other.day and
+                self.hours == other.hours and
+                self.minutes == other.minutes and
+                self.seconds == other.seconds and
+                self.hour == other.hour and
+                self.minute == other.minute and
+                self.second == other.second and
+                self.weekday == other.weekday):
+                return 1
+            else:
+                return 0
+        else:
+            raise TypeError,"can't compare the two types"
+
+    def __hash__(self):
+
+        if self._hash is not None:
+            return self._hash
+        x = 1234
+        for value in (self.years, self.months, self.days,
+                      self.year, self.day,
+                      self.hours, self.minutes, self.seconds,
+                      self.hour, self.minute, self.second,
+                      self.weekday):
+            if value is None:
+                x = 135051820 ^ x
+            else:
+                x = hash(value) ^ x
+        self._hash = x
+        return x
+
     def __str__(self,
 
                 join=_string.join):
 
         l = []
         append = l.append
+
+        # Format date part
         if self.year is not None:
             append('%04i-' % self.year)
         elif self.years:
@@ -747,24 +828,52 @@ class RelativeDateTime:
         if self.weekday:
             append(' %s:%i' % (Weekday[self.weekday[0]][:3],self.weekday[1]))
         append(' ')
+        
+        # Normalize relative time values to avoid fractions
+        hours = self.hours
+        minutes = self.minutes
+        seconds = self.seconds
+        hours_fraction = hours - int(hours)
+        minutes = minutes + hours_fraction * 60.0
+        minutes_fraction = minutes - int(minutes)
+        seconds = seconds + minutes_fraction * 6.0
+        seconds_fraction = seconds - int(seconds)
+
+        if 0:
+            # Normalize to standard time ranges
+            if seconds > 60.0:
+                extra_minutes, seconds = divmod(seconds, 60.0)
+                minutes = minutes + extra_minutes
+            elif seconds < -60.0:
+                extra_minutes, seconds = divmod(seconds, -60.0)
+                minutes = minutes - extra_minutes
+            if minutes >= 60.0:
+                extra_hours, minutes = divmod(minutes, 60.0)
+                hours = hours + extra_hours
+            elif minutes <= -60.0:
+                extra_hours, minutes = divmod(minutes, -60.0)
+                hours = hours - extra_hours
+
+        # Format time part
         if self.hour is not None:
             append('%02i:' % self.hour)
-        elif self.hours:
-            append('(%0+3i):' % self.hours)
+        elif hours:
+            append('(%0+3i):' % hours)
         else:
             append('HH:')
         if self.minute is not None:
             append('%02i:' % self.minute)
-        elif self.minutes:
-            append('(%0+3i):' % self.minutes)
+        elif minutes:
+            append('(%0+3i):' % minutes)
         else:
             append('MM:')
         if self.second is not None:
             append('%02i' % self.second)
-        elif self.seconds:
-            append('(%0+3i)' % self.seconds)
+        elif seconds:
+            append('(%0+3i)' % seconds)
         else:
             append('SS')
+            
         return join(l,'')
 
     def __repr__(self):
@@ -777,9 +886,9 @@ class RelativeDateTime:
 # Alias
 RelativeDate = RelativeDateTime
 
-def RelativeDateTimeFrom(*args,**kws):
+def RelativeDateTimeFrom(*args, **kws):
 
-    """ RelativeDateTimeFrom(*args,**kws)
+    """ RelativeDateTimeFrom(*args, **kws)
 
         Generic RelativeDateTime instance constructor. Can handle
         parsing strings and keywords.
@@ -788,10 +897,9 @@ def RelativeDateTimeFrom(*args,**kws):
     if len(args) == 1:
         # Single argument
         arg = args[0]
-        import types
-        if type(arg) is types.StringType:
+        if _isstring(arg):
             import Parser
-            return Parser.RelativeDateTimeFromString(arg)
+            return apply(Parser.RelativeDateTimeFromString, args, kws)
         elif isinstance(arg, RelativeDateTime):
             return arg
         else:
@@ -895,10 +1003,14 @@ Age = RelativeDateTimeDiff
 
 ###
 
+_current_year = now().year
+_current_century, _current_year_in_century = divmod(_current_year, 100)
+_current_century = _current_century * 100
+
 def add_century(year,
 
-                current_year=now().year,
-                current_century=(now().year / 100) * 100):
+                current_year=_current_year,
+                current_century=_current_century):
     
     """ Sliding window approach to the Y2K problem: adds a suitable
         century to the given year and returns it as integer.
@@ -927,16 +1039,16 @@ def add_century(year,
 
 def gregorian_jdn(year,month,day):
 
+    # XXX These require proper integer division.
     a = (14-month)/12
     y = year+4800-a
     m = month + 12*a - 3
-
     return day + (306*m+5)/10 + y*365 + y/4 - y/100 + y/400 - 32045
 
 def julian_jdn(year,month,day):
 
+    # XXX These require proper integer division.
     a = (14-month)/12
     y = year+4800-a
     m = month + 12*a - 3
-
     return day + (306*m+5)/10 + y*365 + y/4 - 32083
