@@ -11,7 +11,7 @@
     with the number of calibration rounds you wish to apply.
 
     Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 
@@ -24,6 +24,9 @@ import re,errno,exceptions,string,time
 # Enable to produce debugging output
 _debug = 0
 
+# Timeout waiting for a daytime server to respond in seconds
+TIMEOUT = 4.0
+
 # Port numbers
 try:
     DAYTIME_PORT = socket.getservbyname('daytime','tcp')
@@ -100,7 +103,7 @@ def _parse_datetime(daytime):
     """
     m = _daytimeRE.match(daytime)
     if not m:
-        raise ValueError,'unkown daytime format: "%s"' % daytime
+        raise ValueError,'unknown daytime format: "%s"' % daytime
     groups = m.groups()
     l = list(groups)
     for i in range(len(l)):
@@ -116,18 +119,12 @@ def _parse_datetime(daytime):
 
 ### Internal functions
     
-def _connect(host,port,timeout,
-
-             socket=socket.socket,AF_INET=socket.AF_INET,
-             SOCK_STREAM=socket.SOCK_STREAM,select=select.select,
-             DAYTIME_PORT=DAYTIME_PORT,EINPROGRESS=errno.EINPROGRESS,
-             error=socket.error,ip_cache=ip_cache,
-             gethostbyname=socket.gethostbyname):
+def _connect(host, port, timeout):
 
     """ Connect to the given host and port and return a socket
         instance.
 
-        timeout states the number of seconds to wait for the
+        timeout defines the number of seconds to wait for the
         connection to be established.
 
         In case no connection is possible, None is returned.
@@ -139,44 +136,51 @@ def _connect(host,port,timeout,
         host = ip_cache[hostname]
     else:
         try:
-            host = gethostbyname(hostname)
+            host = socket.gethostbyname(hostname)
         except error:
             # Host not found or some other error
             return None
         else:
             ip_cache[hostname] = host
-    s = socket(AF_INET,SOCK_STREAM)
+    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
     # Non-blocking mode
     s.setblocking(0)
-    while 1:
+    rc = -1
+    for i in range(10):
         try:
-            rc = s.connect_ex((host,DAYTIME_PORT))
-        except error,why:
+            rc = s.connect_ex((host, DAYTIME_PORT))
+        except socket.error, why:
             if _debug: print ' error: %s' % why
             return None
-        if _debug: print ' received %s' % errno.errorcode[rc]
-        if rc == EINPROGRESS:
-            # In progress... wait
-            r,w,e = select([],[s],[],timeout)
+        if _debug: print ' received %s' % errno.errorcode.get(rc, rc)
+        if rc in (errno.EAGAIN, errno.EWOULDBLOCK):
+            # Try again
+            continue
+        if rc == errno.EINPROGRESS:
+            # In progress... wait for connect
+            r,w,e = select.select([],[s],[],timeout)
             if w:
                 # Connect
                 s.setblocking(1)
                 try:
-                    rc = s.connect_ex((host,DAYTIME_PORT))
-                except error,why:
+                    rc = s.connect_ex((host, DAYTIME_PORT))
+                except socket.error, why:
                     # Should not happen, but does for e.g. 'host not found'
                     if _debug: print ' error: %s' % why
                     continue
+                if _debug: print ' received %s' % errno.errorcode.get(rc, rc)
             else:
                 if _debug: print ' timeout'
                 rc = -1
         break
+    else:
+        return None
     if rc != 0:
         return None
     else:
         return s
 
-def _read(socket,size,timeout,
+def _read(socket, size, timeout,
 
           select=select.select):
 
@@ -193,7 +197,7 @@ def _read(socket,size,timeout,
         if _debug: print ' read timeout'
         return None
 
-def _get_daytime(daytime_servers=daytime_servers,timeout=4,
+def _get_daytime(daytime_servers=daytime_servers, timeout=TIMEOUT,
 
                  strip=string.strip,_parse_datetime=_parse_datetime,
                  DateTimeFromMJD=DateTime.DateTimeFromMJD,
@@ -209,16 +213,20 @@ def _get_daytime(daytime_servers=daytime_servers,timeout=4,
         daytime protocol. They can be given by name or IP.
 
         timeout is used as timeout when connecting to each server.  It
-        must be given in seconds. The total timeout depends on the
-        number of servers listed in daytime_servers.
+        must be given in seconds and defaults to TIMEOUT. The total
+        timeout depends on the number of servers listed in
+        daytime_servers.
 
     """
+    # We use the timeout value for connect and read, so make sure we
+    # never spend more time on a single server
+    timeout = timeout / 2.0
     if _debug: print 'Contacting NIST servers...'
     for host in daytime_servers:
         if _debug: print 'Trying server: %s' % host
-        s = _connect(host,DAYTIME_PORT,timeout)
+        s = _connect(host, DAYTIME_PORT, timeout)
         if s:
-            data = _read(s,512,0.5)
+            data = _read(s, 512, timeout)
             if data is None:
                 continue
             daytime = strip(data)
