@@ -9,7 +9,7 @@
     It is safe to do a 'from mx.Log import *'.
 
     Copyright (c) 1999-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 
@@ -17,7 +17,7 @@
 # Be careful with these symbols since we want 'from mx.Log import *'
 # to not clutter up the users namespace.
 import sys,string,types,traceback
-from mx import DateTime
+from mx import DateTime,Tools
 _sys,_string,_DateTime,_types,_traceback = sys,string,DateTime,types,traceback
 del sys,string,DateTime,types,traceback
 
@@ -100,34 +100,48 @@ def print_tb_vars(tb=None, *args, **kws):
     # Show the global variables
     apply(print_frame_globals, (tb.tb_frame,) + args, kws)
 
+def filter_private_attributes(key):
+
+    """ Filters all attributes which start with an underscore.
+
+        For use as filter argument to one of the print_xxx() APIs.
+    
+    """
+    if type(key) is _types.StringType and \
+       key[:1] == '_':
+        return 0
+    return 1
+
 def print_frame_locals(frame,file=_sys.stdout,
                        indent='#  ',
                        title='### Dump of local variables:\n',
                        levels=2,
-                       nonrecursive=_NON_RECURSIVE):
+                       nonrecursive=_NON_RECURSIVE,
+                       filter=None):
 
     """ Print a listing of the locals() in frame to file.
     """
     if title:
         file.write(title)
     print_dict(frame.f_locals,file,indent,levels,
-               nonrecursive=nonrecursive)
+               nonrecursive=nonrecursive,filter=filter)
 
 def print_frame_globals(frame,file=_sys.stdout,
                         indent='#  ',
                         title='### Dump of global variables:\n',
                         levels=2,
-                        nonrecursive=_NON_RECURSIVE):
+                        nonrecursive=_NON_RECURSIVE,
+                        filter=None):
 
     """ Print a listing of the globals() in frame to file.
     """
     if title:
         file.write(title)
     print_dict(frame.f_globals,file,indent,levels,
-               nonrecursive=nonrecursive)
+               nonrecursive=nonrecursive,filter=filter)
 
 def print_dict(dict,file,indent='',levels=1,reprkeys=0,
-               nonrecursive=()):
+               nonrecursive=(),filter=None):
 
     """ Print dictionary to file.
 
@@ -135,6 +149,10 @@ def print_dict(dict,file,indent='',levels=1,reprkeys=0,
         indent is prepended to all lines. levels indicates the number
         of recursion levels to be printed.
 
+        filter may be given as callable taking the dictionary key as
+        input. It should then return 1 for keys which should be
+        displayed and 0 for ones which should be skipped.
+
     """
     l = dict.items()
     l.sort()
@@ -142,6 +160,9 @@ def print_dict(dict,file,indent='',levels=1,reprkeys=0,
         l = l[:_DICT_LEN_LIMIT]
         l.append(('...','...truncated...'))
     for k,v in l:
+        if filter is not None:
+            if not filter(k):
+                continue
         if reprkeys:
             try:
                 n = repr(k)
@@ -164,18 +185,18 @@ def print_dict(dict,file,indent='',levels=1,reprkeys=0,
             file.write('%s.%-15s = %s\n' % (indent,n,r))
         if levels > 1:
             print_recursive(v,file,indent + '  ',levels-1,
-                            nonrecursive=nonrecursive)
+                            nonrecursive=nonrecursive, filter=filter)
 
 def print_sequence(obj,file=_sys.stdout,indent='',levels=2,
                    nonrecursive=(),
 
-                   StringType=_types.StringType,join=_string.join):
+                   StringType=_types.StringType):
 
     l = []
     unfold = 0
     try:
         length = len(obj)
-    except (AttributeError, ValueError):
+    except (AttributeError, ValueError, TypeError):
         return
     for i in trange(min(length,_VALUE_LEN_LIMIT)):
         try:
@@ -202,7 +223,7 @@ def print_sequence(obj,file=_sys.stdout,indent='',levels=2,
                                 nonrecursive=nonrecursive)
 
 def print_recursive(obj,file=_sys.stdout,indent='',levels=1,
-                    nonrecursive=(),
+                    nonrecursive=(),filter=None,
 
                     StringType=_types.StringType):
 
@@ -217,12 +238,13 @@ def print_recursive(obj,file=_sys.stdout,indent='',levels=1,
         return
     
     # Print the object depending on its interface
-    if hasattr(obj,'__dict__'):
+    if hasattr(obj,'__dict__') and \
+       obj.__dict__ is not None:
         print_dict(obj.__dict__,file,indent,levels,
-                   nonrecursive=nonrecursive)
+                   nonrecursive=nonrecursive, filter=filter)
     elif hasattr(obj,'items'):
         print_dict(obj,file,indent,levels,1,
-                   nonrecursive=nonrecursive)
+                   nonrecursive=nonrecursive, filter=filter)
     elif issequence(obj) and not type(obj) == StringType:
         print_sequence(obj,file,indent,levels,
                        nonrecursive=nonrecursive)
@@ -230,11 +252,11 @@ def print_recursive(obj,file=_sys.stdout,indent='',levels=1,
         d = {}
         for attr in obj.__members__:
             d[attr] = getattr(obj,attr)
-        print_dict(d,file,indent,levels,
-                   nonrecursive=nonrecursive)
+        print_dict(d, file, indent, levels,
+                   nonrecursive=nonrecursive, filter=filter)
 
 def print_obj(obj,file=_sys.stdout,indent='',levels=2,
-              nonrecursive=()):
+              nonrecursive=(),filter=None):
 
     try:
         r = repr(obj)
@@ -244,7 +266,7 @@ def print_obj(obj,file=_sys.stdout,indent='',levels=2,
     if len(r) > _VALUE_LEN_LIMIT:
         r = r[:_VALUE_LEN_LIMIT] + '...'
     file.write('%s%s\n' % (indent,r))
-    print_recursive(obj,file,indent + '  ',levels-1)
+    print_recursive(obj, file, indent + '  ', levels-1, filter=filter)
 
 def print_stack_with_locals(file=_sys.stdout,levels=100,offset=0):
 
@@ -254,8 +276,8 @@ def print_stack(file=_sys.stdout,levels=100,offset=0,locals=0):
 
     # Prepare frames
     try:
-        1/0
-    except ZeroDivisionError:
+        raise ValueError
+    except ValueError:
         # Go back offset+1 frames...
         f = _sys.exc_info()[2].tb_frame
         for i in range(offset + 1):
@@ -300,19 +322,23 @@ def print_loaded_modules(file=_sys.stdout,
                 # module
                 file.write(' %s\n' % (n))
 
-def format_traceback():
+def format_traceback(title='### Dump of local variables:\n',
+                     filter=None):
 
     """ Formats a traceback of the current exception and
         returns it as string.
 
-        Includes a listing of the locals() causing that were
-        active during the exception.
+        Includes a listing of the locals() causing that were active
+        during the exception. title is used to separate the traceback
+        from the listing of variables. filter can be given to filter
+        out certain variables from the listing. It takes one argument:
+        the variable name and has to return 1/0.
 
     """
     import cStringIO
     f = cStringIO.StringIO()
     _traceback.print_exc(_TRACEBACK_LIMIT,f)
-    print_tb_locals(_sys.exc_info()[2],f)
+    print_tb_locals(_sys.exc_info()[2], f, title=title, filter=filter)
     return f.getvalue()
 
 ### Log class
@@ -464,7 +490,9 @@ class Log:
         # Print traceback and locals to file
         f.write(79*'-'+ '\n')
         _traceback.print_exc(_TRACEBACK_LIMIT,f)
-        print_tb_locals(_sys.exc_info()[2],f)
+        print_tb_locals(_sys.exc_info()[2], f,
+                        title='### Dump of local variables (omitting private ones):\n',
+                        filter=filter_private_attributes)
         f.write(79*'-'+'\n')
         f.flush()
 
@@ -621,18 +649,16 @@ class Log:
         if errno >= self.ignore_level: return
         f = self.open_log_file or self.open()
         f.write("%04i %s [%s] %s\n" %
-                (errno,_DateTime.now(),self.log_id,
-                                       objname))
+                (errno, _DateTime.now(), self.log_id, objname))
         f.write(79*'-'+'\n')
         print_obj(object,f,levels=levels)
         f.write(79*'-'+'\n')
         f.flush()
         
-
     def write(self,data):
 
         """ File interface, so that the object can be used as
-            _sys.stderr replacement.
+            sys.stderr replacement.
 
             Errorcode 0 is used. The output is line buffered.
             
@@ -642,7 +668,7 @@ class Log:
         if '\n' not in buffer:
             return
 
-        # Spit out all complete lines stored in line_buffer
+        # Write out all complete lines stored in line_buffer
         lines = _string.split(buffer,'\n')
         if buffer[-1:] != '\n':
             # last line is incomplete
@@ -654,13 +680,21 @@ class Log:
         f = self.open_log_file or self.open()
         now = _DateTime.now()
         id = self.log_id
+        fwrite = f.write
         for line in lines:
             if not _string.strip(line):
                 # avoid emtpy lines
                 continue
-            f.write("%04i %s [%s] %s\n" % (0,now,id,line))
+            fwrite("%04i %s [%s] %s\n" % (0,now,id,line))
         f.flush()
 
+    def flush(self):
+
+        """ Flush the underlying log stream.
+        """
+        if self.open_log_file:
+            self.open_log_file.flush()
+
     def exception(self,errno,*args):
 
         """ Log a real exception in the log file.
@@ -692,7 +726,7 @@ class Log:
         del exc_class,exc_inst,exc_traceback
 
 # Create a main log object
-if __debug__:
+if __debug__ and Tools.debugging():
     log = Log(SYSTEM_LOG_EVERYTHING)
 else:
     log = Log(SYSTEM_INFO+1)
