@@ -2,7 +2,7 @@
   mxProxy -- Proxy wrapper type
 
   Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author (mailto:mal@lemburg.com).
 
@@ -20,11 +20,14 @@
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXPROXY
 
+/* Mark the module as Py_ssize_t clean. */
+#define PY_SSIZE_T_CLEAN 1
+
 #include "mx.h"
 #include "mxProxy.h"
 
 /* Version number: Major.Minor.Patchlevel */
-#define VERSION "2.0.2"
+#define MXPROXY_VERSION "3.0.0"
 
 /* Define these to have the module use free lists (saves malloc calls) */
 #define MXPROXY_FREELIST
@@ -33,10 +36,10 @@
 
 static char *Module_docstring = 
 
- MXPROXY_MODULE" -- Generic proxy wrapper type. Version "VERSION"\n\n"
+ MXPROXY_MODULE" -- Generic proxy wrapper type. Version "MXPROXY_VERSION"\n\n"
 
  "Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com\n"
- "Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
+ "Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -63,6 +66,9 @@ static mxProxyObject *mxProxy_FreeList = NULL;
 
 static PyObject *mxProxy_WeakReferences;	
 
+/* Flag telling us whether the module was initialized or not. */
+static int mxProxy_Initialized = 0;
+
 /* --- forward declarations ----------------------------------------------- */
 
 staticforward PyTypeObject mxProxy_Type;
@@ -90,8 +96,31 @@ PyObject *insexc(PyObject *moddict,
 {
     PyObject *v;
     char fullname[256];
+    char *modname;
+    char *dot;
     
-    sprintf(fullname,MXPROXY_MODULE".%s",name);
+    v = PyDict_GetItemString(moddict, "__name__");
+    if (v == NULL)
+	modname = NULL;
+    else
+	modname = PyString_AsString(v);
+    if (modname == NULL) {
+	PyErr_Clear();
+	modname = MXPROXY_MODULE;
+    }
+    /* The symbols from this extension are imported into
+       mx.<packagename>. We trim the name to not confuse the user with
+       an overly long package path. */
+    strcpy(fullname, modname);
+    dot = strchr(fullname, '.');
+    if (dot)
+	dot = strchr(dot+1, '.');
+    if (dot)
+	strcpy(dot+1, name);
+    else
+	sprintf(fullname, "%s.%s", modname, name);
+
+
     v = PyErr_NewException(fullname, base, NULL);
     if (v == NULL)
 	return NULL;
@@ -146,7 +175,7 @@ void insobj(PyObject *dict,
 static
 PyObject *seq2dict(PyObject *sequence)
 {
-    int i,len;
+    Py_ssize_t i,len;
     PyObject *v,*w;
 
     len = PySequence_Length(sequence);
@@ -216,8 +245,8 @@ int mxProxy_RegisterWeakReference(mxProxyObject *proxy,
     if (id == NULL)
 	goto onError;
     DPRINTF("mxProxy_RegisterWeakReference: "
-	    "proxy at %0x, object at %0x -> id at %0x\n",
-	    (int)proxy,(int)object,(int)id);
+	    "proxy at %0lx, object at %0lx -> id at %0lx\n",
+	    (long)proxy,(long)object,(long)id);
     Py_Assert(mxProxy_WeakReferences != NULL && 
 	      mxProxy_WeakReferences->ob_refcnt > 0,
 	      mxProxy_InternalError,
@@ -262,8 +291,8 @@ int mxProxy_RegisterWeakReference(mxProxyObject *proxy,
 	if (rc != 0)
 	    goto onError;
 	DPRINTF("mxProxy_RegisterWeakReference: "
-		"added a new weak dict entry id %0x in dict %0x\n",
-		(int)id,(int)mxProxy_WeakReferences);
+		"added a new weak dict entry id %0lx in dict %0lx\n",
+		(long)id,(long)mxProxy_WeakReferences);
 	DPRINTF("                               "
 		"refcounts... id:%i, dict:%i\n",
 		id->ob_refcnt,mxProxy_WeakReferences->ob_refcnt);
@@ -286,7 +315,7 @@ int mxProxy_RegisterWeakReference(mxProxyObject *proxy,
 static
 int mxProxy_DefuncWeakProxies(mxProxyObject *proxy)
 {
-    DPRINTF("mxProxy_DefuncWeakProxies: proxy at %0x\n",(int)proxy);
+    DPRINTF("mxProxy_DefuncWeakProxies: proxy at %0lx\n",(long)proxy);
     do {
 	mxProxy_DefuncObjectReference(proxy);
 	proxy = proxy->next_weak_proxy;
@@ -316,8 +345,8 @@ int mxProxy_CollectWeakReference(mxProxyObject *proxy)
 
     id = proxy->object;
     DPRINTF("mxProxy_CollectWeakReference: "
-	    "proxy at %0x, id at %0x\n",
-	    (int)proxy,(int)id);
+	    "proxy at %0lx, id at %0lx\n",
+	    (long)proxy,(long)id);
     Py_Assert(mxProxy_WeakReferences != NULL && 
 	      mxProxy_WeakReferences->ob_refcnt > 0,
 	      mxProxy_InternalError,
@@ -350,8 +379,8 @@ int mxProxy_CollectWeakReference(mxProxyObject *proxy)
 
 	/* Remove the mxProxy_WeakReferences entry for this object */
 	DPRINTF("mxProxy_CollectWeakReference: "
-		"removing the weak dict entry id %0x from dict %0x\n",
-		(int)id,(int)mxProxy_WeakReferences);
+		"removing the weak dict entry id %0lx from dict %0lx\n",
+		(long)id,(long)mxProxy_WeakReferences);
 	DPRINTF("                              "
 		"refcounts... id:%i, dict:%i\n",
 		id->ob_refcnt,mxProxy_WeakReferences->ob_refcnt);
@@ -381,8 +410,8 @@ int mxProxy_DeregisterWeakReference(mxProxyObject *proxy)
     PyObject *id,*v,*w;
 
     id = proxy->object;
-    DPRINTF("mxProxy_DeregisterWeakReference: proxy at %0x, id at %0x\n",
-	    (int)proxy,(int)id);
+    DPRINTF("mxProxy_DeregisterWeakReference: proxy at %0lx, id at %0lx\n",
+	    (long)proxy,(long)id);
     Py_Assert(mxProxy_WeakReferences != NULL && 
 	      mxProxy_WeakReferences->ob_refcnt > 0,
 	      mxProxy_InternalError,
@@ -473,8 +502,8 @@ PyObject *mxProxy_GetWeakReferenceObject(mxProxyObject *proxy)
     PyObject *id,*v,*w;
 
     id = proxy->object;
-    DPRINTF("mxProxy_GetWeakReferenceObject: proxy at %0x, id at %0x\n",
-	    (int)proxy,(int)id);
+    DPRINTF("mxProxy_GetWeakReferenceObject: proxy at %0lx, id at %0lx\n",
+	    (long)proxy,(long)id);
     Py_Assert(mxProxy_WeakReferences != NULL && 
 	      mxProxy_WeakReferences->ob_refcnt > 0,
 	      mxProxy_InternalError,
@@ -496,8 +525,8 @@ PyObject *mxProxy_GetWeakReferenceObject(mxProxyObject *proxy)
 	}
 	w = PyTuple_GET_ITEM(v,0);
 	Py_INCREF(w);
-	DPRINTF("mxProxy_GetWeakReferenceObject: returning object at %0x\n",
-		(int)w);
+	DPRINTF("mxProxy_GetWeakReferenceObject: returning object at %0lx\n",
+		(long)w);
 	return w;
     }
     else
@@ -513,7 +542,7 @@ int _mxProxy_CollectWeakReferences(int force)
 {
     PyObject *id,*v,*collect = 0;
     mxProxyObject *proxy;
-    int i;
+    Py_ssize_t i;
 
     Py_Assert(mxProxy_WeakReferences != NULL &&
 	      mxProxy_WeakReferences->ob_refcnt > 0,
@@ -538,8 +567,8 @@ int _mxProxy_CollectWeakReferences(int force)
 	    if (!proxy)
 		goto onError;
 	    DPRINTF("mxProxy_CheckWeakReferenceDict: "
-		    "found proxy at %0x, object at %0x -> id at %0x\n",
-		    (int)proxy,(int)PyTuple_GET_ITEM(v,0),(int)id);
+		    "found proxy at %0lx, object at %0lx -> id at %0lx\n",
+		    (long)proxy,(long)PyTuple_GET_ITEM(v,0),(long)id);
 	    PyList_Append(collect,(PyObject *)proxy);
 	}
     }
@@ -671,8 +700,11 @@ mxProxyObject *mxProxy_New(PyObject *object,
     }
     else {
 	/* Store the index object instead of the object itself. */
-	if (mxProxy_RegisterWeakReference(proxy,object))
+	if (mxProxy_RegisterWeakReference(proxy, object)) {
+	    mxPy_UNREF(proxy);
+	    PyObject_Del(proxy);
 	    goto onError;
+	}
     }
 
     /* Errors may not occur after this line, or the onError-mechanism
@@ -723,7 +755,7 @@ void mxProxy_Free(mxProxyObject *proxy)
 {
     PyObject *error_type, *error_value, *error_traceback;
 
-    DPRINTF("mxProxy_Free: instance at %0x\n",(int)proxy);
+    DPRINTF("mxProxy_Free: instance at %0lx\n",(long)proxy);
 
     /* Allow cleanup of the wrapped object; errors are printed to stderr */
     if (proxy->cleanup) {
@@ -813,7 +845,7 @@ void mxProxy_Free(mxProxyObject *proxy)
     *(mxProxyObject **)proxy = mxProxy_FreeList;
     mxProxy_FreeList = proxy;
 #else
-    PyMem_DEL(proxy);
+    PyObject_Del(proxy);
 #endif
     DPRINTF("mxProxy_Free: dealloced\n");
 }
@@ -834,8 +866,8 @@ int mxProxy_SlotAccessAllowed(mxProxyObject *self,
     PyObject *v;
     
     DPRINTF("mxProxy_SlotAccessAllowed: "
-	    "proxy at %0x, object/index at %0x, slot '%.200s' -- ",
-	    (int)self,(int)self->object,PyString_AS_STRING(name));
+	    "proxy at %0lx, object/index at %0lx, slot '%.200s' -- ",
+	    (long)self,(long)self->object,PyString_AS_STRING(name));
 
     /* Is interface access restricted ? */
     if (!self->interface) {
@@ -864,8 +896,8 @@ static
 int mxProxy_DefuncObjectReference(mxProxyObject *self)
 {
     DPRINTF("mxProxy_DefuncObjectReference: "
-	    "proxy at %0x, object/index at %0x\n",
-	    (int)self,(int)self->object);
+	    "proxy at %0lx, object/index at %0lx\n",
+	    (long)self,(long)self->object);
     Py_XDECREF(self->object);
     self->object = NULL;
     return 0;
@@ -874,9 +906,10 @@ int mxProxy_DefuncObjectReference(mxProxyObject *self)
 /* Get attribute name from the wrapped object */
 
 static
-PyObject *mxProxy_GetattrObject(mxProxyObject *self,
+PyObject *mxProxy_GetattrObject(PyObject *obj,
 				PyObject *name)
 {
+    mxProxyObject *self = (mxProxyObject *)obj;
     PyObject *v;
 
     DPRINTF("mxProxy_GetattrObject: %.200s\n",
@@ -967,10 +1000,11 @@ PyObject *mxProxy_GetattrObject(mxProxyObject *self,
    case of error, 0 otherwise. */
 
 static
-int mxProxy_SetattrObject(mxProxyObject *self,
+int mxProxy_SetattrObject(PyObject *obj,
 			  PyObject *name,
 			  PyObject *value)
 {
+    mxProxyObject *self = (mxProxyObject *)obj;
     PyObject *v;
     
     DPRINTF("mxProxy_SetattrObject: %.200s to %.200s\n",
@@ -1097,7 +1131,7 @@ Py_C_Function( mxProxy_proxy_getattr,
     PyObject *name;
     
     Py_GetArg("O",name);
-    return mxProxy_GetattrObject(proxy,name);
+    return mxProxy_GetattrObject(self, name);
     
  onError:
     return NULL;
@@ -1112,7 +1146,7 @@ Py_C_Function( mxProxy_proxy_setattr,
     PyObject *name,*value;
     
     Py_Get2Args("OO",name,value);
-    if (mxProxy_SetattrObject(proxy,name,value))
+    if (mxProxy_SetattrObject(self, name, value))
 	goto onError;
     Py_ReturnNone();
     
@@ -1125,8 +1159,9 @@ Py_C_Function( mxProxy_proxy_setattr,
 /* --- slots --- */
 
 static
-PyObject *mxProxy_Repr(mxProxyObject *self)
+PyObject *mxProxy_Repr(PyObject *obj)
 {
+    mxProxyObject *self = (mxProxyObject *)obj;
     char t[100];
 
     if (self->isWeak) {
@@ -1144,8 +1179,9 @@ PyObject *mxProxy_Repr(mxProxyObject *self)
 
 #define SLOT_API_PPP(name,slot,function,arg1type,arg2type,arg3type,rctype,errorrc) \
 static										   \
-rctype name(mxProxyObject *self, arg1type v, arg2type w, arg3type x)		   \
+rctype name(PyObject *obj, arg1type v, arg2type w, arg3type x)		   \
 {										   \
+    mxProxyObject *self = (mxProxyObject *)obj;		\
     static PyObject *slotstr;							   \
 										   \
     if (slotstr == NULL)								   \
@@ -1172,8 +1208,9 @@ rctype name(mxProxyObject *self, arg1type v, arg2type w, arg3type x)		   \
 
 #define SLOT_API_PP(name,slot,function,arg1type,arg2type,rctype,errorrc)	   \
 static										   \
-rctype name(mxProxyObject *self, arg1type v, arg2type w)			   \
+rctype name(PyObject *obj, arg1type v, arg2type w)			   \
 {										   \
+    mxProxyObject *self = (mxProxyObject *)obj;		\
     static PyObject *slotstr;							   \
 										   \
     if (slotstr == NULL)								   \
@@ -1200,8 +1237,9 @@ rctype name(mxProxyObject *self, arg1type v, arg2type w)			   \
 
 #define SLOT_API_P(name,slot,function,arg1type,rctype,errorrc)			   \
 static										   \
-rctype name(mxProxyObject *self, arg1type v)					   \
+rctype name(PyObject *obj, arg1type v)					   \
 {										   \
+    mxProxyObject *self = (mxProxyObject *)obj;		\
     static PyObject *slotstr;							   \
 										   \
     if (slotstr == NULL)								   \
@@ -1226,32 +1264,33 @@ rctype name(mxProxyObject *self, arg1type v)					   \
     return errorrc;								   \
 }
 
-#define SLOT_API(name,slot,function,rctype,errorrc)				   \
-static										   \
-rctype name(mxProxyObject *self)						   \
-{							       			   \
-    static PyObject *slotstr;							   \
-										   \
-    if (slotstr == NULL)								   \
-	slotstr = PyString_InternFromString(#slot);			   \
-    Py_Assert(mxProxy_SlotAccessAllowed(self,slotstr),				   \
-	      mxProxy_AccessError,						   \
-	      #slot" access denied");					   \
-    if (self->isWeak) {								   \
-        PyObject *object;							   \
-        rctype rc;								   \
-	object = mxProxy_GetWeakReferenceObject(self);				   \
-	if (!object)								   \
-	    goto onError;							   \
-        rc = function(object);							   \
-        Py_DECREF(object);							   \
-	return rc;								   \
-    }										   \
-    else									   \
-        return function(self->object);						   \
-										   \
- onError:									   \
-    return errorrc;								   \
+#define SLOT_API(name,slot,function,rctype,errorrc)	\
+static							\
+rctype name(PyObject *obj)				\
+{							\
+    mxProxyObject *self = (mxProxyObject *)obj;		\
+    static PyObject *slotstr;				\
+							\
+    if (slotstr == NULL)				\
+	slotstr = PyString_InternFromString(#slot);	\
+    Py_Assert(mxProxy_SlotAccessAllowed(self,slotstr),	\
+	      mxProxy_AccessError,			\
+	      #slot" access denied");			\
+    if (self->isWeak) {					\
+        PyObject *object;				\
+        rctype rc;					\
+	object = mxProxy_GetWeakReferenceObject(self);	\
+	if (!object)					\
+	    goto onError;				\
+        rc = function(object);				\
+        Py_DECREF(object);				\
+	return rc;					\
+    }							\
+    else						\
+        return function(self->object);			\
+							\
+ onError:						\
+    return errorrc;					\
 }
 
 /* Basic slots */
@@ -1263,15 +1302,15 @@ SLOT_API_P(mxProxy_Compare,__cmp__,PyObject_Compare,PyObject*,int,-1)
 /* Mapping */
 SLOT_API_P(mxProxy_GetItem,__getitem__,PyObject_GetItem,PyObject*,PyObject*,NULL)
 SLOT_API_PP(mxProxy_SetItem,__setitem__,PyObject_SetItem,PyObject*,PyObject*,int,-1)
-SLOT_API(mxProxy_Length,__len__,PyObject_Length,int,-1)
+SLOT_API(mxProxy_Length,__len__,PyObject_Length,Py_ssize_t,-1)
 
 /* Sequence */
 SLOT_API_P(mxProxy_Concat,__add__,PySequence_Concat,PyObject*,PyObject*,NULL)
-SLOT_API_P(mxProxy_Repeat,__repeat__,PySequence_Repeat,int,PyObject*,NULL)
-SLOT_API_P(mxProxy_GetIndex,__getitem__,PySequence_GetItem,int,PyObject*,NULL)
-SLOT_API_PP(mxProxy_GetSlice,__getslice__,PySequence_GetSlice,int,int,PyObject*,NULL)
-SLOT_API_PP(mxProxy_SetIndex,__setitem__,PySequence_SetItem,int,PyObject*,int,-1)
-SLOT_API_PPP(mxProxy_SetSlice,__getitem__,PySequence_SetSlice,int,int,PyObject*,int,-1)
+SLOT_API_P(mxProxy_Repeat,__repeat__,PySequence_Repeat,Py_ssize_t,PyObject*,NULL)
+SLOT_API_P(mxProxy_GetIndex,__getitem__,PySequence_GetItem,Py_ssize_t,PyObject*,NULL)
+SLOT_API_PP(mxProxy_GetSlice,__getslice__,PySequence_GetSlice,Py_ssize_t,Py_ssize_t,PyObject*,NULL)
+SLOT_API_PP(mxProxy_SetIndex,__setitem__,PySequence_SetItem,Py_ssize_t,PyObject*,int,-1)
+SLOT_API_PPP(mxProxy_SetSlice,__getitem__,PySequence_SetSlice,Py_ssize_t,Py_ssize_t,PyObject*,int,-1)
 
 /* Number
 
@@ -1305,45 +1344,45 @@ SLOT_API(mxProxy_Float,__float__,PyNumber_Float,PyObject*,NULL)
 /* Python Type Tables */
 
 static PyNumberMethods mxProxy_NumberSlots = {
-	(binaryfunc)mxProxy_Add, /*nb_add*/
-	(binaryfunc)mxProxy_Sub, /*nb_subtract*/
-	(binaryfunc)mxProxy_Multiply, /*nb_multiply*/
-	(binaryfunc)mxProxy_Divide, /*nb_divide*/
-	(binaryfunc)mxProxy_Remainder, /*nb_remainder*/
-	(binaryfunc)mxProxy_Divmod, /*nb_divmod*/
-	(ternaryfunc)mxProxy_Power, /*nb_power*/
-	(unaryfunc)mxProxy_Negative, /*nb_negative*/
-	(unaryfunc)mxProxy_Positive, /*nb_positive*/
-	(unaryfunc)mxProxy_Absolute, /*nb_absolute*/
-	(inquiry)mxProxy_IsTrue, /*nb_nonzero*/
-	(unaryfunc)mxProxy_Invert, /*nb_invert*/
-	(binaryfunc)mxProxy_Lshift, /*nb_lshift*/
-	(binaryfunc)mxProxy_Rshift, /*nb_rshift*/
-	(binaryfunc)mxProxy_And, /*nb_and*/
-	(binaryfunc)mxProxy_Xor, /*nb_xor*/
-	(binaryfunc)mxProxy_Or, /*nb_or*/
-	(coercion)0,		/*nb_coerce*/
-	(unaryfunc)mxProxy_Int, /*nb_int*/
-	(unaryfunc)mxProxy_Long, /*nb_long*/
-	(unaryfunc)mxProxy_Float, /*nb_float*/
-	(unaryfunc)0, /*nb_oct*/
-	(unaryfunc)0, /*nb_hex*/
+    mxProxy_Add, 			/*nb_add*/
+    mxProxy_Sub, 			/*nb_subtract*/
+    mxProxy_Multiply, 			/*nb_multiply*/
+    mxProxy_Divide, 			/*nb_divide*/
+    mxProxy_Remainder, 			/*nb_remainder*/
+    mxProxy_Divmod, 			/*nb_divmod*/
+    mxProxy_Power, 			/*nb_power*/
+    mxProxy_Negative, 			/*nb_negative*/
+    mxProxy_Positive, 			/*nb_positive*/
+    mxProxy_Absolute, 			/*nb_absolute*/
+    mxProxy_IsTrue, 			/*nb_nonzero*/
+    mxProxy_Invert, 			/*nb_invert*/
+    mxProxy_Lshift, 			/*nb_lshift*/
+    mxProxy_Rshift, 			/*nb_rshift*/
+    mxProxy_And, 			/*nb_and*/
+    mxProxy_Xor, 			/*nb_xor*/
+    mxProxy_Or, 			/*nb_or*/
+    0,					/*nb_coerce*/
+    mxProxy_Int, 			/*nb_int*/
+    mxProxy_Long, 			/*nb_long*/
+    mxProxy_Float, 			/*nb_float*/
+    0, 					/*nb_oct*/
+    0, 					/*nb_hex*/
 };
 
 static PySequenceMethods mxProxy_SequenceSlots = {
-	(inquiry)mxProxy_Length, 	/*sq_length*/
-	(binaryfunc)mxProxy_Concat, 	/*sq_concat*/
-	(intargfunc)mxProxy_Repeat, 	/*sq_repeat*/
-	(intargfunc)mxProxy_GetIndex, 	/*sq_item*/
-	(intintargfunc)mxProxy_GetSlice,/*sq_slice*/
-	(intobjargproc)mxProxy_SetIndex,/*sq_ass_item*/
-	(intintobjargproc)mxProxy_SetSlice,/*sq_ass_slice*/
+    mxProxy_Length, 			/*sq_length*/
+    mxProxy_Concat, 			/*sq_concat*/
+    mxProxy_Repeat, 			/*sq_repeat*/
+    mxProxy_GetIndex, 			/*sq_item*/
+    mxProxy_GetSlice,			/*sq_slice*/
+    mxProxy_SetIndex,			/*sq_ass_item*/
+    mxProxy_SetSlice,			/*sq_ass_slice*/
 };
 
 static PyMappingMethods mxProxy_MappingSlots = {
-	(inquiry)mxProxy_Length,	/*mp_length*/
-	(binaryfunc)mxProxy_GetItem, 	/*mp_subscript*/
-	(objobjargproc)mxProxy_SetItem,	/*mp_ass_subscript*/
+    mxProxy_Length,			/*mp_length*/
+    mxProxy_GetItem, 			/*mp_subscript*/
+    mxProxy_SetItem,			/*mp_ass_subscript*/
 };
 
 statichere
@@ -1355,22 +1394,22 @@ PyTypeObject mxProxy_Type = {
     0,			  		/*tp_itemsize*/
     /* slots */
     (destructor)mxProxy_Free,		/*tp_dealloc*/
-    (printfunc)0,		  	/*tp_print*/
-    (getattrfunc)0,		  	/*tp_getattr -- see tp_getattro */
-    (setattrfunc)0,		  	/*tp_setattr -- see tp_setattro */
-    (cmpfunc)mxProxy_Compare, 		/*tp_compare*/
-    (reprfunc)mxProxy_Repr,	  	/*tp_repr*/
+    0,		 		 	/*tp_print*/
+    0,				  	/*tp_getattr -- see tp_getattro */
+    0,				  	/*tp_setattr -- see tp_setattro */
+    mxProxy_Compare, 	       		/*tp_compare*/
+    mxProxy_Repr,	  		/*tp_repr*/
     &mxProxy_NumberSlots, 		/*tp_as_number*/
     &mxProxy_SequenceSlots,		/*tp_as_sequence*/
     &mxProxy_MappingSlots,		/*tp_as_mapping*/
-    (hashfunc)mxProxy_Hash,		/*tp_hash*/
-    (ternaryfunc)mxProxy_Call,		/*tp_call*/
-    (reprfunc)mxProxy_Str,		/*tp_str*/
-    (getattrofunc)mxProxy_GetattrObject,/*tp_getattro*/
-    (setattrofunc)mxProxy_SetattrObject,/*tp_setattro*/
+    mxProxy_Hash,			/*tp_hash*/
+    mxProxy_Call,			/*tp_call*/
+    mxProxy_Str,			/*tp_str*/
+    mxProxy_GetattrObject,		/*tp_getattro*/
+    mxProxy_SetattrObject,		/*tp_setattro*/
     0,					/*tp_as_buffer*/
     0,					/*tp_xxx4*/
-    (char*) 0				/*tp_doc*/
+    0					/*tp_doc*/
 };
 
 /* Python Method Table
@@ -1486,7 +1525,7 @@ void mxProxyModule_Cleanup(void)
 	while (d != NULL) {
 	    mxProxyObject *v = d;
 	    d = *(mxProxyObject **)d;
-	    free(v);
+	    PyObject_Del(v);
 	}
 	mxProxy_FreeList = NULL;
     }
@@ -1501,6 +1540,9 @@ void mxProxyModule_Cleanup(void)
 		"memory leak: weakref dictionary still alive\n");
     }
     mxProxy_WeakReferences = NULL;
+
+    /* Reset mxProxy_Initialized flag */
+    mxProxy_Initialized = 0;
 }
 
 /* Create PyMethodObjects and register them in the module's dict */
@@ -1509,6 +1551,10 @@ MX_EXPORT(void)
 {
     PyObject *module, *moddict;
 
+    if (mxProxy_Initialized)
+	Py_Error(PyExc_SystemError,
+		 "can't initialize "MXPROXY_MODULE" more than once");
+
     /* Init type objects */
     PyType_Init(mxProxy_Type);
 
@@ -1521,6 +1567,11 @@ MX_EXPORT(void)
     if (module == NULL)
 	goto onError;
 
+    /* Init globals */
+#ifdef MXPROXY_FREELIST
+    mxProxy_FreeList = NULL;
+#endif
+
     /* Register cleanup function */
     if (Py_AtExit(mxProxyModule_Cleanup))
 	/* XXX what to do if we can't register that function ??? */;
@@ -1533,7 +1584,7 @@ MX_EXPORT(void)
     moddict = PyModule_GetDict(module);
     if (moddict == NULL)
 	goto onError;
-    insstr(moddict,"__version__",VERSION);
+    insstr(moddict,"__version__",MXPROXY_VERSION);
 
     /* Errors */
     if (!(mxProxy_AccessError = insexc(moddict,
@@ -1556,6 +1607,9 @@ MX_EXPORT(void)
     DPRINTF("initmxProxy: module dict refcnt = %i\n",
 	    moddict->ob_refcnt);
 
+    /* We are now initialized */
+    mxProxy_Initialized = 1;
+
  onError:
     /* Check for errors and report them */
     if (PyErr_Occurred())
