@@ -2,7 +2,7 @@
   mxQueue -- A queue implementation
 
   Copyright (c) 1999-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author (mailto:mal@lemburg.com).
 
@@ -20,10 +20,13 @@
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXQUEUE
 
+/* Mark the module as Py_ssize_t clean. */
+#define PY_SSIZE_T_CLEAN 1
+
 #include "mx.h"
 #include "mxQueue.h"
 
-#define VERSION "2.0.3"
+#define MXQUEUE_VERSION "3.0.0"
 
 /* The minimal size of queues; must be > 1 */
 #define MINIMAL_QUEUESIZE	4
@@ -41,10 +44,10 @@
 
 static char *Module_docstring = 
 
- MXQUEUE_MODULE" -- A queue implementation. Version "VERSION"\n\n"
+ MXQUEUE_MODULE" -- A queue implementation. Version "MXQUEUE_VERSION"\n\n"
 
  "Copyright (c) 1999-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com\n"
- "Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
+ "Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -54,6 +57,11 @@ static char *Module_docstring =
 
 static PyObject *mxQueue_Error;			/* Error Exception
 						   object */
+static PyObject *mxQueue_EmptyError;		/* EmptyError Exception
+						   object */
+
+/* Flag telling us whether the module was initialized or not. */
+static int mxQueue_Initialized = 0;
 
 /* --- forward declarations ----------------------------------------------- */
 
@@ -161,7 +169,7 @@ PyObject *notimplemented3(PyObject *u, PyObject *v, PyObject *w)
 /* --- allocation --- */
 
 static
-mxQueueObject *mxQueue_New(int size)
+mxQueueObject *mxQueue_New(Py_ssize_t size)
 {
     mxQueueObject *queue;
     PyObject **w;
@@ -181,8 +189,8 @@ mxQueueObject *mxQueue_New(int size)
     queue->array = w;
     queue->size = size;
     queue->tail = queue->head = size - 1;
-    DPRINTF("Created Queue object at 0x%0x with size %i\n",
-	    (int)queue,size);
+    DPRINTF("Created Queue object at 0x%0lx with size %i\n",
+	    (long)queue,size);
 
     return queue;
 }
@@ -193,8 +201,8 @@ static
 void mxQueue_Free(register mxQueueObject *queue)
 {
     if (queue->array) {
-	register int i;
-	int head = queue->head,
+	register Py_ssize_t i;
+	Py_ssize_t head = queue->head,
 	    size = queue->size;
 	
 	for (i = queue->tail; i != head; i = (i+1) % size)
@@ -209,19 +217,25 @@ void mxQueue_Free(register mxQueueObject *queue)
 /* --- API functions --- */
 
 static
-int mxQueue_Length(register mxQueueObject *queue)
+Py_ssize_t _mxQueue_Length(register mxQueueObject *queue)
 {
-    register int len = queue->head - queue->tail;
+    register Py_ssize_t len = queue->head - queue->tail;
     if (len < 0)
 	len += queue->size;
     return len;
 }
 
+static
+Py_ssize_t mxQueue_Length(register PyObject *queue)
+{
+    return _mxQueue_Length((mxQueueObject *)queue);
+}
+
 static
 int mxQueue_Push(register mxQueueObject *queue,
 		 PyObject *v)
 {
-    int tail,size = queue->size;
+    Py_ssize_t tail,size = queue->size;
     
     if (queue == NULL) {
 	PyErr_BadInternalCall();
@@ -236,7 +250,7 @@ int mxQueue_Push(register mxQueueObject *queue,
     /* Grow the queue first, if we have touching ends */
     if (queue->head == tail) {
 	PyObject **w;
-	int oldsize,oldtail;
+	Py_ssize_t oldsize,oldtail;
 
 	oldsize = size;
 	oldtail = queue->tail;
@@ -276,9 +290,9 @@ static
 int mxQueue_PushMany(register mxQueueObject *queue,
 		     PyObject *seq)
 {
-    int tail;
-    register int i;
-    int length;
+    Py_ssize_t tail;
+    register Py_ssize_t i;
+    Py_ssize_t length;
 
     if (queue == NULL) {
 	PyErr_BadInternalCall();
@@ -294,7 +308,7 @@ int mxQueue_PushMany(register mxQueueObject *queue,
     /* Grow the queue, if we hit the array boundary */
     if (tail + length >= queue->size) {
 	register PyObject **w;
-	register int size;
+	register Py_ssize_t size;
 
 	size = queue->size;
 	while (tail + length >= size)
@@ -346,14 +360,14 @@ static
 PyObject *mxQueue_Pop(register mxQueueObject *queue)
 {
     PyObject *v;
-    int head = queue->head;
+    Py_ssize_t head = queue->head;
     
     if (queue == NULL) {
 	PyErr_BadInternalCall();
 	goto onError;
     }
     Py_Assert(head != queue->tail,
-	      mxQueue_Error,
+	      mxQueue_EmptyError,
 	      "queue is empty");
 
     head--;
@@ -373,10 +387,10 @@ PyObject *mxQueue_Pop(register mxQueueObject *queue)
 
 static
 PyObject *mxQueue_PopMany(register mxQueueObject *queue,
-			  int n)
+			  Py_ssize_t n)
 {
     PyObject *t;
-    register int i;
+    register Py_ssize_t i;
     
     if (queue == NULL) {
 	PyErr_BadInternalCall();
@@ -397,8 +411,8 @@ PyObject *mxQueue_PopMany(register mxQueueObject *queue,
 static
 int mxQueue_Clear(register mxQueueObject *queue)
 {
-    register int i;
-    int size = queue->size,
+    register Py_ssize_t i;
+    Py_ssize_t size = queue->size,
 	head = queue->head;
 	
     if (head != queue->tail)
@@ -438,7 +452,7 @@ static
 PyObject *mxQueue_AsTuple(register mxQueueObject *queue)
 {
     PyObject *t = 0;
-    int i,len;
+    Py_ssize_t i,len;
     
     if (queue == NULL) {
 	PyErr_BadInternalCall();
@@ -467,7 +481,7 @@ static
 PyObject *mxQueue_AsList(register mxQueueObject *queue)
 {
     PyObject *l = 0;
-    int i,len;
+    Py_ssize_t i,len;
     
     if (queue == NULL) {
 	PyErr_BadInternalCall();
@@ -496,7 +510,7 @@ PyObject *mxQueue_AsList(register mxQueueObject *queue)
 #if 0
 static
 int mxQueue_Resize(register mxQueueObject *self,
-		   int size)
+		   Py_ssize_t size)
 {
     register PyObject **w;
 
@@ -534,9 +548,9 @@ Py_C_Function( mxQueue_new,
 	       "Queue([intialsize])")
 {
     PyObject *v;
-    int size = 0;
+    Py_ssize_t size = 0;
 
-    Py_GetArg("|i",size);
+    Py_GetArg("|"Py_SSIZE_T_PARSERMARKER, size);
 
     v = (PyObject *)mxQueue_New(size);
     if (v == NULL)
@@ -600,9 +614,9 @@ Py_C_Function( mxQueue_pop_many,
 	       "pop_many(n)")
 {
     register PyObject *v;
-    int n;
+    Py_ssize_t n;
 
-    Py_GetSingleArg("i",n);
+    Py_GetSingleArg(Py_SSIZE_T_PARSERMARKER, n);
     
     v = mxQueue_PopMany(queue,n);
     if (v == NULL)
@@ -663,9 +677,9 @@ Py_C_Function( mxQueue_as_list,
 Py_C_Function( mxQueue_resize,
 	       "resize([size=len(queue)])")
 {
-    int size = mxQueue_Length(queue);
+    Py_ssize_t size = mxQueue_Length(queue);
 
-    Py_GetArg("|i",size);
+    Py_GetArg("|"Py_SSIZE_T_PARSERMARKER,size);
     
     if (mxQueue_Resize(queue,size))
 	goto onError;
@@ -682,11 +696,12 @@ Py_C_Function( mxQueue_resize,
 /* --- slots --- */
 
 static
-int mxQueue_Print(mxQueueObject *self,
+int mxQueue_Print(PyObject *obj,
 		  FILE *fp, 
 		  int flags)
 {
-    int i,
+    mxQueueObject *self = (mxQueueObject *)obj;
+    Py_ssize_t i,
 	head = self->head,
 	tail = self->tail,
 	size = self->size;
@@ -714,8 +729,9 @@ int mxQueue_Print(mxQueueObject *self,
 }
 
 static
-PyObject *mxQueue_Repr(mxQueueObject *self)
+PyObject *mxQueue_Repr(PyObject *obj)
 {
+    mxQueueObject *self = (mxQueueObject *)obj;
     char s[256];
 
 #if 1
@@ -730,13 +746,15 @@ PyObject *mxQueue_Repr(mxQueueObject *self)
 }
 
 static
-int mxQueue_Compare(mxQueueObject *v,
-		    mxQueueObject *w)
+int mxQueue_Compare(PyObject *left,
+		    PyObject *right)
 {
-    int v_len = mxQueue_Length(v);
-    int w_len = mxQueue_Length(w);
-    int len = min(v_len,w_len);
-    int i,j,k;
+    mxQueueObject *v = (mxQueueObject *)left;
+    mxQueueObject *w = (mxQueueObject *)right;
+    Py_ssize_t v_len = _mxQueue_Length(v);
+    Py_ssize_t w_len = _mxQueue_Length(w);
+    Py_ssize_t len = min(v_len,w_len);
+    Py_ssize_t i,j,k;
 
     for (i = v->tail, j = w->tail, k = 0; 
 	 k < len; 
@@ -749,16 +767,17 @@ int mxQueue_Compare(mxQueueObject *v,
 }
 
 static
-PyObject *mxQueue_Getattr(mxQueueObject *self,
+PyObject *mxQueue_Getattr(PyObject *obj,
 			  char *name)
 {
-    return Py_FindMethod(mxQueue_Methods,
-			 (PyObject *)self,name);
+    return Py_FindMethod(mxQueue_Methods, obj, name);
 }
 
 static
-int mxQueue_NonZero(mxQueueObject *self)
+int mxQueue_NonZero(PyObject *obj)
 {
+    mxQueueObject *self = (mxQueueObject *)obj;
+
     return self->head != self->tail;
 }
 
@@ -792,26 +811,29 @@ int mxQueue_Coerce(PyObject **pv,
 }
 
 static
-PyObject *mxQueue_LeftShift(mxQueueObject *v,
-			    PyObject *w)
+PyObject *mxQueue_LeftShift(PyObject *left,
+			    PyObject *right)
 {
+    mxQueueObject *self = (mxQueueObject *)left;
+
     /* Queue << Object -> (Queue.push(Object),Queue) */
-    mxQueue_Push(v,w);
-    Py_INCREF(v);
-    return (PyObject *)v;
+    mxQueue_Push(self, right);
+    Py_INCREF(self);
+    return left;
 }
 
 static
-PyObject *mxQueue_RightShift(mxQueueObject *v,
-			     PyObject *w)
+PyObject *mxQueue_RightShift(PyObject *left,
+			     PyObject *right)
 {
+    mxQueueObject *self = (mxQueueObject *)left;
     long n;
     
     /* Queue >> Integer -> Queue.pop_many(Integer) */
-    if (!PyInt_Check(w))
+    if (!PyInt_Check(right))
 	Py_Error(PyExc_TypeError,
 		 "right side of >> must an integer");
-    n = PyInt_AS_LONG(w);
+    n = PyInt_AS_LONG(right);
 
     /* XXX Doesn't work yet... */
     n = 1;
@@ -820,11 +842,11 @@ PyObject *mxQueue_RightShift(mxQueueObject *v,
 	      PyExc_TypeError,
 	      "you can only pop 1 or more entries");
     if (n == 1)
-	return mxQueue_Pop(v);
+	return mxQueue_Pop(self);
     else
-	return mxQueue_PopMany(v,n);
+	return mxQueue_PopMany(self, n);
 #else
-    return mxQueue_Pop(v);
+    return mxQueue_Pop(self);
 #endif
 
  onError:
@@ -835,44 +857,44 @@ PyObject *mxQueue_RightShift(mxQueueObject *v,
 
 static
 PySequenceMethods mxQueue_TypeAsSequence = {
-    (inquiry)mxQueue_Length,		/*sq_length*/
-    (binaryfunc)0,			/*sq_concat*/
-    (intargfunc)0,			/*sq_repeat*/
-    (intargfunc)0,			/*sq_item*/
-    (intintargfunc)0,			/*sq_slice*/
-    (intobjargproc)0,			/*sq_ass_item*/
-    (intintobjargproc)0,		/*sq_ass_slice*/
+    mxQueue_Length,			/*sq_length*/
+    0,					/*sq_concat*/
+    0,					/*sq_repeat*/
+    0,					/*sq_item*/
+    0,					/*sq_slice*/
+    0,					/*sq_ass_item*/
+    0,					/*sq_ass_slice*/
 };
 
 static
 PyNumberMethods mxQueue_TypeAsNumber = {
 
     /* These slots are not NULL-checked, so we must provide dummy functions */
-    (binaryfunc)notimplemented2,		/*nb_add*/
-    (binaryfunc)notimplemented2,		/*nb_subtract*/
-    (binaryfunc)notimplemented2,		/*nb_multiply*/
-    (binaryfunc)notimplemented2,		/*nb_divide*/
-    (binaryfunc)notimplemented2,		/*nb_remainder*/
-    (binaryfunc)notimplemented2,		/*nb_divmod*/
-    (ternaryfunc)notimplemented3,		/*nb_power*/
-    (unaryfunc)notimplemented1,			/*nb_negative*/
-    (unaryfunc)notimplemented1,			/*nb_positive*/
+    notimplemented2,			/*nb_add*/
+    notimplemented2,			/*nb_subtract*/
+    notimplemented2,			/*nb_multiply*/
+    notimplemented2,			/*nb_divide*/
+    notimplemented2,			/*nb_remainder*/
+    notimplemented2,			/*nb_divmod*/
+    notimplemented3,			/*nb_power*/
+    notimplemented1,			/*nb_negative*/
+    notimplemented1,			/*nb_positive*/
 
     /* Everything below this line EXCEPT nb_nonzero (!) is NULL checked */
-    (unaryfunc)0,				/*nb_absolute*/
-    (inquiry)mxQueue_NonZero,			/*nb_nonzero*/
-    (unaryfunc)0,				/*nb_invert*/
-    (binaryfunc)mxQueue_LeftShift,		/*nb_lshift*/
-    (binaryfunc)mxQueue_RightShift,		/*nb_rshift*/
-    (binaryfunc)notimplemented2,		/*nb_and*/
-    (binaryfunc)notimplemented2,		/*nb_xor*/
-    (binaryfunc)notimplemented2,		/*nb_or*/
-    (coercion)mxQueue_Coerce,			/*nb_coerce*/
-    (unaryfunc)0,				/*nb_int*/
-    (unaryfunc)0,				/*nb_long*/
-    (unaryfunc)0,				/*nb_float*/
-    (unaryfunc)0,				/*nb_oct*/
-    (unaryfunc)0,				/*nb_hex*/
+    0,					/*nb_absolute*/
+    mxQueue_NonZero,			/*nb_nonzero*/
+    0,					/*nb_invert*/
+    mxQueue_LeftShift,			/*nb_lshift*/
+    mxQueue_RightShift,			/*nb_rshift*/
+    notimplemented2,			/*nb_and*/
+    notimplemented2,			/*nb_xor*/
+    notimplemented2,			/*nb_or*/
+    mxQueue_Coerce,			/*nb_coerce*/
+    0,					/*nb_int*/
+    0,					/*nb_long*/
+    0,					/*nb_float*/
+    0,					/*nb_oct*/
+    0,					/*nb_hex*/
 };
 
 statichere
@@ -884,22 +906,22 @@ PyTypeObject mxQueue_Type = {
     0,			  		/*tp_itemsize*/
     /* slots */
     (destructor)mxQueue_Free,		/*tp_dealloc*/
-    (printfunc)mxQueue_Print,	  	/*tp_print*/
-    (getattrfunc)mxQueue_Getattr,  	/*tp_getattr*/
-    (setattrfunc)0,		  	/*tp_setattr*/
-    (cmpfunc)mxQueue_Compare,  		/*tp_compare*/
-    (reprfunc)mxQueue_Repr,	  	/*tp_repr*/
+    mxQueue_Print,	  	       	/*tp_print*/
+    mxQueue_Getattr,  			/*tp_getattr*/
+    0,		  			/*tp_setattr*/
+    mxQueue_Compare,  		       	/*tp_compare*/
+    mxQueue_Repr,	  	       	/*tp_repr*/
     &mxQueue_TypeAsNumber, 		/*tp_as_number*/
     &mxQueue_TypeAsSequence,		/*tp_as_sequence*/
     0,					/*tp_as_mapping*/
-    (hashfunc)0,			/*tp_hash*/
-    (ternaryfunc)0,			/*tp_call*/
-    (reprfunc)0,			/*tp_str*/
-    (getattrofunc)0, 			/*tp_getattro*/
-    (setattrofunc)0, 			/*tp_setattro*/
+    0,					/*tp_hash*/
+    0,					/*tp_call*/
+    0,					/*tp_str*/
+    0, 					/*tp_getattro*/
+    0, 					/*tp_setattro*/
     0,					/*tp_as_buffer*/
     0,					/*tp_xxx4*/
-    (char*) 0				/*tp_doc*/
+    0					/*tp_doc*/
 };
 
 /* Python Method Table */
@@ -966,7 +988,7 @@ mxQueueModule_APIObject mxQueueModuleAPI =
     mxQueue_Push,
     mxQueue_Pop,
     mxQueue_Clear,
-    mxQueue_Length
+    _mxQueue_Length
 #if 0
     ,mxQueue_FromSequence,
     mxQueue_AsTuple,
@@ -976,12 +998,24 @@ mxQueueModule_APIObject mxQueueModuleAPI =
 #endif
 };
 
+/* Cleanup function */
+static 
+void mxQueueModule_Cleanup(void)
+{
+    /* Reset mxQueue_Initialized flag */
+    mxQueue_Initialized = 0;
+}
+
 /* create PyMethodObjects and register them in the module's dict */
 MX_EXPORT(void) 
      initmxQueue(void)
 {
     PyObject *module, *moddict, *api;
 
+    if (mxQueue_Initialized)
+	Py_Error(PyExc_SystemError,
+		 "can't initialize "MXQUEUE_MODULE" more than once");
+
     /* Init type objects */
     PyType_Init(mxQueue_Type);
 
@@ -994,15 +1028,23 @@ MX_EXPORT(void)
     if (module == NULL)
 	goto onError;
 
+    /* Register cleanup function */
+    if (Py_AtExit(mxQueueModule_Cleanup)) {
+	/* XXX what to do if we can't register that function ??? */
+	DPRINTF("* Failed to register mxQueue cleanup function\n");
+    }
+
     /* Add some constants to the module's dict */
     moddict = PyModule_GetDict(module);
     PyDict_SetItemString(moddict, 
 			 "__version__",
-			 PyString_FromString(VERSION));
+			 PyString_FromString(MXQUEUE_VERSION));
 
     /* Errors */
     if (!(mxQueue_Error = insexc(moddict,"Error",PyExc_IndexError)))
 	goto onError;
+    if (!(mxQueue_EmptyError = insexc(moddict,"EmptyError",mxQueue_Error)))
+	goto onError;
 
     /* Type objects */
     Py_INCREF(&mxQueue_Type);
@@ -1016,6 +1058,9 @@ MX_EXPORT(void)
     PyDict_SetItemString(moddict,MXQUEUE_MODULE"API",api);
     Py_DECREF(api);
 
+    /* We are now initialized */
+    mxQueue_Initialized = 1;
+
  onError:
     /* Check for errors and report them */
     if (PyErr_Occurred())
