@@ -16,14 +16,17 @@
   MAL_MEM_DEBUG -- enable malloc output (new,cnew,free,...) [mxstdlib.h]
 
   Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author.
 
+  XXX Most of these macros and defines should really use an mx prefix to
+      avoid collisions with native Python macros and defines. So far this
+      has not happened, which is why this hasn't been implemented yet.
+
 */
 
-/* Include the generic mx header file */
-#include "mxh.h"
+/* --- Platform or compiler specific tweaks ------------------------------- */
 
 /* Add some platform specific symbols to enable work-arounds for the
    static forward declaration of type definitions; note that the GNU C
@@ -32,27 +35,57 @@
    Many thanks to all who have contributed to this list.
 
 */
-#if (!defined(__GNUC__))
-# if (defined(NeXT) || defined(sgi) || defined(_AIX) || (defined(__osf__) && defined(__DECC)) || defined(TrueComaq64) || defined(__VMS))
-#  define BAD_STATIC_FORWARD
+#ifndef BAD_STATIC_FORWARD
+# if (!defined(__GNUC__))
+#  if (defined(NeXT) || defined(sgi) || defined(_AIX) || (defined(__osf__) && defined(__DECC)) || defined(TrueCompaq64) || defined(__VMS))
+#   define BAD_STATIC_FORWARD
+#  endif
 # endif
 #endif
 
-/* Some more tweaks for various platforms. Thanks to Jean-Franois
-   PIRONNE. */
+/* Some more tweaks for various platforms. */
+
+/* VMS needs this define. Thanks to Jean-Franois PIRONNE */
 #if defined(__VMS)
 # define __SC__
 #endif
 
+/* xlC on AIX doesn't like the Python work-around for static forwards
+   in ANSI mode (default), so we switch on extended mode. Thanks to
+   Albert Chin-A-Young */
+#if defined(__xlC__)
+# pragma langlvl extended
+#endif
+
+/* --- Standard header files ---------------------------------------------- */
+
+/* Include the generic mx header file */
+#include "mxh.h"
+
 /* Include nearly all Python symbols & definitions */
 #include "Python.h"
 
 /* Include other standard stuff */
 #include "mxstdlib.h"
 
-/* Include backward compatibility stuff */
+/* Include Python backward compatibility stuff */
 #include "mxpyapi.h"
 
+/* --- Compiler support --------------------------------------------------- */
+
+/* Support for compilers which don't like static forward declarations
+   of arrays; Python 2.3 removed the support for BAD_STATIC_FORWARD
+   which is why we now use our own little helpers here. */
+#undef staticforward
+#undef statichere
+#ifdef BAD_STATIC_FORWARD
+# define staticforward extern
+# define statichere static
+#else
+# define staticforward static
+# define statichere static
+#endif 
+
 /* --- Declare macros ----------------------------------------------------- */
 
 #define Py_NONE (Py_INCREF(Py_None),Py_None)
@@ -82,10 +115,10 @@
 
 #ifdef MAL_REF_DEBUG
 
-static void _Py_INCREF(PyObject *v,
-		       char *name,
-		       char *filename,
-		       int lineno)
+static void mx_Py_INCREF(PyObject *v,
+			 const char *name,
+			 const char *filename,
+			 int lineno)
 {
     if (!Py_DebugFlag) {
 	Py_XINCREF(v);
@@ -103,10 +136,10 @@ static void _Py_INCREF(PyObject *v,
     }
 }
 
-static void _Py_DECREF(PyObject *v,
-		       char *name,
-		       char *filename,
-		       int lineno) 
+static void mx_Py_DECREF(PyObject *v,
+			 const char *name,
+			 const char *filename,
+			 int lineno) 
 {
     if (!Py_DebugFlag) {
 	Py_XDECREF(v);
@@ -130,10 +163,10 @@ static void _Py_DECREF(PyObject *v,
     }
 }
 
-static void _Py_PRINT_REFCOUNT(PyObject *v,
-			       char *name,
-			       char *filename,
-			       int lineno) 
+static void mx_Py_PRINT_REFCOUNT(PyObject *v,
+				 const char *name,
+				 const char *filename,
+				 int lineno) 
 {
     if (!v)
 	mxDebugPrintf("[%s:%5i] Py_PRINT_REFCOUNT( %-8s == NULL );\n",
@@ -146,15 +179,15 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 }
 
 # undef Py_INCREF
-# define Py_INCREF(x) _Py_INCREF((PyObject *)x,#x,__FILE__,__LINE__)
+# define Py_INCREF(x) mx_Py_INCREF((PyObject *)x,#x,__FILE__,__LINE__)
 # undef Py_DECREF
-# define Py_DECREF(x) _Py_DECREF((PyObject *)x,#x,__FILE__,__LINE__)
+# define Py_DECREF(x) mx_Py_DECREF((PyObject *)x,#x,__FILE__,__LINE__)
 # undef Py_XINCREF
-# define Py_XINCREF(x) _Py_INCREF((PyObject *)x,#x,__FILE__,__LINE__)
+# define Py_XINCREF(x) mx_Py_INCREF((PyObject *)x,#x,__FILE__,__LINE__)
 # undef Py_XDECREF
-# define Py_XDECREF(x) _Py_DECREF((PyObject *)x,#x,__FILE__,__LINE__)
-# define Py_DELETE(x) {if (x->ob_refcnt > 1) mxDebugPrintf("[%s:%5i] Py_DELETE( "#x" ) WARNING: Refcount = %i > 0\n",__FILE__,__LINE__,(int)x->ob_refcnt-1);Py_DECREF(x);}
-# define Py_PRINT_REFCOUNT(x) _Py_PRINT_REFCOUNT((PyObject *)x,#x,__FILE__,__LINE__)
+# define Py_XDECREF(x) mx_Py_DECREF((PyObject *)x,#x,__FILE__,__LINE__)
+# define Py_DELETE(x) {if (x->ob_refcnt > 1) mxDebugPrintf("[%s:%5i] Py_DELETE( "#x" ) WARNING: Refcount = %i > 1\n",__FILE__,__LINE__,(int)x->ob_refcnt);Py_DECREF(x);}
+# define Py_PRINT_REFCOUNT(x) mx_Py_PRINT_REFCOUNT((PyObject *)x,#x,__FILE__,__LINE__)
 #else
 # define Py_DELETE(x) Py_DECREF(x)
 # define Py_PRINT_REFCOUNT(x) 
@@ -162,11 +195,28 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 
 #define Py_DEC_REF(x) {Py_XDECREF(x); x=0;} /* doing this once too often doesn't hurt */
 
+/* Unreference a Python object. This is only used in Python debug
+   builds and needed to keep track of all allocated references. Use in
+   object constructors or free list implementations.  */
+#ifndef _Py_DEC_REFTOTAL
+# ifdef Py_REF_DEBUG
+#  define _Py_DEC_REFTOTAL _Py_RefTotal--
+# else
+#  define _Py_DEC_REFTOTAL
+# endif
+#endif
+#define mxPy_UNREF(x) _Py_DEC_REFTOTAL
+
 /* --- Argument passing and checking -------------------------------------- */
 
 /* No arguments expected; also use Py_MethodListEntryNoArgs() for this
-   kind of fct */
-#define Py_NoArgsCheck() {if (!PyArg_NoArgs(args)) goto onError;}
+   kind of fct; this check is no longer needed in Python 2.3 and
+   later */
+#if PY_VERSION_HEX >= 0x02030000
+# define Py_NoArgsCheck() {if (0) goto onError;}
+#else
+# define Py_NoArgsCheck() {if (!PyArg_NoArgs(args)) goto onError;}
+#endif
 
 /* For functions with old style args (Py_MethodListEntrySingleArg) */
 #define Py_GetArgObject(a) {a = args; if (!a) {PyErr_SetString(PyExc_TypeError,"function/method requires an argument"); goto onError;}}
@@ -182,10 +232,13 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 #define Py_Get7Args(format,a1,a2,a3,a4,a5,a6,a7) {if (!PyArg_ParseTuple(args,format,&a1,&a2,&a3,&a4,&a5,&a6,&a7)) goto onError;}
 #define Py_Get8Args(format,a1,a2,a3,a4,a5,a6,a7,a8) {if (!PyArg_ParseTuple(args,format,&a1,&a2,&a3,&a4,&a5,&a6,&a7,&a8)) goto onError;}
 
+/* The keywords array string type */
+#define Py_KEYWORDS_STRING_TYPE char
+
 /* For functions with keywords -- the first macro parameter must be
    the keywords array given as e.g.
 
-   static char *keywords[] = {"first","second","third", 0};
+   static const char *keywords[] = {"first","second","third", 0};
 
    with an entry for every argument (in the correct order). The
    functions must be included in the method list using
@@ -193,14 +246,14 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
    Py_C_Function_WithKeywords().
 
 */
-#define Py_KeywordGetArg(keywords,format,a1) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1)) goto onError;}
-#define Py_KeywordGet2Args(keywords,format,a1,a2) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2)) goto onError;}
-#define Py_KeywordGet3Args(keywords,format,a1,a2,a3) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3)) goto onError;}
-#define Py_KeywordGet4Args(keywords,format,a1,a2,a3,a4) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3,&a4)) goto onError;}
-#define Py_KeywordGet5Args(keywords,format,a1,a2,a3,a4,a5) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3,&a4,&a5)) goto onError;}
-#define Py_KeywordGet6Args(keywords,format,a1,a2,a3,a4,a5,a6) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6)) goto onError;}
-#define Py_KeywordGet7Args(keywords,format,a1,a2,a3,a4,a5,a6,a7) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6,&a7)) goto onError;}
-#define Py_KeywordGet8Args(keywords,format,a1,a2,a3,a4,a5,a6,a7,a8) {if (!PyArg_ParseTupleAndKeywords(args,kw,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6,&a7,&a8)) goto onError;}
+#define Py_KeywordGetArg(keywords,format,a1) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1)) goto onError;}
+#define Py_KeywordGet2Args(keywords,format,a1,a2) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2)) goto onError;}
+#define Py_KeywordGet3Args(keywords,format,a1,a2,a3) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3)) goto onError;}
+#define Py_KeywordGet4Args(keywords,format,a1,a2,a3,a4) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3,&a4)) goto onError;}
+#define Py_KeywordGet5Args(keywords,format,a1,a2,a3,a4,a5) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3,&a4,&a5)) goto onError;}
+#define Py_KeywordGet6Args(keywords,format,a1,a2,a3,a4,a5,a6) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6)) goto onError;}
+#define Py_KeywordGet7Args(keywords,format,a1,a2,a3,a4,a5,a6,a7) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6,&a7)) goto onError;}
+#define Py_KeywordGet8Args(keywords,format,a1,a2,a3,a4,a5,a6,a7,a8) {if (!PyArg_ParseTupleAndKeywords(args,kws,format,keywords,&a1,&a2,&a3,&a4,&a5,&a6,&a7,&a8)) goto onError;}
 
 /* New style macros fof functions supporting keywords -- the C
    variable names are used as template for the keyword list, i.e. they
@@ -239,14 +292,14 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
     }
 
 */
-#define Py_KeywordsGetArg(format,a1) {static char *kwlist[] = {#a1,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1)) goto onError;}
-#define Py_KeywordsGet2Args(format,a1,a2) {static char *kwlist[] = {#a1,#a2,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2)) goto onError;}
-#define Py_KeywordsGet3Args(format,a1,a2,a3) {static char *kwlist[] = {#a1,#a2,#a3,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3)) goto onError;}
-#define Py_KeywordsGet4Args(format,a1,a2,a3,a4) {static char *kwlist[] = {#a1,#a2,#a3,#a4,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3,&a4)) goto onError;}
-#define Py_KeywordsGet5Args(format,a1,a2,a3,a4,a5) {static char *kwlist[] = {#a1,#a2,#a3,#a4,#a5,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3,&a4,&a5)) goto onError;}
-#define Py_KeywordsGet6Args(format,a1,a2,a3,a4,a5,a6) {static char *kwlist[] = {#a1,#a2,#a3,#a4,#a5,#a6,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3,&a4,&a5,&a6)) goto onError;}
-#define Py_KeywordsGet7Args(format,a1,a2,a3,a4,a5,a6,a7) {static char *kwlist[] = {#a1,#a2,#a3,#a4,#a5,#a6,#a7,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3,&a4,&a5,&a6,&a7)) goto onError;}
-#define Py_KeywordsGet8Args(format,a1,a2,a3,a4,a5,a6,a7,a8) {static char *kwlist[] = {#a1,#a2,#a3,#a4,#a5,#a6,#a7,#a8,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kw,format,kwlist,&a1,&a2,&a3,&a4,&a5,&a6,&a7,&a8)) goto onError;}
+#define Py_KeywordsGetArg(format,a1) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1)) goto onError;}
+#define Py_KeywordsGet2Args(format,a1,a2) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2)) goto onError;}
+#define Py_KeywordsGet3Args(format,a1,a2,a3) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3)) goto onError;}
+#define Py_KeywordsGet4Args(format,a1,a2,a3,a4) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,#a4,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3,&a4)) goto onError;}
+#define Py_KeywordsGet5Args(format,a1,a2,a3,a4,a5) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,#a4,#a5,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3,&a4,&a5)) goto onError;}
+#define Py_KeywordsGet6Args(format,a1,a2,a3,a4,a5,a6) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,#a4,#a5,#a6,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3,&a4,&a5,&a6)) goto onError;}
+#define Py_KeywordsGet7Args(format,a1,a2,a3,a4,a5,a6,a7) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,#a4,#a5,#a6,#a7,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3,&a4,&a5,&a6,&a7)) goto onError;}
+#define Py_KeywordsGet8Args(format,a1,a2,a3,a4,a5,a6,a7,a8) {static Py_KEYWORDS_STRING_TYPE *kwslist[] = {#a1,#a2,#a3,#a4,#a5,#a6,#a7,#a8,NULL}; if (!PyArg_ParseTupleAndKeywords(args,kws,format,kwslist,&a1,&a2,&a3,&a4,&a5,&a6,&a7,&a8)) goto onError;}
 
 /* --- Returning values to Python ----------------------------------------- */
 
@@ -278,20 +331,27 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 
 /* --- Declaring Python builtin functions/methods ------------------------- */
 
+/* The doc-string type changed to const char* in Python 2.5 */
+#if PY_VERSION_HEX >= 0x02050000
+# define Py_DOC_STRING_TYPE const char
+#else
+# define Py_DOC_STRING_TYPE char
+#endif
+
 /* Declare C function/method fct, having docstring docstr; may use vargargs */
 #define Py_C_Function(fct,docstr) \
-        static char fct##_docstring[] = docstr;\
+        static Py_DOC_STRING_TYPE fct##_docstring[] = docstr;\
         static PyObject *fct(PyObject *self, PyObject *args)
 
 /* Declare C function/method fct, having keywords keywordsarray and a
    docstring docstr; may use vargargs & keywords */
 #define Py_C_Function_WithKeywords(fct,docstr) \
-        static char fct##_docstring[] = docstr;\
-        static PyObject *fct(PyObject *self, PyObject *args, PyObject *kw)
+        static Py_DOC_STRING_TYPE fct##_docstring[] = docstr;\
+        static PyObject *fct(PyObject *self, PyObject *args, PyObject *kws)
 
 /* These declare: self -- instance pointer for methods, NULL for functions
                   args -- argument tuple
-		  kw   -- keywords dict (if applicable)
+		  kws  -- keywords dict (if applicable)
    plus as statics:
                   <function name>_docstring -- the docstring as given
 		  <function name>_keywords  -- the keyword array as given
@@ -311,8 +371,14 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 #define Py_MethodListEntry(pyname,cname) {pyname,(PyCFunction)cname,METH_VARARGS,cname##_docstring}
 
 /* Add a C function/method cname to the module dict as pyname; the
-   function takes no args */
-#define Py_MethodListEntryNoArgs(pyname,cname) {pyname,(PyCFunction)cname,0,cname##_docstring}
+   function takes no args; in Python 2.3 a new flag was added for
+   these which implements the no args check in the interpreter
+   itself. */
+#ifdef METH_NOARGS
+# define Py_MethodListEntryNoArgs(pyname,cname) {pyname,(PyCFunction)cname,METH_NOARGS,cname##_docstring}
+#else
+# define Py_MethodListEntryNoArgs(pyname,cname) {pyname,(PyCFunction)cname,0,cname##_docstring}
+#endif
 
 /* Add a C function/method cname to the module dict as pyname; the
    function takes one argument: the object is passed in directly
@@ -345,15 +411,34 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 		start = stop;			\
 	}
 
+/* --- Number macros ------------------------------------------------------ */
+
+/* Test for PyFloat_AsDouble() compatible object */
+#define PyFloat_Compatible(obj)					\
+        (PyInstance_Check(obj) 					\
+	 ? PyObject_HasAttrString(obj, "__float__")		\
+	 : (obj->ob_type->tp_as_number->nb_float != NULL))
+
 /* --- Text macros -------------------------------------------------------- */
 
-/* Check a given text slice and apply the usual rules for negative indices */
+/* Check a given text slice and apply the usual rules for negative
+   indices */
 #define Py_CheckBufferSlice(textlen,start,stop) \
         Py_CheckSequenceSlice(textlen,start,stop)
 
 /* Dito for string objects */
+#define Py_CheckStringSlice(textobj,start,stop) \
+        Py_CheckSequenceSlice(PyString_GET_SIZE(textobj),start,stop)
+
+/* For b/w compatibility */
 #define Py_CheckSlice(textobj,start,stop) \
-        Py_CheckBufferSlice(PyString_GET_SIZE(textobj),start,stop)
+        Py_CheckStringSlice(textobj,start,stop)
+
+/* Dito for Unicode objects */
+#ifdef PyUnicode_GET_SIZE
+# define Py_CheckUnicodeSlice(unicode,start,stop) \
+         Py_CheckSequenceSlice(PyUnicode_GET_SIZE(unicode),start,stop)
+#endif
 
 /* This assumes that fixed is a constant char array; the strcmp
    function is only called in case the attribute name length exceeds
@@ -433,7 +518,7 @@ static void _Py_PRINT_REFCOUNT(PyObject *v,
 #define PyType_Init(x)						\
 {								\
     x.ob_type = &PyType_Type; 					\
-    Py_Assert(x.tp_basicsize >= sizeof(PyObject),		\
+    Py_Assert(x.tp_basicsize >= (int)sizeof(PyObject),	        \
 	      PyExc_SystemError,				\
 	      "Internal error: tp_basicsize of "#x" too small");\
 }
