@@ -2,7 +2,7 @@
   mxStack -- A stack implementation
 
   Copyright (c) 1997-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author (mailto:mal@lemburg.com).
 
@@ -20,10 +20,13 @@
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXSTACK
 
+/* Mark the module as Py_ssize_t clean. */
+#define PY_SSIZE_T_CLEAN 1
+
 #include "mx.h"
 #include "mxStack.h"
 
-#define VERSION "2.0.3"
+#define MXSTACK_VERSION "3.0.0"
 
 /* The minimal size of stacks; must be > 1 */
 #define MINIMAL_STACKSIZE	4
@@ -41,10 +44,10 @@
 
 static char *Module_docstring = 
 
- MXSTACK_MODULE" -- A stack implementation. Version "VERSION"\n\n"
+ MXSTACK_MODULE" -- A stack implementation. Version "MXSTACK_VERSION"\n\n"
 
  "Copyright (c) 1997-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com\n"
- "Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
+ "Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -54,6 +57,11 @@ static char *Module_docstring =
 
 static PyObject *mxStack_Error;			/* Error Exception
 						   object */
+static PyObject *mxStack_EmptyError;		/* EmptyError Exception
+						   object */
+
+/* Flag telling us whether the module was initialized or not. */
+static int mxStack_Initialized = 0;
 
 /* --- forward declarations ----------------------------------------------- */
 
@@ -161,7 +169,7 @@ PyObject *notimplemented3(PyObject *u, PyObject *v, PyObject *w)
 /* --- allocation --- */
 
 static
-mxStackObject *mxStack_New(int size)
+mxStackObject *mxStack_New(Py_ssize_t size)
 {
     mxStackObject *stack;
     PyObject **w;
@@ -294,7 +302,8 @@ int mxStack_PushMany(register mxStackObject *stack,
 	    if (!v) {
 		/* Rollback */
 		while (i > 0) {
-		    Py_DECREF(stack->array[top--]);
+		    Py_DECREF(stack->array[top]);
+		    top--;
 		    i--;
 		}
 		stack->top = top;
@@ -318,7 +327,7 @@ PyObject *mxStack_Pop(register mxStackObject *stack)
 	goto onError;
     }
     Py_Assert(stack->top >= 0,
-	      mxStack_Error,
+	      mxStack_EmptyError,
 	      "stack is empty");
     return stack->array[stack->top--];
  onError:
@@ -330,10 +339,10 @@ PyObject *mxStack_Pop(register mxStackObject *stack)
 
 static
 PyObject *mxStack_PopMany(register mxStackObject *stack,
-			  int n)
+			  Py_ssize_t n)
 {
     PyObject *t;
-    register int i;
+    register Py_ssize_t i;
     
     if (stack == NULL) {
 	PyErr_BadInternalCall();
@@ -353,7 +362,7 @@ PyObject *mxStack_PopMany(register mxStackObject *stack,
 static
 int mxStack_Clear(register mxStackObject *stack)
 {
-    register int i;
+    register Py_ssize_t i;
 	
     for (i = 0; i <= stack->top; i++)
 	Py_DECREF(stack->array[i]);
@@ -387,7 +396,7 @@ static
 PyObject *mxStack_AsTuple(register mxStackObject *stack)
 {
     PyObject *t = 0;
-    int i,len;
+    Py_ssize_t i,len;
     
     if (stack == NULL) {
 	PyErr_BadInternalCall();
@@ -416,7 +425,7 @@ static
 PyObject *mxStack_AsList(register mxStackObject *stack)
 {
     PyObject *l = 0;
-    int i,len;
+    Py_ssize_t i,len;
     
     if (stack == NULL) {
 	PyErr_BadInternalCall();
@@ -442,14 +451,20 @@ PyObject *mxStack_AsList(register mxStackObject *stack)
 }
 
 static
-int mxStack_Length(register mxStackObject *self)
+Py_ssize_t _mxStack_Length(register mxStackObject *self)
 {
     return self->top + 1;
 }
 
+static
+Py_ssize_t mxStack_Length(register PyObject *self)
+{
+    return _mxStack_Length((mxStackObject *)self);
+}
+
 static
 int mxStack_Resize(register mxStackObject *self,
-		   int size)
+		   Py_ssize_t size)
 {
     register PyObject **w;
 
@@ -479,11 +494,11 @@ int mxStack_Resize(register mxStackObject *self,
 }
 
 static
-PyObject *mxStack_GetItem(register mxStackObject *self,
-			  int index)
+PyObject *_mxStack_GetItem(register mxStackObject *self,
+			   Py_ssize_t index)
 {
     PyObject *v;
-    int len = self->top + 1;
+    Py_ssize_t len = self->top + 1;
 
     if (index < 0)
 	index += len;
@@ -498,6 +513,14 @@ PyObject *mxStack_GetItem(register mxStackObject *self,
     return NULL;
 }
 
+static
+PyObject *mxStack_GetItem(register PyObject *obj,
+			  Py_ssize_t index)
+{
+    return _mxStack_GetItem((mxStackObject *)obj, index);
+}
+
+
 /* --- methods --- (should have lowercase extension) */
 
 #define stack ((mxStackObject*)self)
@@ -506,9 +529,9 @@ Py_C_Function( mxStack_new,
 	       "Stack([intialsize])")
 {
     PyObject *v;
-    int size = 0;
+    Py_ssize_t size = 0;
 
-    Py_GetArg("|i",size);
+    Py_GetArg("|"Py_SSIZE_T_PARSERMARKER, size);
 
     v = (PyObject *)mxStack_New(size);
     if (v == NULL)
@@ -569,9 +592,9 @@ Py_C_Function( mxStack_pop_many,
 	       "pop_many(n)")
 {
     register PyObject *v;
-    int n;
+    Py_ssize_t n;
 
-    Py_GetSingleArg("i",n);
+    Py_GetSingleArg(Py_SSIZE_T_PARSERMARKER, n);
     
     v = mxStack_PopMany(stack,n);
     if (v == NULL)
@@ -630,9 +653,9 @@ Py_C_Function( mxStack_as_list,
 Py_C_Function( mxStack_resize,
 	       "resize([size=len(stack)])")
 {
-    int size = mxStack_Length(stack);
+    Py_ssize_t size = _mxStack_Length(stack);
 
-    Py_GetArg("|i",size);
+    Py_GetArg("|"Py_SSIZE_T_PARSERMARKER, size);
     
     if (mxStack_Resize(stack,size))
 	goto onError;
@@ -648,11 +671,12 @@ Py_C_Function( mxStack_resize,
 /* --- slots --- */
 
 static
-int mxStack_Print(mxStackObject *self,
+int mxStack_Print(PyObject *obj,
 		  FILE *fp, 
 		  int flags)
 {
-    int i,top;
+    mxStackObject *self = (mxStackObject *)obj;
+    Py_ssize_t i,top;
     
     top = self->top;
     fprintf(fp, "Stack[");
@@ -669,8 +693,9 @@ int mxStack_Print(mxStackObject *self,
 }
 
 static
-PyObject *mxStack_Repr(mxStackObject *self)
+PyObject *mxStack_Repr(PyObject *obj)
 {
+    mxStackObject *self = (mxStackObject *)obj;
     char s[256];
 
     sprintf(s,"<Stack object at %lx>",(long)self);
@@ -678,11 +703,13 @@ PyObject *mxStack_Repr(mxStackObject *self)
 }
 
 static
-int mxStack_Compare(mxStackObject *v,
-		    mxStackObject *w)
+int mxStack_Compare(PyObject *left,
+		    PyObject *right)
 {
-    int len = min(v->top,w->top) + 1;
-    int i;
+    mxStackObject *v = (mxStackObject *)left;
+    mxStackObject *w = (mxStackObject *)right;
+    Py_ssize_t len = min(v->top,w->top) + 1;
+    Py_ssize_t i;
 
     for (i = 0; i < len; i++) {
 	int cmp = PyObject_Compare(v->array[i],w->array[i]);
@@ -693,16 +720,17 @@ int mxStack_Compare(mxStackObject *v,
 }
 
 static
-PyObject *mxStack_Getattr(mxStackObject *self,
+PyObject *mxStack_Getattr(PyObject *obj,
 			  char *name)
 {
-    return Py_FindMethod(mxStack_Methods,
-			 (PyObject *)self,name);
+    return Py_FindMethod(mxStack_Methods, obj, name);
 }
 
 static
-int mxStack_NonZero(mxStackObject *self)
+int mxStack_NonZero(PyObject *obj)
 {
+    mxStackObject *self = (mxStackObject *)obj;
+
     return self->top >= 0;
 }
 
@@ -737,41 +765,44 @@ int mxStack_Coerce(PyObject **pv,
 }
 
 static
-PyObject *mxStack_LeftShift(mxStackObject *v,
-			    PyObject *w)
+PyObject *mxStack_LeftShift(PyObject *left,
+			    PyObject *right)
 {
+    mxStackObject *self = (mxStackObject *)left;
+
     /* Stack << Object -> (Stack.push(Object),Stack) */
-    if (!_mxStack_Check(v)) {
+    if (!_mxStack_Check(left)) {
 	PyErr_BadInternalCall();
 	return NULL;
     }
-    mxStack_Push(v,w);
-    Py_INCREF(v);
-    return (PyObject *)v;
+    mxStack_Push(self, right);
+    Py_INCREF(self);
+    return left;
 }
 
 static
-PyObject *mxStack_RightShift(mxStackObject *v,
-			     PyObject *w)
+PyObject *mxStack_RightShift(PyObject *left,
+			     PyObject *right)
 {
+    mxStackObject *self = (mxStackObject *)left;
     long n;
     
     /* Stack >> Integer -> Stack.pop_many(Integer) */
-    if (!_mxStack_Check(v)) {
+    if (!_mxStack_Check(left)) {
 	PyErr_BadInternalCall();
 	return NULL;
     }
-    if (!PyInt_Check(w))
+    if (!PyInt_Check(right))
 	Py_Error(PyExc_TypeError,
 		 "right side of >> must an integer");
-    n = PyInt_AS_LONG(w);
+    n = PyInt_AS_LONG(right);
     Py_Assert(n > 0,
 	      PyExc_TypeError,
 	      "you can only pop 1 or more entries");
     if (n == 1)
-	return mxStack_Pop(v);
+	return mxStack_Pop(self);
     else
-	return mxStack_PopMany(v,n);
+	return mxStack_PopMany(self, n);
  onError:
     return NULL;
 }
@@ -780,44 +811,44 @@ PyObject *mxStack_RightShift(mxStackObject *v,
 
 static
 PySequenceMethods mxStack_TypeAsSequence = {
-    (inquiry)mxStack_Length,		/*sq_length*/
-    (binaryfunc)0,			/*sq_concat*/
-    (intargfunc)0,			/*sq_repeat*/
-    (intargfunc)mxStack_GetItem,	/*sq_item*/
-    (intintargfunc)0,			/*sq_slice*/
-    (intobjargproc)0,			/*sq_ass_item*/
-    (intintobjargproc)0,		/*sq_ass_slice*/
+    mxStack_Length,			/*sq_length*/
+    0,					/*sq_concat*/
+    0,					/*sq_repeat*/
+    mxStack_GetItem,			/*sq_item*/
+    0,					/*sq_slice*/
+    0,					/*sq_ass_item*/
+    0,					/*sq_ass_slice*/
 };
 
 static
 PyNumberMethods mxStack_TypeAsNumber = {
 
     /* These slots are not NULL-checked, so we must provide dummy functions */
-    (binaryfunc)notimplemented2,		/*nb_add*/
-    (binaryfunc)notimplemented2,		/*nb_subtract*/
-    (binaryfunc)notimplemented2,		/*nb_multiply*/
-    (binaryfunc)notimplemented2,		/*nb_divide*/
-    (binaryfunc)notimplemented2,		/*nb_remainder*/
-    (binaryfunc)notimplemented2,		/*nb_divmod*/
-    (ternaryfunc)notimplemented3,		/*nb_power*/
-    (unaryfunc)notimplemented1,			/*nb_negative*/
-    (unaryfunc)notimplemented1,			/*nb_positive*/
+    notimplemented2,			/*nb_add*/
+    notimplemented2,			/*nb_subtract*/
+    notimplemented2,			/*nb_multiply*/
+    notimplemented2,			/*nb_divide*/
+    notimplemented2,			/*nb_remainder*/
+    notimplemented2,			/*nb_divmod*/
+    notimplemented3,			/*nb_power*/
+    notimplemented1,			/*nb_negative*/
+    notimplemented1,			/*nb_positive*/
 
     /* Everything below this line EXCEPT nb_nonzero (!) is NULL checked */
-    (unaryfunc)0,				/*nb_absolute*/
-    (inquiry)mxStack_NonZero,			/*nb_nonzero*/
-    (unaryfunc)0,				/*nb_invert*/
-    (binaryfunc)mxStack_LeftShift,		/*nb_lshift*/
-    (binaryfunc)mxStack_RightShift,		/*nb_rshift*/
-    (binaryfunc)notimplemented2,		/*nb_and*/
-    (binaryfunc)notimplemented2,		/*nb_xor*/
-    (binaryfunc)notimplemented2,		/*nb_or*/
-    (coercion)mxStack_Coerce,			/*nb_coerce*/
-    (unaryfunc)0,				/*nb_int*/
-    (unaryfunc)0,				/*nb_long*/
-    (unaryfunc)0,				/*nb_float*/
-    (unaryfunc)0,				/*nb_oct*/
-    (unaryfunc)0,				/*nb_hex*/
+    0,					/*nb_absolute*/
+    mxStack_NonZero,			/*nb_nonzero*/
+    0,					/*nb_invert*/
+    mxStack_LeftShift,			/*nb_lshift*/
+    mxStack_RightShift,			/*nb_rshift*/
+    notimplemented2,			/*nb_and*/
+    notimplemented2,			/*nb_xor*/
+    notimplemented2,			/*nb_or*/
+    mxStack_Coerce,			/*nb_coerce*/
+    0,					/*nb_int*/
+    0,					/*nb_long*/
+    0,					/*nb_float*/
+    0,					/*nb_oct*/
+    0,					/*nb_hex*/
 };
 
 statichere
@@ -829,22 +860,22 @@ PyTypeObject mxStack_Type = {
     0,			  		/*tp_itemsize*/
     /* slots */
     (destructor)mxStack_Free,		/*tp_dealloc*/
-    (printfunc)mxStack_Print,	  	/*tp_print*/
-    (getattrfunc)mxStack_Getattr,  	/*tp_getattr*/
-    (setattrfunc)0,		  	/*tp_setattr*/
-    (cmpfunc)mxStack_Compare,  		/*tp_compare*/
-    (reprfunc)mxStack_Repr,	  	/*tp_repr*/
+    mxStack_Print,	  		/*tp_print*/
+    mxStack_Getattr,  			/*tp_getattr*/
+    0,		  			/*tp_setattr*/
+    mxStack_Compare,  			/*tp_compare*/
+    mxStack_Repr,	  		/*tp_repr*/
     &mxStack_TypeAsNumber, 		/*tp_as_number*/
     &mxStack_TypeAsSequence,		/*tp_as_sequence*/
     0,					/*tp_as_mapping*/
-    (hashfunc)0,			/*tp_hash*/
-    (ternaryfunc)0,			/*tp_call*/
-    (reprfunc)0,			/*tp_str*/
-    (getattrofunc)0, 			/*tp_getattro*/
-    (setattrofunc)0, 			/*tp_setattro*/
+    0,					/*tp_hash*/
+    0,					/*tp_call*/
+    0,					/*tp_str*/
+    0, 					/*tp_getattro*/
+    0, 					/*tp_setattro*/
     0,					/*tp_as_buffer*/
     0,					/*tp_xxx4*/
-    (char*) 0				/*tp_doc*/
+    0					/*tp_doc*/
 };
 
 /* Python Method Table */
@@ -896,6 +927,14 @@ PyMethodDef Module_methods[] =
     {NULL,NULL} /* end of list */
 };
 
+/* Cleanup function */
+static 
+void mxStackModule_Cleanup(void)
+{
+    /* Reset mxStack_Initialized flag */
+    mxStack_Initialized = 0;
+}
+
 /* C API table */
 static
 mxStackModule_APIObject mxStackModuleAPI =
@@ -910,8 +949,8 @@ mxStackModule_APIObject mxStackModuleAPI =
     mxStack_PopMany,
     mxStack_PushMany,
     mxStack_Clear,
-    mxStack_Length,
-    mxStack_GetItem
+    _mxStack_Length,
+    _mxStack_GetItem
 };
 
 /* create PyMethodObjects and register them in the module's dict */
@@ -920,6 +959,10 @@ MX_EXPORT(void)
 {
     PyObject *module, *moddict, *api;
 
+    if (mxStack_Initialized)
+	Py_Error(PyExc_SystemError,
+		 "can't initialize "MXSTACK_MODULE" more than once");
+
     /* Init type objects */
     PyType_Init(mxStack_Type);
 
@@ -932,15 +975,23 @@ MX_EXPORT(void)
     if (module == NULL)
 	goto onError;
 
+    /* Register cleanup function */
+    if (Py_AtExit(mxStackModule_Cleanup)) {
+	/* XXX what to do if we can't register that function ??? */
+	DPRINTF("* Failed to register mxStack cleanup function\n");
+    }
+
     /* Add some constants to the module's dict */
     moddict = PyModule_GetDict(module);
     PyDict_SetItemString(moddict, 
 			 "__version__",
-			 PyString_FromString(VERSION));
+			 PyString_FromString(MXSTACK_VERSION));
 
     /* Errors */
     if (!(mxStack_Error = insexc(moddict,"Error",PyExc_IndexError)))
 	goto onError;
+    if (!(mxStack_EmptyError = insexc(moddict,"EmptyError",mxStack_Error)))
+	goto onError;
 
     /* Type objects */
     Py_INCREF(&mxStack_Type);
@@ -954,6 +1005,9 @@ MX_EXPORT(void)
     PyDict_SetItemString(moddict,MXSTACK_MODULE"API",api);
     Py_DECREF(api);
 
+    /* We are now initialized */
+    mxStack_Initialized = 1;
+
  onError:
     /* Check for errors and report them */
     if (PyErr_Occurred())
