@@ -1,7 +1,7 @@
 """ Tools - Add-ons for Python written in C for performance.
 
     Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 
@@ -460,7 +460,7 @@ def import_code(name,code):
 
 def pairs2tuples(tab):
 
-    """ Format a sequence of adjancent pairs into a list
+    """ Format a sequence of adjacent pairs into a list
         of 2-tuples, e.g. 'abcdef' gives [('a','b'),('c','d'),
         ('e','f')]
     """
@@ -572,7 +572,7 @@ def class_attributes(classobj,of_class=None,
             if isinstance(v,of_class):
                 d[k] = v
     else:
-        d.update(obj.__dict__)
+        d.update(classobj.__dict__)
     return d
 
 def inst_attributes(obj,of_class=None):
@@ -895,7 +895,7 @@ _basemethod_cache = {}
 def basemethod(object,method=None,
 
                cache = _basemethod_cache,InstanceType=types.InstanceType,
-               ClassType=types.ClassType,None=None):
+               ClassType=types.ClassType):
 
     """ Return the unbound method that is defined *after* method in the
         inheritance order of object with the same name as method
@@ -928,7 +928,7 @@ def basemethod(object,method=None,
         objclass = object.im_class
 
     # Check cache
-    cacheentry = (defclass, name)
+    cacheentry = (objclass, name)
     basemethod = cache.get(cacheentry, None)
     if basemethod is not None:
         if not issubclass(objclass, basemethod.im_class):
@@ -1102,7 +1102,7 @@ def _uu_decode(data):
     # Find start of encoded data
     while 1:
         s = readline()
-        print '...',s,
+        #print '...',s,
         if not s:
             raise ValueError, 'Missing "begin" line in input data'
         if s[:5] == 'begin':
