@@ -2,7 +2,7 @@
   mxTools -- Misc. utilities for Python
 
   Copyright (c) 1997-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-  Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+  Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com
   See the documentation for further copyright information or contact
   the author (mailto:mal@lemburg.com).
 
@@ -21,6 +21,9 @@
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXTOOLS
 
+/* Mark the module as Py_ssize_t clean. */
+#define PY_SSIZE_T_CLEAN 1
+
 #include "mx.h"
 #include "mxTools.h"
 
@@ -31,7 +34,7 @@
 /* Needed for mxTools_cur_frame() */
 #include "frameobject.h"
 
-#define VERSION "2.0.2"
+#define MXTOOLS_VERSION "3.0.0"
 
 /* To open up the playground for new ideas... */
 /*#define INCLUDE_FUNSTUFF*/
@@ -43,10 +46,10 @@
 
 static char *Module_docstring = 
 
- MXTOOLS_MODULE" -- A tool collection. Version "VERSION"\n\n"
+ MXTOOLS_MODULE" -- A tool collection. Version "MXTOOLS_VERSION"\n\n"
 
  "Copyright (c) 1997-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com\n"
- "Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
+ "Copyright (c) 2000-2007, eGenix.com Software GmbH; mailto:info@egenix.com\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -57,11 +60,16 @@ static char *Module_docstring =
 static PyObject *mxTools_Error;			/* Error Exception
 						   object */
 
+static PyObject *mxTools_ProgrammingError;	/* ProgrammingError Exception
+						   object */
+
 static PyObject *mxNotGiven;			/* NotGiven singleton */
 
 static PyObject *mxTools_BaseobjAttribute;      /* Attribute to use for
 						   acquire() per default */
 
+/* Flag telling us whether the module was initialized or not. */
+static int mxTools_Initialized = 0;
 
 /* --- forward declarations ----------------------------------------------- */
 
@@ -75,13 +83,37 @@ static PyObject *mxTools_BaseobjAttribute;      /* Attribute to use for
 
 static 
 PyObject *insexc(PyObject *moddict,
-		 char *name)
+		 char *name,
+		 PyObject *baseclass)
 {
     PyObject *v;
     char fullname[256];
+    char *modname;
+    char *dot;
     
-    sprintf(fullname,MXTOOLS_MODULE".%s",name);
-    v = PyErr_NewException(fullname, NULL, NULL);
+    v = PyDict_GetItemString(moddict, "__name__");
+    if (v == NULL)
+	modname = NULL;
+    else
+	modname = PyString_AsString(v);
+    if (modname == NULL) {
+	PyErr_Clear();
+	modname = MXTOOLS_MODULE;
+    }
+    /* The symbols from this extension are imported into
+       mx.<packagename>. We trim the name to not confuse the user with
+       an overly long package path. */
+    strcpy(fullname, modname);
+    dot = strchr(fullname, '.');
+    if (dot)
+	dot = strchr(dot+1, '.');
+    if (dot)
+	strcpy(dot+1, name);
+    else
+	sprintf(fullname, "%s.%s", modname, name);
+
+
+    v = PyErr_NewException(fullname, baseclass, NULL);
     if (v == NULL)
 	return NULL;
     if (PyDict_SetItemString(moddict,name,v))
@@ -196,12 +228,14 @@ Py_C_Function( mxTools_napply,
 	       "Calls the function number_of_calls times with the same\n"
 	       "arguments and returns a tuple with the return values.")
 {
-    int count;
-    register int i;
+    Py_ssize_t count;
+    register Py_ssize_t i;
     PyObject *func,*arg = 0,*kw = 0;
     PyObject *w = 0;
 
-    Py_Get4Args("iO|OO",count,func,arg,kw);
+    Py_Get4Args(Py_SSIZE_T_PARSERMARKER
+		"O|OO",
+		count,func,arg,kw);
 
     Py_XINCREF(arg);
 
@@ -224,7 +258,7 @@ Py_C_Function( mxTools_napply,
 	int flags = PyCFunction_GET_FLAGS(func);
 
 	if (!(flags & METH_VARARGS)) {
-	    int size = PyTuple_GET_SIZE(arg);
+	    Py_ssize_t size = PyTuple_GET_SIZE(arg);
 	    if (size == 1)
 		arg = PyTuple_GET_ITEM(arg, 0);
 	    else if (size == 0)
@@ -280,13 +314,14 @@ Py_C_Function( mxTools_mapply,
 	       "Calls the callable_objects in the given order with the same\n"
 	       "arguments and returns a tuple with the return values.")
 {
-    register int i;
-    int n;
+    register Py_ssize_t i;
+    Py_ssize_t n;
     PyObject *objects;
     PyObject *arg = 0,*kw = 0;
     PyObject *w = 0;
 
-    Py_Get3Args("O|OO",objects,arg,kw);
+    Py_Get3Args("O|OO",
+		objects,arg,kw);
 
     Py_XINCREF(arg);
 
@@ -318,7 +353,7 @@ Py_C_Function( mxTools_mapply,
 	    PyObject *args = arg;
 	    
 	    if (!(PyCFunction_GET_FLAGS(func) & METH_VARARGS)) {
-		int size = PyTuple_GET_SIZE(args);
+		Py_ssize_t size = PyTuple_GET_SIZE(args);
 		if (size == 1)
 		    args = PyTuple_GET_ITEM(args, 0);
 		else if (size == 0)
@@ -383,14 +418,15 @@ Py_C_Function( mxTools_method_mapply,
 	       "order with the same arguments and returns a tuple with\n"
 	       "the return values.")
 {
-    register int i;
-    int n;
+    register Py_ssize_t i;
+    Py_ssize_t n;
     PyObject *objects;
     PyObject *arg = 0,*kw = 0;
     PyObject *w = 0;
     char *methodname;
 
-    Py_Get4Args("Os|OO",objects,methodname,arg,kw);
+    Py_Get4Args("Os|OO",
+		objects,methodname,arg,kw);
 
     Py_XINCREF(arg);
 
@@ -431,7 +467,7 @@ Py_C_Function( mxTools_method_mapply,
 	    PyObject *args = arg;
 	    
 	    if (!(PyCFunction_GET_FLAGS(func) & METH_VARARGS)) {
-		int size = PyTuple_GET_SIZE(args);
+		Py_ssize_t size = PyTuple_GET_SIZE(args);
 		if (size == 1)
 		    args = PyTuple_GET_ITEM(args, 0);
 		else if (size == 0)
@@ -494,13 +530,17 @@ Py_C_Function( mxTools_trange,
 	       "trange([start=0,]stop[,step=1])\n\n"
 	       "Returns tuple(range(start,stop,step))")
 {
-    int start,stop=INT_MAX,step=INT_MAX;
-    int n;
-    register int i;
-    register int index;
+    Py_ssize_t start, stop=INT_MAX, step=INT_MAX;
+    Py_ssize_t n;
+    register Py_ssize_t i;
+    register Py_ssize_t index;
     PyObject *t = 0;
 
-    Py_Get3Args("i|ii",start,stop,step);
+    Py_Get3Args(Py_SSIZE_T_PARSERMARKER
+		"|"
+		Py_SSIZE_T_PARSERMARKER
+		Py_SSIZE_T_PARSERMARKER,
+		start,stop,step);
 
     /* Get the boundaries right... */
     if (stop == INT_MAX) {
@@ -563,8 +603,8 @@ Py_C_Function( mxTools_indices,
 	       "indices(object)\n\n"
 	       "Returns tuple(range(len(object))).")
 {
-    register int n;
-    register int index;
+    register Py_ssize_t n;
+    register Py_ssize_t index;
     register PyObject *t = 0;
     PyObject *w;
 
@@ -597,8 +637,8 @@ Py_C_Function( mxTools_range_len,
 	       "range_len(object)\n"
 	       "Returns range(len(object)).")
 {
-    register int n;
-    register int index;
+    register Py_ssize_t n;
+    register Py_ssize_t index;
     register PyObject *t = 0;
     PyObject *w;
 
@@ -632,13 +672,14 @@ Py_C_Function( mxTools_irange,
 	       "for each item in the sequence indices or, if this is not\n"
 	       "given, in trange(len(object)).")
 {
-    register int n;
-    register int index;
+    register Py_ssize_t n;
+    register Py_ssize_t index;
     PyObject *t = 0;
     PyObject *w;
     PyObject *indices = 0;
 
-    Py_Get2Args("O|O",w,indices);
+    Py_Get2Args("O|O",
+		w,indices);
 
     if (!indices) {
 	n = PyObject_Length(w);
@@ -718,7 +759,8 @@ Py_C_Function( mxTools_get,
     PyObject *x, *w, *index;
     PyObject *def = mxNotGiven;
 
-    Py_Get3Args("OO|O",w,index,def);
+    Py_Get3Args("OO|O",
+		w,index,def);
 
     x = PyObject_GetItem(w,index);
     if (!x) {
@@ -743,14 +785,15 @@ Py_C_Function( mxTools_extract,
 	       "as indices and is used to provide default values in case\n"
 	       "the lookup fails.")
 {
-    int n = 0;
-    register int index;
+    Py_ssize_t n = 0;
+    register Py_ssize_t index;
     PyObject *t = 0;
     PyObject *w;
     PyObject *indices;
     PyObject *defaults = 0;
 
-    Py_Get3Args("OO|O",w,indices,defaults);
+    Py_Get3Args("OO|O",
+		w,indices,defaults);
 
     n = PyObject_Length(indices);
     if (n < 0)
@@ -769,8 +812,8 @@ Py_C_Function( mxTools_extract,
 	    v = PySequence_GetItem(indices,index);
 	    if (!v) {
 		PyErr_Format(PyExc_IndexError,
-			     "index value nr. %i not accessible",
-			     index);
+			     "index value nr. %ld not accessible",
+			     (long)index);
 		goto onError;
 	    }
 	    x = PyObject_GetItem(w,v);
@@ -780,9 +823,10 @@ Py_C_Function( mxTools_extract,
 		PyErr_Clear();
 		x = PySequence_GetItem(defaults,index);
 		if (!x) {
-		    PyErr_Format(PyExc_IndexError,
-			      "default value for index nr. %i not accessible",
-				 index);
+		    PyErr_Format(
+			     PyExc_IndexError,
+			     "default value for index nr. %ld not accessible",
+			     (long)index);
 		    goto onError;
 		}
 	    }
@@ -796,16 +840,16 @@ Py_C_Function( mxTools_extract,
 	    v = PySequence_GetItem(indices,index);
 	    if (!v) {
 		PyErr_Format(PyExc_IndexError,
-			     "index value nr. %i not accessible",
-			     index);
+			     "index value nr. %ld not accessible",
+			     (long)index);
 		goto onError;
 	    }
 	    x = PyObject_GetItem(w,v);
 	    Py_DECREF(v);
 	    if (!x) {
 		PyErr_Format(PyExc_IndexError,
-			     "default value for index nr. %i not accessible",
-			     index);
+			     "default value for index nr. %ld not accessible",
+			     (long)index);
 		goto onError;
 	    }
 	    PyList_SET_ITEM(t,index,x);
@@ -826,12 +870,13 @@ Py_C_Function( mxTools_iremove,
 	       "left in an undefined state)."
 	       )
 {
-    int n = 0;
-    register int index;
+    Py_ssize_t n = 0;
+    register Py_ssize_t index;
     PyObject *w;
     PyObject *indices;
 
-    Py_Get2Args("OO",w,indices);
+    Py_Get2Args("OO",
+		w,indices);
 
     n = PyObject_Length(indices);
     if (n < 0)
@@ -846,8 +891,8 @@ Py_C_Function( mxTools_iremove,
 	    v = PySequence_GetItem(indices,index);
 	    if (!v) {
 		PyErr_Format(PyExc_IndexError,
-			     "index %i not accessible",
-			     index);
+			     "index %ld not accessible",
+			     (long)index);
 		goto onError;
 	    }
 	    rc = PyMapping_DelItem(w,v);
@@ -857,17 +902,17 @@ Py_C_Function( mxTools_iremove,
 	}
 
     else if (PySequence_Check(w)) {
-	int prev_index = INT_MAX;
+	Py_ssize_t prev_index = INT_MAX;
 	
 	for (index = n - 1; index >= 0; index--) {
 	    register PyObject *v;
-	    register int i;
+	    register Py_ssize_t i;
 	    
 	    v = PySequence_GetItem(indices,index);
 	    if (!v && PyInt_Check(v)) {
 		PyErr_Format(PyExc_IndexError,
-			     "index %i not accessible or not an integer",
-			     index);
+			     "index %ld not accessible or not an integer",
+			     (long)index);
 		goto onError;
 	    }
 	    i = PyInt_AS_LONG(v);
@@ -900,15 +945,16 @@ Py_C_Function( mxTools_ifilter,
 	       "the sequence indices (defaulting to indices(object)).\n"
 	       "Order is preserved. condition must be a callable object.")
 {
-    int n;
-    register int index;
+    Py_ssize_t n;
+    register Py_ssize_t index;
     PyObject *t = 0;
     PyObject *w;
     PyObject *indices = 0;
     PyObject *condition;
     PyObject *argtuple = 0;
 
-    Py_Get3Args("OO|O",condition,w,indices);
+    Py_Get3Args("OO|O",
+		condition,w,indices);
 
     if (!indices) {
 	n = PyObject_Length(w);
@@ -1031,9 +1077,9 @@ Py_C_Function( mxTools_tuples,
 	       "from the other sequences are filled in with None."
 	       )
 {
-    int n,m;
-    register int i;
-    register int j;
+    Py_ssize_t n,m;
+    register Py_ssize_t i;
+    register Py_ssize_t j;
     PyObject *l = 0;
     PyObject *arg,*w;
 
@@ -1113,9 +1159,9 @@ Py_C_Function( mxTools_lists,
 	       "Same as tuples(), except that a tuple of lists is created."
 	       )
 {
-    int n,m;
-    register int i;
-    register int j;
+    Py_ssize_t n,m;
+    register Py_ssize_t i;
+    register Py_ssize_t j;
     PyObject *t = 0;
     PyObject *arg,*w;
 
@@ -1199,11 +1245,12 @@ Py_C_Function( mxTools_count,
     PyObject *condition;
     PyObject *list;
     PyObject *argtuple = 0;
-    register int i;
-    register int n;
-    int length;
+    register Py_ssize_t i;
+    register Py_ssize_t n;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",condition,list);
+    Py_Get2Args("OO",
+		condition,list);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1251,11 +1298,12 @@ Py_C_Function( mxTools_exists,
     PyObject *condition;
     PyObject *list;
     PyObject *argtuple = 0;
-    register int i;
-    int n;
-    int length;
+    register Py_ssize_t i;
+    Py_ssize_t n;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",condition,list);
+    Py_Get2Args("OO",
+		condition,list);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1306,11 +1354,12 @@ Py_C_Function( mxTools_forall,
     PyObject *condition;
     PyObject *list;
     PyObject *argtuple = 0;
-    register int i;
-    int n;
-    int length;
+    register Py_ssize_t i;
+    Py_ssize_t n;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",condition,list);
+    Py_Get2Args("OO",
+		condition,list);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1361,11 +1410,12 @@ Py_C_Function( mxTools_index,
     PyObject *condition;
     PyObject *list = 0;
     PyObject *argtuple = 0;
-    register int i;
-    int n;
-    int length;
+    register Py_ssize_t i;
+    Py_ssize_t n;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",condition,list);
+    Py_Get2Args("OO",
+		condition,list);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1416,7 +1466,7 @@ Py_C_Function( mxTools_sizeof,
 	       "Note that this doesn't show any extra space allocated by\n"
 	       "the object.")
 {
-    int size;
+    Py_ssize_t size;
     register PyObject *w;
     PyTypeObject *t;
 
@@ -1438,10 +1488,11 @@ Py_C_Function( mxTools_findattr,
 {
     PyObject *list;
     PyObject *name;
-    register int i;
-    int length;
+    register Py_ssize_t i;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",list,name);
+    Py_Get2Args("OO",
+		list,name);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1482,10 +1533,11 @@ Py_C_Function( mxTools_attrlist,
 {
     PyObject *list,*l=0;
     PyObject *name;
-    register int i;
-    int length;
+    register Py_ssize_t i;
+    Py_ssize_t length;
 
-    Py_Get2Args("OO",list,name);
+    Py_Get2Args("OO",
+		list,name);
 
     length = PySequence_Length(list);
     if (length < 0)
@@ -1532,9 +1584,9 @@ Py_C_Function( mxTools_dict,
 	       "the first entry being interpreted as the key and the second as\n"
 	       "the value.")
 {
-    int n;
+    Py_ssize_t n;
     PyObject *seq,*d = 0;
-    register int i;
+    register Py_ssize_t i;
     register PyObject *k = 0;
     register PyObject *v = 0;
 
@@ -1560,7 +1612,8 @@ Py_C_Function( mxTools_dict,
 
 	if (!k || !v) {
 	    PyErr_Format(PyExc_TypeError,
-			 "item %i in sequence doesn't have two entries",i);
+			 "item %ld in sequence doesn't have two entries",
+			 (long)i);
 	    goto onError;
 	}
 	if (PyDict_SetItem(d,k,v))
@@ -1584,11 +1637,12 @@ Py_C_Function( mxTools_setdict,
 	       "dictionary keys. The values are all set to value."
 	       )
 {
-    int n;
+    Py_ssize_t n;
     PyObject *seq,*d = 0,*value = 0;
-    register int i;
+    register Py_ssize_t i;
 
-    Py_Get2Args("O|O",seq,value);
+    Py_Get2Args("O|O",
+		seq,value);
     n = PySequence_Length(seq);
     if (n < 0)
 	Py_Error(PyExc_TypeError,
@@ -1630,7 +1684,7 @@ Py_C_Function( mxTools_invdict,
 	       "Creates a dictionary with inverse mappings from the\n"
 	       "given dictionary d.")
 {
-    int pos;
+    Py_ssize_t pos;
     PyObject *key,*value;
     PyObject *d,*inv = 0;
 
@@ -1666,8 +1720,8 @@ Py_C_Function( mxTools_reverse,
     Py_GetArgObject(seq);
 
     if (PyTuple_Check(seq)) {
-	register int i;
-	register int n = PyTuple_GET_SIZE(seq);
+	register Py_ssize_t i;
+	register Py_ssize_t n = PyTuple_GET_SIZE(seq);
 	
 	w = PyTuple_New(n);
 	if (!w)
@@ -1683,8 +1737,8 @@ Py_C_Function( mxTools_reverse,
 	}
     }
     else if (PyList_Check(seq)) {
-	register int i;
-	register int n = PyList_GET_SIZE(seq);
+	register Py_ssize_t i;
+	register Py_ssize_t n = PyList_GET_SIZE(seq);
 	
 	w = PyList_New(n);
 	if (!w)
@@ -1700,8 +1754,8 @@ Py_C_Function( mxTools_reverse,
 	}
     }
     else {
-	register int i;
-	register int n = PySequence_Length(seq);
+	register Py_ssize_t i;
+	register Py_ssize_t n = PySequence_Length(seq);
 
 	if (n < 0)
 	    Py_Error(PyExc_TypeError,
@@ -1717,7 +1771,8 @@ Py_C_Function( mxTools_reverse,
 	    o = PySequence_GetItem(seq,i);
 	    if (!o)
 		PyErr_Format(PyExc_IndexError,
-			     "item %i of sequence",i);
+			     "item %ld of sequence",
+			     (long)i);
 	    Py_INCREF(o);
 	    i++;
 	    PyList_SET_ITEM(w,n-i,o);
@@ -1740,12 +1795,12 @@ void mxTools_free(void *p)
 Py_C_Function( mxTools_malloc,
 	       "malloc(x)\n\n")
 {
-    int x;
+    Py_ssize_t x;
     char *p;
-    register int i;
+    register Py_ssize_t i;
     register char *c;
 
-    Py_GetSingleArg("i",x);
+    Py_GetSingleArg(Py_SSIZE_T_PARSERMARKER, x);
 
     /* Allocate virtual memory */
     p = (char *)malloc(x);
@@ -1771,7 +1826,7 @@ Py_C_Function( mxTools_caching_eval,
 {
     PyObject *codestr;
 
-    Py_GetArg("O",codestr);
+    Py_GetArg("O", codestr);
 
     Py_Assert(PyString_Check(codestr),
 	      PyExc_TypeError,
@@ -1802,7 +1857,8 @@ Py_C_Function( mxTools_acquire,
 	      PyExc_SystemError,
 	      "maximum acquire() recursion depth exceeded");
 
-    Py_Get3Args("OO|O",obj,name,baseobjattr);
+    Py_Get3Args("OO|O",
+		obj,name,baseobjattr);
 
     Py_Assert(PyString_Check(name),
 	      PyExc_TypeError,
@@ -1842,7 +1898,7 @@ Py_C_Function( mxTools_verbosity,
     int value = Py_VerboseFlag,
 	old_value = value;
 
-    Py_GetArg("|i",value);
+    Py_GetArg("|i", value);
     Py_VerboseFlag = value;
 
     return PyInt_FromLong((long)old_value);
@@ -1861,7 +1917,7 @@ Py_C_Function( mxTools_debugging,
     int value = Py_DebugFlag,
 	old_value = value;
 
-    Py_GetArg("|i",value);
+    Py_GetArg("|i", value);
     Py_DebugFlag = value;
 
     return PyInt_FromLong((long)old_value);
@@ -1880,7 +1936,7 @@ Py_C_Function( mxTools_optimization,
     int value = Py_OptimizeFlag,
 	old_value = value;
 
-    Py_GetArg("|i",value);
+    Py_GetArg("|i", value);
     Py_OptimizeFlag = value;
 
     return PyInt_FromLong((long)old_value);
@@ -1889,6 +1945,25 @@ Py_C_Function( mxTools_optimization,
     return NULL;
 }
 
+Py_C_Function( mxTools_interactive,
+	       "interactive([level])\n\n"
+	       "Sets the value of the interpreter's interactive flag.\n"
+	       "Returns the flag's value before changing it or, when called\n"
+	       "without level, the current value."
+	       )
+{
+    int value = Py_InteractiveFlag,
+	old_value = value;
+
+    Py_GetArg("|i", value);
+    Py_InteractiveFlag = value;
+
+    return PyInt_FromLong((long)old_value);
+
+ onError:
+    return NULL;
+}
+
 Py_C_Function( mxTools_cur_frame,
 	       "cur_frame([offset=0])\n\n"
 	       "Returns the current execution frame, optionally going up the\n"
@@ -1899,9 +1974,9 @@ Py_C_Function( mxTools_cur_frame,
 {
     PyFrameObject *frame = NULL;
     PyObject *v;
-    int offset = 0;
+    Py_ssize_t offset = 0;
 
-    Py_GetArg("|i",offset);
+    Py_GetArg("|"Py_SSIZE_T_PARSERMARKER, offset);
 
     /* Get requested frame object */
     frame = (PyFrameObject *)PyEval_GetFrame();
@@ -1941,7 +2016,7 @@ Py_C_Function( mxTools_issequence,
     int rc;
     PyObject *v;
 
-    Py_GetArg("O",v);
+    Py_GetArg("O", v);
     if (PyInstance_Check(v)) {
 	PyObject *w;
 	
@@ -1975,7 +2050,7 @@ Py_C_Function( mxTools_truth,
     PyObject *obj;
     int istrue;
     
-    Py_GetArg("O",obj);
+    Py_GetArg("O", obj);
 
     istrue = PyObject_IsTrue(obj);
     if (istrue < 0)
@@ -2000,7 +2075,7 @@ Py_C_Function( mxTools_sign,
     PyObject *obj,*neg = 0;
     int sign;
     
-    Py_GetArg("O",obj);
+    Py_GetArg("O", obj);
 
     neg = PyNumber_Negative(obj);
     if (!neg) 
@@ -2024,10 +2099,10 @@ Py_C_Function( mxTools_makeref,
 	       "USE WITH CARE - since this can cause core dumps !"
 	       )
 {
-    int id;
+    long id;
     PyObject *obj;
     
-    Py_GetArg("i",id);
+    Py_GetArg("l", id);
 
     /* Accessing the memory location at id can cause a core dump ! */
     obj = (PyObject *)id;
@@ -2060,10 +2135,11 @@ Py_C_Function( mxTools_dictscan,
 	       "range."
 	       )
 {
-    int pos = 0;
+    Py_ssize_t pos = 0;
     PyObject *d,*key,*value;
     
-    Py_Get2Args("O|i",d,pos);
+    Py_Get2Args("O|"Py_SSIZE_T_PARSERMARKER,
+		d, pos);
 
     Py_Assert(PyDict_Check(d),
 	      PyExc_TypeError,
@@ -2072,7 +2148,8 @@ Py_C_Function( mxTools_dictscan,
     if (!PyDict_Next(d,&pos,&key,&value))
 	Py_Error(PyExc_IndexError,
 		 "end of scan or illegal index");
-    Py_Return3Args("OOi",key,value,pos);
+    Py_Return3Args("OO"Py_SSIZE_T_PARSERMARKER,
+		   key,value,pos);
 
  onError:
     return NULL;
@@ -2082,13 +2159,13 @@ Py_C_Function( mxTools_dictscan,
 
 static
 int parselevel(char *s,
-	       int len,
-	       int start,
+	       Py_ssize_t len,
+	       Py_ssize_t start,
 	       int *number,
 	       char *rest)
 {
-    int i;
-    int split = -1, number_len;
+    Py_ssize_t i;
+    Py_ssize_t split = -1, number_len;
     char buffer[256];
 
     for (i = start; i < len; i++) {
@@ -2105,12 +2182,12 @@ int parselevel(char *s,
 	split = i;
     }
     else {
-	int rest_len = i - split;
+	Py_ssize_t rest_len = i - split;
 	memcpy(rest,&s[split],rest_len);
 	rest[rest_len] = '\0';
     }
     number_len = split - start;
-    if (0 < number_len < sizeof(buffer)) {
+    if (0 < number_len && number_len < sizeof(buffer)) {
 	memcpy(buffer,&s[start],number_len);
 	buffer[number_len] = '\0';
 	*number = atoi(buffer);
@@ -2127,13 +2204,14 @@ Py_C_Function( mxTools_verscmp,
 	       )
 {
     char *a,*b;
-    int a_len,b_len;
-    int a_i,b_i;
+    Py_ssize_t a_len,b_len;
+    Py_ssize_t a_i,b_i;
     int a_n,b_n;
     char a_x[256],b_x[256];
     int cmp = 0;
     
-    Py_Get4Args("s#s#",a,a_len,b,b_len);
+    Py_Get4Args("s#s#",
+		a, a_len, b, b_len);
 
     Py_Assert(a_len < 255 && b_len < 255,
 	      PyExc_TypeError,
@@ -2142,8 +2220,8 @@ Py_C_Function( mxTools_verscmp,
     DPRINTF("mxTools_verscmp: a: '%s' b: '%s'\n",a,b);
 
     for (a_i = b_i = 0;;) {
-	a_i = parselevel(a,a_len,a_i,&a_n,a_x);
-	b_i = parselevel(b,b_len,b_i,&b_n,b_x);
+	a_i = parselevel(a, a_len, a_i, &a_n, a_x);
+	b_i = parselevel(b, b_len, b_i, &b_n, b_x);
 	DPRINTF("mxTools_verscmp: a: %i %i '%s'\n",a_i,a_n,a_x);
 	DPRINTF("mxTools_verscmp: b: %i %i '%s'\n",b_i,b_n,b_x);
 
@@ -2187,6 +2265,178 @@ Py_C_Function( mxTools_verscmp,
     return NULL;
 }
 
+/* --- Experimental safecall() API ---------------------------------- */
+
+/*#define MXTOOLS_ENABLE_SAFECALL*/
+
+#ifdef MXTOOLS_ENABLE_SAFECALL
+
+#include <setjmp.h>
+#include <signal.h>
+
+/* safecall() environment
+
+   This is *not* thread-safe, but then signals don't work reliably in
+   the presence of threads anyway.
+
+   XXX sigsetjmp et al. are not available on Windows and probably
+       other platforms as well !
+
+*/
+static int safecall_environment_initialized = 0;
+static sigjmp_buf safecall_environment;
+
+/* safecall() signal handler */
+
+static void safecall_handler(int signum) 
+{
+    DPRINTF("SIGSEGV handler called\n");
+
+    /* Force SysV behavior: reset the signal handler to the default
+       handler */
+    signal(signum, SIG_DFL);
+
+    if (safecall_environment_initialized)
+	/* Jump back to the environment saved in the safecall function. */
+	siglongjmp(safecall_environment, signum);
+    else
+	/* Reraise the signal (forcing the use of the default handler) */
+	raise(signum);
+}
+
+/* Initialize the safecall_handler
+
+   Note that a signal that invokes the handler will automatically
+   cause the default handler to be installed again for the signal 
+
+*/
+static int init_safecall_handler(void) 
+{
+    DPRINTF("init SIGSEGV handler\n");
+    signal(SIGSEGV, safecall_handler);
+    signal(SIGBUS, safecall_handler);
+    signal(SIGABRT, safecall_handler);
+    signal(SIGILL, safecall_handler);
+    signal(SIGFPE, safecall_handler);
+    return 0;
+}
+
+
+/* Reset the signal handlers to their defaults */
+
+static int reset_safecall_handler(void) 
+{
+    DPRINTF("reset SIGSEGV handler\n");
+    signal(SIGSEGV, SIG_DFL);
+    signal(SIGBUS, SIG_DFL);
+    signal(SIGABRT, SIG_DFL);
+    signal(SIGILL, SIG_DFL);
+    signal(SIGFPE, SIG_DFL);
+    return 0;
+}
+
+Py_C_Function( mxTools_safecall,
+	       "safecall(callable [,args, kws])\n\n"
+	       "Calls the callable object callable using args and kws,\n"
+               "if given. In the event of the call causing a segfault\n"
+	       "or similar serious problem, the function will raise an\n"
+	       "mx.Tools.ProgrammingError. The application can then terminate\n"
+	       "gracefully. Note: This function is not thread-safe !"
+	       )
+{
+    PyObject *callable, *callargs=NULL, *callkws=NULL;
+    PyObject *result = NULL;
+    PyObject *v;
+    int signum;
+
+    Py_Get3Args("O|OO", 
+		callable, callargs, callkws);
+
+    /* Save the current stack environment and call the callable */
+    DPRINTF("setjmp()\n");
+    signum = sigsetjmp(safecall_environment, 1);
+
+    if (signum == 0) {
+	/* Call the callable object and return gracefully if
+	   possible */
+	safecall_environment_initialized = 1;
+	init_safecall_handler();
+	result = PyEval_CallObjectWithKeywords(callable, callargs, callkws);
+	safecall_environment_initialized = 0;
+	reset_safecall_handler();
+	return result;
+    }
+
+    /* We got here by a longjmp(); signum is the signal triggering
+       this event. */
+    DPRINTF("processing longjmp()\n");
+    safecall_environment_initialized = 0;
+    reset_safecall_handler();
+    v = PyTuple_New(2);
+    PyTuple_SET_ITEM(v, 0, PyString_FromString("Bug in external routine"));
+    PyTuple_SET_ITEM(v, 1, PyInt_FromLong(signum));
+    PyErr_SetObject(mxTools_ProgrammingError, v);
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxTools_segfault,
+	       "segfault()\n\n"
+	       "Cause a hard seg fault."
+	       )
+{
+    char *p = NULL;
+    char buffer[10];
+    double a,b,c;
+    int errtype = 0;
+    
+    Py_GetArg("|i", errtype);
+
+    switch (errtype) {
+
+    case 0:
+	/* Cause a true seg fault by writing to a NULL pointer */
+	memset(p, 0, 100);
+	break;
+
+    case 1:
+	/* Double free() */
+	p = malloc(100);
+	free(p);
+	free(p);
+	break;
+
+    case 2:
+	/* Buffer overrun on stack */
+	memset(buffer, 0, 20);
+	break;
+	
+    case 3:
+	/* Buffer underrun on stack */
+	memset(buffer - 10, 0, 20);
+	break;
+
+    case 4:
+	/* Division by 0 */
+	a = 1;
+	b = 0;
+	c = a / b;
+	break;
+	
+    default:
+	Py_Error(PyExc_ValueError,
+		 "segfault(): unknown error type");
+	
+    }
+    Py_ReturnNone();
+
+ onError:
+    return NULL;
+}
+
+#endif
+
 /* --- NotGiven Singleton -------------------------------------------------- */
 
 static 
@@ -2218,38 +2468,38 @@ PyNumberMethods mxNotGiven_TypeAsNumber = {
     notimplemented1,				/*nb_positive*/
 
     /* Everything below this line EXCEPT nb_nonzero (!) is NULL checked */
-    (unaryfunc)0,				/*nb_absolute*/
+    0,						/*nb_absolute*/
     mxNotGiven_NonZero,				/*nb_nonzero*/
-    (unaryfunc)0,				/*nb_invert*/
-    (binaryfunc)0,				/*nb_lshift*/
-    (binaryfunc)0,				/*nb_rshift*/
-    (binaryfunc)0,				/*nb_and*/
-    (binaryfunc)0,				/*nb_xor*/
-    (binaryfunc)0,				/*nb_or*/
-    (coercion)0,				/*nb_coerce*/
-    (unaryfunc)0,				/*nb_int*/
-    (unaryfunc)0,				/*nb_long*/
-    (unaryfunc)0,				/*nb_float*/
-    (unaryfunc)0,				/*nb_oct*/
-    (unaryfunc)0,				/*nb_hex*/
+    0,						/*nb_invert*/
+    0,						/*nb_lshift*/
+    0,						/*nb_rshift*/
+    0,						/*nb_and*/
+    0,						/*nb_xor*/
+    0,						/*nb_or*/
+    0,						/*nb_coerce*/
+    0,						/*nb_int*/
+    0,						/*nb_long*/
+    0,						/*nb_float*/
+    0,						/*nb_oct*/
+    0,						/*nb_hex*/
 };
 
 static PyTypeObject mxNotGiven_Type = {
-    PyObject_HEAD_INIT(0)		/* init at startup ! */
-    0,		/*ob_size*/
-    "NotGiven",	/*tp_name*/
-    sizeof(mxNotGivenObject),		/*tp_basicsize*/
-    0,		/*tp_itemsize*/
-    0,		/*tp_dealloc; object always stays alive */
-    0,		/*tp_print*/
-    0,		/*tp_getattr*/
-    0,		/*tp_setattr*/
-    0,		/*tp_compare*/
-    mxNotGiven_Repr, /*tp_repr*/
-    &mxNotGiven_TypeAsNumber,		/*tp_as_number*/
-    0,		/*tp_as_sequence*/
-    0,		/*tp_as_mapping*/
-    0,		/*tp_hash */
+    PyObject_HEAD_INIT(0)			/* init at startup ! */
+    0,						/*ob_size*/
+    "NotGiven",					/*tp_name*/
+    sizeof(mxNotGivenObject),			/*tp_basicsize*/
+    0,						/*tp_itemsize*/
+    0,						/*tp_dealloc; object always stays alive */
+    0,						/*tp_print*/
+    0,						/*tp_getattr*/
+    0,						/*tp_setattr*/
+    0,						/*tp_compare*/
+    mxNotGiven_Repr, 				/*tp_repr*/
+    &mxNotGiven_TypeAsNumber,			/*tp_as_number*/
+    0,						/*tp_as_sequence*/
+    0,						/*tp_as_mapping*/
+    0,						/*tp_hash */
 };
 
 /* --- module init --------------------------------------------------------- */
@@ -2289,6 +2539,7 @@ PyMethodDef Module_methods[] =
     Py_MethodListEntry("acquire",mxTools_acquire),
     Py_MethodListEntry("verbosity",mxTools_verbosity),
     Py_MethodListEntry("optimization",mxTools_optimization),
+    Py_MethodListEntry("interactive",mxTools_interactive),
     Py_MethodListEntry("debugging",mxTools_debugging),
     Py_MethodListEntry("cur_frame",mxTools_cur_frame),
     Py_MethodListEntry("truth",mxTools_truth),
@@ -2299,6 +2550,10 @@ PyMethodDef Module_methods[] =
 #ifdef INCLUDE_FUNSTUFF
     Py_MethodListEntrySingleArg("malloc",mxTools_malloc),
     Py_MethodListEntry("caching_eval",mxTools_caching_eval),
+#endif
+#ifdef MXTOOLS_ENABLE_SAFECALL
+    Py_MethodListEntry("safecall",mxTools_safecall),
+    Py_MethodListEntry("segfault",mxTools_segfault),
 #endif
     {NULL,NULL} /* end of list */
 };
@@ -2308,6 +2563,10 @@ static
 void mxToolsModule_Cleanup(void)
 {
     Py_XDECREF(mxTools_BaseobjAttribute);
+    mxTools_BaseobjAttribute = 0;
+
+    /* Reset mxTools_Initialized flag */
+    mxTools_Initialized = 0;
 }
 
 /* create PyMethodObjects and register them in the module's dict */
@@ -2316,6 +2575,10 @@ MX_EXPORT(void)
 {
     PyObject *module, *moddict;
 
+    if (mxTools_Initialized)
+	Py_Error(PyExc_SystemError,
+		 "can't initialize "MXTOOLS_MODULE" more than once");
+
     /* Init type objects */
     PyType_Init(mxNotGiven_Type);
 
@@ -2344,14 +2607,20 @@ MX_EXPORT(void)
 
     /* Add some constants to the module's dict */
     moddict = PyModule_GetDict(module);
-    insstr(moddict,"__version__",VERSION);
+    insstr(moddict,"__version__",MXTOOLS_VERSION);
     PyDict_SetItemString(moddict,
 			 "NotGiven",
 			 mxNotGiven);
 
     /* Errors */
-    if (!(mxTools_Error = insexc(moddict,"Error")))
+    if (!(mxTools_Error = insexc(moddict, "Error", NULL)))
 	goto onError;
+    if (!(mxTools_ProgrammingError = insexc(moddict, "ProgrammingError",
+					    PyExc_RuntimeError)))
+	goto onError;
+
+    /* We are now initialized */
+    mxTools_Initialized = 1;
 
  onError:
     /* Check for errors and report them */
